/***************************************************************************
 *   Copyright (C) 2008 by Muhammad Aamir Cheema   *
 *   macheema@cse.unsw.edu.au   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "CEquation.h"

#include <limits>
#include <string>
#include <math.h>
#include <float.h>



using namespace std;

CEquation::CEquation(double* coeff)
{
  memcpy(coefficient, coeff, sizeof(double) * (DIMENSION+1));
}

CEquation::CEquation(CPoint* q, CPoint *p)
{
  double m;  // m will store coefficient of dimension i of the midpoint
  //double N; // N will store coefficient of dimension i of 
             // normal vector i.e; normal to the bisector of coord1,coord2
             // We don't need N because we directly store it's value in coefficient[i]
             // see for loop below 
  
  //Vector equation for bisector is N.x=N.m, where N is a vector for line P joining points coord1 and
  // coord2 and m is the midpoint m[1],m[2] and so on. x is (x[1],x[2],x[3],...)
  // So coefficient of x[1] is N[1], coefficient of x[2] is N[2] and so on
  // coefficient of c is N.m ==> N[1].m[1] + N[2].m[2] + ...
  
  coefficient[DIMENSION] = 0;
  for(int i=0; i<DIMENSION;i++){
    m = (q->coord[i] + p->coord[i])/2;
    //N = coord1[i] - coord2[i];
    //coefficient[i] = N; 
    coefficient[i] = q->coord[i] - p->coord[i];
    coefficient[DIMENSION] += coefficient[i]*m;  // c
  }
}

CEquation::CEquation(CPoint* p1, CPoint* p2, bool temp)
{
/*
for (x1,y1) and (x2,y2) a line between two points will be

(x2-x1)/(x1-x) = (y2-y1)/(y1-y)
=> (x2-x1)*(y1-y)=(y2-y1)*(x1-x)
=>(x2-x1)y1-(x2-x1)y = (y2-y1)x1-(y2-y1)x
=>(y2-y1)x + (x1-x2)y =(y2-y1)x1 + (x1-x2)y1

A = (y2-y1)
B = (x1-x2)
C = x1*A + y1*B
*/
 /* coefficient[0]=p2->coord[1] - p1->coord[1];
  coefficient[1]=p1->coord[0] - p2->coord[0];
  coefficient[2]=p1->coord[0]*coefficient[0]+p1->coord[1]*coefficient[1];*/
  if(temp){
    coefficient[0]=p2->coord[1] - p1->coord[1];
    coefficient[1]=p1->coord[0] - p2->coord[0];
    coefficient[2]=p1->coord[0]*coefficient[0]+p1->coord[1]*coefficient[1];  ///Original one Ax+By=C
 }else
  {
    coefficient[0]=p2->coord[1] - p1->coord[1];
    coefficient[1]=p1->coord[0] - p2->coord[0];
    coefficient[2]=-1*(p1->coord[0]*coefficient[0]+p1->coord[1]*coefficient[1]);  ///Changed  Ax+By+C=0
 }

}

CEquation::CEquation(CPoint* q, CPoint *p, CPoint* passThrough)
{
  coefficient[DIMENSION] = 0;
  for(int i=0; i<DIMENSION;i++){
    coefficient[i] = q->coord[i] - p->coord[i];
    coefficient[DIMENSION]+=coefficient[i]*passThrough->coord[i];  // c
  }
}




CEquation::CEquation(CPoint* p, float angle, bool temp)
{
 // RETURNS in form Ax + BX  + C = 0;
 if(temp)
 {
   float slope = tan(angle*PI/180);
  coefficient[1] = 1;
  coefficient[0] = -1*slope;
  coefficient[2] = -1*(slope*p->coord[0] - p->coord[1]);     //Ax+Bx=C
 }
 else{
    float slope = tan(angle*PI/180);
  coefficient[1] = 1;
  coefficient[0] = -1*slope;
  coefficient[2] = (slope*p->coord[0] - p->coord[1]);      // Original one Ax + BX  + C = 0;
}
}


CEquation::~CEquation(void)
{
  // we do not need to do the deletion of array because array is not dynamic
// //   delete[] coefficient;
}


double CEquation::getCoeff(int index)
{
  return coefficient[index];
}

void CEquation::print(void)
{
  for(int i=0; i<DIMENSION; i++)
  {
    if (coefficient[i]>=0)
      cout << " +" << coefficient[i] << "X" << i ;
    else
      cout << " " << coefficient[i] << "X" << i ;
  }
  cout << " >= " << coefficient[DIMENSION] << endl;
}

void CEquation::moveToOrigin(void)
{
   coefficient[DIMENSION] = 0;
}

void CEquation::fixZeroes(void)
{
  for(int i=0; i<DIMENSION; i++)
  {
    if(coefficient[i]==0)
      coefficient[i] = 0.00001;
  }
}

double CEquation::solveFor(double* solveFor)
{
  // Let equation be Ax + By + Cz = 0
  // unkown value z is ==>  (-Ax - By)/C

  double unknown=0; // The coefficient of unknown dimension
  double sum=0;  // 

  for(int i=0; i<DIMENSION-1; i++)
  {
    sum = sum + this->coefficient[i]*solveFor[i];
  }

  unknown = -1*sum/this->coefficient[DIMENSION-1]; 
  return unknown;
}

bool CEquation::liesInSpace(double* solveFor)
{
  // let the space defined by Ax + By + Cz < 0
  // to check whether a point (x,y,z) lies in the plane or not is trivial 
  // put the value of x,y and z in equation and if it is lesser than 0, it lies
  // in space otherwise it doesn't

  double res=0;

  for(int i=0; i<DIMENSION; i++){
    res = res + this->coefficient[i]*solveFor[i];
  }

  if(res<0){
//     cout << "res is " << res << endl;
    return true;

  }
  else{
    return false;
  }
}



CPoint* CEquation::getIntersection(CEquation* line)
{
//   Line-Line Intersection
// Regardless of how the lines are specified, you should be able to generate two different points along the line, and then generate A, B and C. Now, lets say that you have lines, given by the equations:
//           A1x + B1y = C1
//           A2x + B2y = C2
//           To find the point at which the two lines intersect, we simply need to solve the two equations for the two unknowns, x and y.
// 
//           double det = A1*B2 - A2*B1
//                 if(det == 0){
//         //Lines are parallel
//                 }else{
//                   double x = (B2*C1 - B1*C2)/det
//                         double y = (A1*C2 - A2*C1)/det
//                 }

  CPoint* intersection = new CPoint();
  intersection->coord[0] = -1;
  intersection->coord[1] = -1;
  
  // Since equation should be in form Ax + Bx  = C and our equation is Ax + Bx + C = 0
  float A1 = this->coefficient[0] ;
  float B1 = this->coefficient[1] ;
  float C1 = this->coefficient[2]* -1;
  
  float A2 = line->coefficient[0] ;
  float B2 = line->coefficient[1] ;
  float C2 = line->coefficient[2]* -1;
  

  float det = A1*B2 - A2*B1;
  if(det==0)
    return intersection;
  
  intersection->coord[0] = (B2*C1 -B1*C2)/det;
  intersection->coord[1] = (A1*C2 - A2*C1)/det;
  
  return intersection;

}



bool CEquation::pliesInSpace(CPoint *p)
{
  // let the space defined by Ax + By + Cz < 0
  // to check whether a point p lies in the plane or not is trivial 
  // put the value of x,y and z in equation and if it is lesser than 0, it lies
  // in space otherwise it doesn't

  double res=-1*coefficient[2];

  for(int i=0; i<DIMENSION; i++){
    res = res + this->coefficient[i]*p->coord[i]; 
  }

  if(res<0){
//     cout << "res is " << res << endl;
    return true;

  }
  else{
    return false;
  }
}


CPoint* CEquation::IntersectionPoint(CEquation *eq)
{
/* One of the most common tasks you will find in geometry problems is line intersection. Despite the fact that it is so common, a lot of coders still have trouble with it. The first question is, what form are we given our lines in, and what form would we like them in? Ideally, each of our lines will be in the form Ax+By=C, where A, B and C are the numbers which define the line. However, we are rarely given lines in this format, but we can easily generate such an equation from two points. Say we are given two different points, (x1, y1) and (x2, y2), and want to find A, B and C for the equation above. We can do so by setting
A = y2-y1
B = x1-x2
C = A*x1+B*y1
Regardless of how the lines are specified, you should be able to generate two different points along the line, and then generate A, B and C. Now, lets say that you have lines, given by the equations:
A1x + B1y = C1
A2x + B2y = C2
To find the point at which the two lines intersect, we simply need to solve the two equations for the two unknowns, x and y.
    double det = A1*B2 - A2*B1
    if(det == 0){
        //Lines are parallel
    }else{
        double x = (B2*C1 - B1*C2)/det
        double y = (A1*C2 - A2*C1)/det
    }
To see where this comes from, consider multiplying the top equation by B2, and the bottom equation by B1. This gives you
A1B2x + B1B2y = B2C1
A2B1x + B1B2y = B1C2
Now, subtract the bottom equation from the top equation to get
A1B2x - A2B1x = B2C1 - B1C2
Finally, divide both sides by A1B2 - A2B1, and you get the equation for x. The equation for y can be derived similarly.
This gives you the location of the intersection of two lines, but what if you have line segments, not lines. In this case, you need to make sure that the point you found is on both of the line segments. If your line segment goes from (x1,y1) to (x2,y2), then to check if (x,y) is on that segment, you just need to check that min(x1,x2) ≤ x ≤ max(x1,x2), and do the same thing for y. You must be careful about double precision issues though. If your point is right on the edge of the segment, or if the segment is horizontal or vertical, a simple comparison might be problematic. In these cases, you can either do your comparisons with some tolerance, or else use a fraction class.
*/
  CPoint* temp = new CPoint();
  double del=this->coefficient[0]*eq->coefficient[1]-eq->coefficient[0]*this->coefficient[1];
//   double t1,t2;
//   double a;
//   double b;
  if(del==0){
    temp->coord[0]=-1;
    temp->coord[1]=-1;
  }
  else{
// 	a = this->coefficient[2]*eq->coefficient[1] - eq->coefficient[2]*this->coefficient[1];
// 	b = this->coefficient[0]*eq->coefficient[2]- this->coefficient[2]*eq->coefficient[0];
// 	t1 = a/del;
//         t2 = b/del;
    temp->coord[0]=(this->coefficient[2]*eq->coefficient[1] - eq->coefficient[2]*this->coefficient[1])/del;

     temp->coord[1]=((this->coefficient[0]*eq->coefficient[2])-(this->coefficient[2]*eq->coefficient[0]))/del;
  }
//    cout << " returning intersection point (" << t1 << "," << t2 <<")" ;
// 	temp->coord[0] = t1;
// 	temp->coord[1] = t2;
//    temp->print();
//    cout << endl;
  return temp;
}
