/***************************************************************************
 *   Copyright (C) 2008 by Muhammad Aamir Cheema   *
 *   macheema@cse.unsw.edu.au   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#ifndef EQUATION_H
#define EQUATION_H

#include <iostream>
#include "utility.h"
#include <cfloat>
#include "cpoint.h"


/**
	@author Muhammad Aamir Cheema <macheema@cse.unsw.edu.au>
*/
class CEquation
{
public:

    double coefficient[DIMENSION+1];

public:

    CEquation ( double* coeff );

    // Get the equation of bisector between two points q and p; passes through the mid-point
    CEquation ( CPoint* q, CPoint* p );

    // get the equation of bisector between q and p; passes through passThrough
    CEquation ( CPoint* q, CPoint *p, CPoint* passThrough );

// Will create a equaltion passing through p1 and p2 (It's not a bisector)
    CEquation ( CPoint* p1, CPoint* p2, bool temp );
    CEquation(CPoint* p, float angle, bool temp);

    ~CEquation();

    // prints the equation in 4X1 + 3X2 - 3X3 >= 4 format
    void print ( void );


    double getCoeff ( int index );

    // Change the equation so that it passes through origin
    void moveToOrigin ( void );

    // Pass DIMENSION-1 attributes and solve equation to find the coeff of unknown DIMENSION
    double solveFor ( double* solveFor );

    // pass a point and check whether it lies in the subspace defined by this->equation or not
    bool liesInSpace ( double* solveFor );

//Checks where the point p is inside the space or not
    bool pliesInSpace ( CPoint *p );

    // if the coefficient for any dimension is zero, set it to some real small value
    void fixZeroes ( void );

    //Finding the intersecting points of two equation
    CPoint *IntersectionPoint ( CEquation *eq );
    
        // Get intersection of this->line with line2
   CPoint* getIntersection(CEquation* line2);
};

#endif
