/***************************************************************************
 *   Copyright (C) 2008 by Muhammad Aamir Cheema   *
 *   macheema@cse.unsw.edu.au   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "cpoint.h"

#include <limits>
#include <string>
#include <math.h>
#include <float.h>
#include "hilbert.h"

double dummy[DIMENSION];
CPoint::CPoint(double* p)
{
  
  memcpy(coord, p, sizeof(double) * (DIMENSION));
}


// USE ONLY FOR TWO DIMENSIONAL
CPoint::CPoint(double x, double y)
{
  assert(DIMENSION == 2);
  if(DIMENSION!=2)
    cerr << "This constructor should be used only for 2D point" << endl;
  memcpy(coord, dummy, sizeof(double)*2);
  coord[0] = x;
  coord[1] = y;
  
}

// use only for two dimensional
CPoint::CPoint(CPoint* p, double angle, double range)
{
  angle = angle*PI/180;
  assert(DIMENSION == 2);
  memcpy(coord, dummy, sizeof(double)*2);
  
  double x = p->coord[0] + range*cos(angle);
  double y = p->coord[1] + range*sin(angle);
  coord[0] = x;
  coord[1] = y;
  
 // assert(fabs(this->getDist(p) - range*range) < EPSILON);
}

CPoint::CPoint(void)
{
  memcpy(coord, dummy, sizeof(double)*(DIMENSION));
  for(int i=0; i<DIMENSION; i++)
    coord[i] = 0;
}

CPoint::~CPoint(void)
{
  // we do not need to do the deletion of array because array is not dynamic
//   delete[] coord;
}


void CPoint::print(void)
{
//   cout.precision(20);

  cout << "(";
  for(int i=0; i<DIMENSION-1; i++)
  {
    cout.precision(10);
    cout << coord[i] << ",";
  }
  cout.precision(10);
  cout << coord[DIMENSION-1] << ")";
}

double CPoint::getDist(CPoint* p)
{
  double dist = 0.0;
  for(int i = 0; i < DIMENSION; i++ )
  {
    dist += pow(coord[i]-p->coord[i],2);
  }
  return dist;
}



long CPoint::getOrientationCode()
{
  long code = 0;
  for(int i=0; i < DIMENSION; i++)
  {
    if(coord[i]==3) // overlaps 
    {
      code = 0;
      break;
    }
    code = code*10 + coord[i];
    
  }
  
  return code;

}

CPoint* CPoint::getCenter(CPoint* p)
{
  CPoint* cen = new CPoint();
  for(int i = 0; i < DIMENSION; i++)
  {
    cen->coord[i] = (this->coord[i] + p->coord[i])/2;
  }
  return cen;
}

int CPoint::getCellID()
{
  int cellID = 0;
  cellID = (int)(this->coord[0]*CGlobal::GRID_WIDTH);
  cellID = cellID * CGlobal::GRID_WIDTH + (int)(this->coord[1]*CGlobal::GRID_WIDTH);
  return cellID;
}

void CPoint::printCell()
{
  int i = (int)(this->coord[0]*CGlobal::GRID_WIDTH);
  int j = (int)(this->coord[1]*CGlobal::GRID_WIDTH);
  cout << "[" << i << "," << j << "] "; 
}

bool CPoint::isSame(CPoint *p)
{
  if(fabs(this->coord[0] - p->coord[0])< .0000001 && fabs(this->coord[1] - p->coord[1])< .0000001)
 //if(fabs(this->coord[0] - p->coord[0])< EPSILON && fabs(this->coord[1] - p->coord[1])< EPSILON)
    return true;
  else
    return false;
}

bool CPoint::pliesInSpace(CPoint* q, CPoint* obj)
{
   double distQ = this->getDist(q);
   double distObj = this->getDist(obj);

//    cout << "distObj: " << distObj << " distQ: " << distQ << " diff: " << fabs(distObj - distQ) << endl;
   if(distObj < distQ && (fabs(distObj - distQ) > EPSILON2))
    return true;
   else
    return false;

}

// Get the angle between the point p and point f
double CPoint::getAngle(CPoint* f)
{
  
  double y_diff = f->coord[1] - this->coord[1];
  double x_diff = f->coord[0] - this->coord[0];
  
 
//   cout << " x " << x_diff << " y " << y_diff << endl;
  
//   assert( y_diff!=0 || x_diff!=0);
  double angle;
  if(y_diff==0 && x_diff==0)
    angle =0;
  else
    angle = atan2(y_diff,x_diff);
  
  if(angle<0)
    angle += 2*PI;
  
  angle = angle * 180/PI;
  
  
//   cout << " angle is " << angle << endl;
  return angle; 


}



void rot( int n , int *x , int *y , int rx, int ry)
{
  if( ry == 0)
  {
    if( rx == 1 )
    {
      *x = n-1 - *x;
      *y = n-1 - *y;
    }
    
    int t = *x ;
    *x = *y;
    *y = t;
  }
}

// Algorithm from http://en.wikipedia.org/wiki/Hilbert_curve
// Modified by Shiyu Yang 
// 2D 
double CPoint::getHilbertValue ( int n )
{
  int x = this->coord[0]*n;
  int y = this->coord[1]*n;
  
  int rx, ry, s, d=0;
  for( s=n/2 ; s > 0 ; s/=2 )
  {
    rx = ( x & s ) > 0;
    ry = ( y & s ) > 0;
    d += s * s * ( ( 3 * rx ) ^ ry );
    rot( s , &x , &y , rx , ry );   
  }
   
   return d;
}


//Code from Yufei Tao
//

bitmask_t CPoint::cal_hilbert(int _num_bit, int _dim,  float _umax)
{
	bitmask_t factor = 1 << _num_bit;
	bitmask_t *coord_t=new bitmask_t[_dim];
	for (int i=0; i<_dim; i++)
	{
		coord_t[i]=bitmask_t( this->coord[i]/_umax*int64_t (factor));
	}
	bitmask_t hil=hilbert_c2i(_dim, _num_bit, coord_t);
	delete []coord_t;
	return hil;
}






