/***************************************************************************
 *   Copyright (C) 2008 by Muhammad Aamir Cheema   *
 *   macheema@cse.unsw.edu.au   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#ifndef CPOINT_H
#define CPOINT_H

#include <iostream>
#include "utility.h"
#include "hilbert.h"
#include <cfloat>

/**
  @author Muhammad Aamir Cheema <macheema@cse.unsw.edu.au>
*/
class CPoint{
public:
    
  double coord[DIMENSION];

  
  public:
  
  CPoint(double* p);
  CPoint(void);
  // USE ONLY FOR TWO DIMENSIONAL
  CPoint(double x, double y);
  // Two dimensional: get a point that lies on the circle centered at p with range r and has an angle "angle" from p
  CPoint(CPoint* p, double angle, double range);
  
    ~CPoint();
  void print(void);
  double getDist(CPoint* p);
  long getOrientationCode();
  
  //return center of this->point and p
  CPoint* getCenter(CPoint* p);
  
  // return the cellID in which this point lies
  int getCellID();
  
  // print the cell in which this point lies
  void printCell();

  bool isSame(CPoint *p);

  
  // get the HiblertValue between this point and p  
  // By Shiyu Yang 
  double getHilbertValue(int n);
  
  bitmask_t cal_hilbert(int _num_bit, int _dim, float _umax);

  // returns true if this->point lies closer to obj than to NN 
  bool pliesInSpace(CPoint* q, CPoint* obj);
  
  // get the angle between this->point and f
  double getAngle(CPoint* f);
  
  
 
  


};

#endif
