#include "heapnode.h"

CHeapNode::CHeapNode(void)
: m_key(0.0),
  m_data(NULL),
  m_parent(NULL),
  m_leftChild(NULL),
  m_rightChild(NULL),
  m_previous(NULL),
  m_next(NULL)
{
}

CHeapNode::CHeapNode(double key, void* data)
: m_key(key),
  m_data(data),
  m_parent(NULL),
  m_leftChild(NULL),
  m_rightChild(NULL),
  m_previous(NULL),
  m_next(NULL)
{
}

CHeapNode::~CHeapNode(void)
{
}

double CHeapNode::getKey(void) const
{
	return m_key;
}

void* CHeapNode::getData(void)
{
  return m_data;
}

CHeapNode* CHeapNode::getParent(void) const
{
	return m_parent;
}

CHeapNode* CHeapNode::getLeftChild(void) const
{
	return m_leftChild;
}

CHeapNode* CHeapNode::getRightChild(void) const
{
	return m_rightChild;
}

CHeapNode* CHeapNode::getPrevious(void) const
{
	return m_previous;
}

CHeapNode* CHeapNode::getNext(void) const
{
	return m_next;
}

bool CHeapNode::isLeftChild(void) const
{
	return(this == m_parent->m_leftChild);
}

void CHeapNode::setParent(CHeapNode* pNode)
{
	m_parent = pNode;
}

void CHeapNode::setLeftChild(CHeapNode* pNode)
{
	m_leftChild = pNode;
}

void CHeapNode::setRightChild(CHeapNode* pNode)
{
	m_rightChild = pNode;
}

void CHeapNode::setPrevious(CHeapNode* pNode)
{
	m_previous = pNode;
}

void CHeapNode::setNext(CHeapNode* pNode)
{
	m_next = pNode;
}

void CHeapNode::swapData(CHeapNode* pNode)
{
    double key = m_key;
    m_key = pNode->m_key;
    pNode->m_key = key;
  
    void* data = m_data;
    m_data = pNode->m_data;
    pNode->m_data = data;
}

void CHeapNode::dump(ostream& os, long level, bool details) const
{
	CHeapNode* node = getRightChild();
	if(node != NULL)
	{
		node->dump(os, level + 1, details);
	}

	for(long i = 0; i < level; i ++)
	{
		os << "    ";
	}
	os << "|== ";

	os << m_key << " " << ((CNode*)m_data)->m_id << endl;

	node = getLeftChild();
	if(node != NULL)
	{
		node->dump(os, level + 1, details);
	}
}
