#include "indexnode.h"

#include <limits>

using namespace std;


CIndexNode::CIndexNode(long id)
    : CNode(id),
    m_entries(NULL)
{
}

CIndexNode::CIndexNode(long level, CRTree* rtree)
    : CNode(level, rtree)
{
  m_entries = new CNode*[rtree->m_nodeCapacity];
}

CIndexNode::~CIndexNode(void)
{
  CNode* tmpNode;
  for (long i = 0; i < m_size; ++ i)
  {
     delete m_entries[i];
  }

  if ( m_entries != NULL )
    delete[] m_entries;
}

CEntry* CIndexNode::getEntry(long index)
{
  return m_entries[index];
}

void CIndexNode::getBytesLength(long& nodeLen, long& entryLen)
{

  nodeLen = sizeof(long) + // m_level
            sizeof(long) + // m_size
            sizeof(long) + // aggregate number of data
            sizeof(double);   //added by Bin

  
  entryLen = sizeof(long) +           // id
             sizeof(double) * DIMENSION * 2;         // m_mbr
}


long CIndexNode::getBytesLength(void)
{
  long len = sizeof(long) +           // m_level
             sizeof(long) +         // m_size
             sizeof(long) +         // aggregate number of data
             sizeof(double);

  // sizeof(double) * DIMENSION * 2 +  // m_mbr
  // id of child nodes only
  len += m_size*(sizeof(long)+sizeof(double)*DIMENSION * 2); 

  return len;
}

void CIndexNode::toBytes(long& len, char** data)
{
  len = getBytesLength();
  *data = new char[len];
  char* ptr = *data;

  // m_level
  memcpy(ptr, &m_level, sizeof(long));
  ptr += sizeof(long);

  // m_size
  memcpy(ptr, &m_size, sizeof(long));
  ptr += sizeof(long);

  // aggreate number of data
  memcpy(ptr, &m_dataCnt, sizeof(long));
  ptr += sizeof(long);
  
  //added by Bin
  memcpy(ptr, &m_p, sizeof(double));
  ptr += sizeof(double);
  
  /*
  memcpy(ptr, &m_exp, sizeof(double)*DIMENSION);
  ptr += sizeof(double)*DIMENSION;
  memcpy(ptr, &m_exp_key, sizeof(double));
  ptr += sizeof(double);
  */

  for (long i = 0; i < m_size; ++ i)
  {
    memcpy(ptr, &(m_entries[i]->m_id), sizeof(long));
    ptr += sizeof(long);
    // write the MBR of the child node
    CRectangle* r = m_entries[i]->m_mbr;
    memcpy(ptr, r->getLow(), sizeof(double) * DIMENSION);
    ptr += sizeof(double) * DIMENSION;
    memcpy(ptr, r->getHigh(), sizeof(double) * DIMENSION);
    ptr += sizeof(double) * DIMENSION;
  }
}

void CIndexNode::load(void)
{
  CGlobal::global_IO++;
  if ( m_size > 0 )
    return;
CGlobal::global_IO_Buff++;

  char* buf;
  m_rtree->m_file->readEntry(m_id, &buf);
  fromBytes(buf);
  delete []buf;
}



void CIndexNode::fromBytes(char* data)
{
  // m_level
  memcpy(&m_level, data, sizeof(long));
  data += sizeof(long);

  // m_size
  memcpy(&m_size, data, sizeof(long));
  data += sizeof(long);

  double low[DIMENSION];
  double high[DIMENSION];

  // mbr NO self MBR now

  // aggregate number of data
  memcpy(&m_dataCnt, data, sizeof(long));
  data += sizeof(long);
  
  //added by Bin
  memcpy(&m_p, data, sizeof(double));
  data += sizeof(double);
  
/*  
  memcpy(&m_exp, data, sizeof(double)*DIMENSION);
  data += sizeof(double)*DIMENSION;
  memcpy(&m_exp_key, data, sizeof(double));
  data += sizeof(double);
*/
  
  if ( m_level==1 )
  {
    if (m_entries == NULL)
    {
      m_entries = (CNode**)(new CLeafNode*[m_size]);
    }
  }
  else
  {
    if (m_entries == NULL)
    {
      m_entries = (CNode**)(new CIndexNode*[m_size]);
    }
  }

  long id =0;
  for (long i = 0; i < m_size; ++ i)
  {
    memcpy(&id, data, sizeof(long));
    data += sizeof(long);

    if (m_level == 1) // child entry is leaf node
    {

      m_entries[i] = new CLeafNode(id);
      // get mbr
      memcpy(low, data, sizeof(double) * DIMENSION);
      data += sizeof(double) * DIMENSION;
      memcpy(high, data, sizeof(double) * DIMENSION);
      data += sizeof(double) * DIMENSION;

      if ( m_entries[i]->m_mbr == NULL)
        m_entries[i]->m_mbr = new CRectangle(low,high );

      m_entries[i]->m_rtree = this->m_rtree;

    }
    else  // child entry is index node
    {
      m_entries[i] = new CIndexNode(id);
      // get mbr
      memcpy(low, data, sizeof(double) * DIMENSION);
      data += sizeof(double) * DIMENSION;
      memcpy(high, data, sizeof(double) * DIMENSION);
      data += sizeof(double) * DIMENSION;
      if ( m_entries[i]->m_mbr == NULL)
        m_entries[i]->m_mbr = new CRectangle(low,high );

      m_entries[i]->m_rtree = this->m_rtree;
    }
  }

}

CNode* CIndexNode::chooseSubtree(CRectangle* mbr, long level,
                                 CStack<CNode*>& path)
{
  long index = 0;

  CIndexNode* node = this;

  while (node->m_level != level &&
          node->m_level > 1)  // child entry is index node
  {
    path.push(node);

    node = node->findLeastEnlargement(mbr);
  }

  if (node->m_level == level)
  {
    return node;
  }

  if (node->m_level == 1) // child entry is leaf node
  {
    path.push(node);

    return node->findLeastOverlap(mbr);
  }
}

CIndexNode* CIndexNode::findLeastEnlargement(CRectangle* mbr)
{
  double area = 0.0;
  double enlargement = 0.0;

  double leastEnlargement = numeric_limits<double>::max();
  long best = -1;

  CRectangle* combinedMBR = new CRectangle();

  for (long i = 0; i < m_size; ++ i)
  {
    combinedMBR->combine(m_entries[i]->m_mbr, mbr);

    area = m_entries[i]->m_mbr->getArea();
    enlargement = combinedMBR->getArea() - area;

    if (enlargement < leastEnlargement)
    {
      leastEnlargement = enlargement;
      best = i;
    }
    else if (enlargement == leastEnlargement)
    {
      if (area < m_entries[best]->m_mbr->getArea())
      {
        best = i;
      }
    }
  }

  //yingz++
  delete combinedMBR;

  return (CIndexNode*)m_entries[best];
}

CLeafNode* CIndexNode::findLeastOverlap(CRectangle* mbr)
{
  double overlap = 0.0;
  double leastOverlap = numeric_limits<double>::max();
  double leastEnlargement = numeric_limits<double>::max();

  COverlapEntry* best;

  COverlapEntry** entries = new COverlapEntry*[m_size];


  for (long i = 0; i < m_size; ++ i)
  {
    entries[i] = new COverlapEntry();

    entries[i]->m_index = i;
    entries[i]->m_oriMBR = m_entries[i]->m_mbr;
    entries[i]->m_combMBR->combine(m_entries[i]->m_mbr, mbr);
    entries[i]->m_oriArea = entries[i]->m_oriMBR->getArea();
    entries[i]->m_combArea = entries[i]->m_combMBR->getArea();

    entries[i]->m_enlargement = entries[i]->m_combArea -
                                entries[i]->m_oriArea;

    if (entries[i]->m_enlargement < leastEnlargement)
    {
      leastEnlargement = entries[i]->m_enlargement;
      best = entries[i];
    }
    else if (entries[i]->m_enlargement == leastEnlargement)
    {
      if (entries[i]->m_oriArea < best->m_oriArea)
      {
        best = entries[i];
      }
    }
  }

  // yingz ++

  if (leastEnlargement > 0.0)
  {
    long cnt = 0;

    if (m_size > m_rtree->m_nearMinimumOverlapFactor)
    {
      qsort(entries, m_size, sizeof(COverlapEntry*),
            COverlapEntry::compare);

      cnt = (long)m_rtree->m_nearMinimumOverlapFactor;
    }
    else
    {
      cnt = m_size;
    }

    double f = 0.0;
    double diff = 0.0;
    COverlapEntry* e;

    for (long i = 0; i < cnt; ++ i)
    {
      e = entries[i];
      diff = 0.0;

      for (long j = 0; j < m_size; ++ j)
      {
        if (e->m_index != j)
        {
          f = e->m_combMBR->getIntersectArea(m_entries[j]->m_mbr);

          if (f != 0.0)
          {
            diff += f - e->m_oriMBR->getIntersectArea(
                        m_entries[j]->m_mbr);
          }
        }
      }

      if (diff < leastOverlap)
      {
        leastOverlap = diff;
        best = e;
      }
      else if (diff == leastOverlap)
      {
        if (e->m_enlargement == best->m_enlargement)
        {
          if (e->m_oriArea < best->m_oriArea)
          {
            best = e;
          }
        }
        else if (e->m_enlargement < best->m_enlargement)
        {
          best = e;
        }
      }
    }
  }


  long rtn = best->m_index;

  //yingz++
  for (long i= 0; i < m_size; ++ i)
  {
    delete entries[i];
  }
  delete []entries;


  return (CLeafNode*)m_entries[rtn];
}
