#ifndef INDEXNODE_H
#define INDEXNODE_H

#include "mystack.h"

#include "rtree.h"
#include "node.h"
#include "leafnode.h"
#include "rectangle.h"

using namespace std;


class CRTree;
class CNode;
class CLeafNode;

class CIndexNode : public CNode
{
private:
  // m_entries: the child entries (i.e., index or leaf node), the capacity
  // is defined by by rtree's leaf capacity
  CNode** m_entries;

  class COverlapEntry
  {
  public:
    long m_index;     // the index of child entry in this node
    double m_enlargement;
    CRectangle* m_oriMBR;
    CRectangle* m_combMBR;
    double m_oriArea;
    double m_combArea;

    COverlapEntry(void)
    {
      m_combMBR = new CRectangle();
    };

    ~COverlapEntry(void)
    {
      delete m_combMBR;
    };

    static int compare(const void* v1, const void* v2)
    {
      COverlapEntry* p1 = *(COverlapEntry**)v1;
      COverlapEntry* p2 = *(COverlapEntry**)v2;

      if (p1->m_enlargement < p2->m_enlargement)
      {
        return -1;
      }
      else if (p1->m_enlargement > p2->m_enlargement)
      {
        return 1;
      }
      else
      {
        return 0;
      }
    }
  };

public:
  CIndexNode(long id);
  CIndexNode(long level, CRTree* rtree);

  virtual ~CIndexNode(void);

  // getEntry: get the child entry specified by 'index'
  CEntry* getEntry(long index);

  static void getBytesLength(long& nodeLen, long& entryLen);

  long getBytesLength(void);
  void toBytes(long& len, char** data);
  void fromBytes(char* data);

  void load(void);

private:
  // chooseSubtree: find the proper leaf node for insertion
  CNode* chooseSubtree(CRectangle* mbr, long level,
                       CStack<CNode*>& path);

  // findLeastEnlargement: for choose subtree when child node is index node
  CIndexNode* findLeastEnlargement(CRectangle* mbr);

  // findLeastOverlap: for choose subtree when child node is leaf node
  CLeafNode* findLeastOverlap(CRectangle* mbr);

  friend class CRTree;
  friend class CNode;
  friend class CLeafNode;
};

#endif
