#include "leafnode.h"


CLeafNode::CLeafNode(long id)
    : CNode(id),
    m_entries(NULL)
{

}

CLeafNode::CLeafNode(CRTree* rtree)
    : CNode(0, rtree)
{
  m_entries = new CEntry*[rtree->m_nodeCapacity];
}

CLeafNode::~CLeafNode(void)
{
  for (long i = 0; i < m_size; ++ i)
  {
    delete m_entries[i];
  }

  if ( m_entries != NULL )
    delete[] m_entries;
}

void CLeafNode::load(void)
{
  CGlobal::global_IO++;
  if ( m_size > 0 )
    return;
CGlobal::global_IO_Buff++;
  char* buf;
  m_rtree->m_file->readEntry(m_id, &buf);

  fromBytes(buf);
  delete[] buf;
}


CEntry* CLeafNode::getEntry(long index)
{
  return m_entries[index];
}

void CLeafNode::getBytesLength(long& nodeLen, long& entryLen)
{
  nodeLen = sizeof(long) +          // m_level
      sizeof(long) +           // m_size
      sizeof(double);   //added by Bin

  // yingz++
  entryLen = sizeof(long) +   // m_id
             sizeof(double) +  
             sizeof(double) * DIMENSION * 2;// m_mbr
}

long CLeafNode::getBytesLength(void)
{
  long len = sizeof(long) +           // m_level
            sizeof(long) +           // m_size
            sizeof(double);
  // sizeof(double) * DIMENSION * 2 +  // m_mbr No self MBR now

  // yingz++
  len += m_size * (sizeof(long) +   // m_id
                  sizeof(double) +  //added by Bin
                  sizeof(double)*DIMENSION*2) ; // m_mbr->m_low

  return len;
}

void CLeafNode::toBytes(long& len, char** data)
{
  len = getBytesLength();

  *data = new char[len];
  char* ptr = *data;

  // m_level
  memcpy(ptr, &m_level, sizeof(long));
  ptr += sizeof(long);

  // m_size
  memcpy(ptr, &m_size, sizeof(long));
  ptr += sizeof(long);
  
  // added by Bin
  memcpy(ptr, &m_p, sizeof(double));
  ptr += sizeof(double);
/*  
  memcpy(ptr, &m_exp, sizeof(double)*DIMENSION);
  ptr += sizeof(double)*DIMENSION;
  memcpy(ptr, &m_exp_key, sizeof(double));
  ptr += sizeof(double);
*/

  for (long i = 0; i < m_size; ++ i)
  {
    // id
    memcpy(ptr, &(m_entries[i]->m_id), sizeof(long));
    ptr += sizeof(long);
    
    //added by Bin
    memcpy(ptr, &(m_entries[i]->m_p), sizeof(double));
    ptr += sizeof(double);
    
    /*
    memcpy(ptr, &(m_entries[i]->m_exp), sizeof(double)*DIMENSION);
    ptr += sizeof(double)*DIMENSION;    
    memcpy(ptr, &(m_entries[i]->m_exp_key), sizeof(double));
    ptr += sizeof(double);
    */
    // mbr
    memcpy(ptr, m_entries[i]->m_mbr->getLow(),
           sizeof(double) * DIMENSION);
    ptr += sizeof(double) * DIMENSION;
    memcpy(ptr, m_entries[i]->m_mbr->getHigh(),
           sizeof(double) * DIMENSION);
    ptr += sizeof(double) * DIMENSION;

  }
}

void CLeafNode::fromBytes(char* data)
{
  // m_level
  memcpy(&m_level, data, sizeof(long));
  data += sizeof(long);

  // m_size
  memcpy(&m_size, data, sizeof(long));
  data += sizeof(long);
 
  m_dataCnt = m_size;
  
  //added by Bin
  memcpy(&m_p, data, sizeof(double));
  data += sizeof(double);

  double low[DIMENSION];
  double high[DIMENSION];

  if (m_entries == NULL)
  {
    m_entries = new CEntry*[m_size];
  }
  
  
  long id = 0;
  double prob;
  double exp[DIMENSION];
  double exp_key;

  for (long i = 0; i < m_size; ++ i)
  {
    // id
    memcpy(&id, data, sizeof(long));
    data += sizeof(long);

    //added by Bin
    memcpy(&prob, data, sizeof(double));
    data += sizeof(double);
    
    // mbr
    memcpy(low, data, sizeof(double) * DIMENSION);
    data += sizeof(double) * DIMENSION;
    memcpy(high, data, sizeof(double) * DIMENSION);
    data += sizeof(double) * DIMENSION;

    m_entries[i] = new CEntry(id, new CRectangle(low, high), 1, prob);
    
  }
}

CNode* CLeafNode::chooseSubtree(CRectangle* mbr, long level,
                                CStack<CNode*>& path)
{
  return this;
}
