#ifndef LEAFNODE_H
#define LEAFNODE_H

#include "rtree.h"
#include "node.h"
#include "indexnode.h"
#include "rectangle.h"


class CRTree;
class CNode;
class CIndexNode;

class CLeafNode : public CNode
{
private:
  // m_entries: all data points in this leaf node, the capacity is defined
  // by rtree's leaf capacity
  CEntry** m_entries;

public:
  CLeafNode(long id);
  CLeafNode(CRTree* rtree);

  virtual ~CLeafNode(void);

  // getEntry: get the child entry specified by 'index'
  CEntry* getEntry(long index);

  // getBytesLength: get the size of this node object in terms of bytes
  static void getBytesLength(long& nodeLen, long& entryLen);
  long getBytesLength(void);
  void toBytes(long& len, char** data);
  void fromBytes(char* data);
  void load(void);

private:
  // chooseSubtree: find the proper leaf node for insertion
  CNode* chooseSubtree(CRectangle* mbr, long level,
                       CStack<CNode*>& path);

  friend class CRTree;
  friend class CNode;
  friend class CIndexNode;
};

#endif
