/***************************************************************************
 *   Copyright (C) 2013 by Shiyu Yang and Muhammad Aamir Cheema            *
 *   Influence Zone was implemented by Muhammad Aamir Cheema               *  
 *   Six-regions and Slice were implemented by Shiyu Yang                  *
 *   If you find any bugs, please let us know at macheema@cse.unsw.edu.au  *
 *   or yangs@cse.unsw.edu.au                                              * 
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "rknn.h"
#include <iostream>
#include <cstdlib>
#include <sstream>



/*   INPUT PARAMETER     */

char treeName[256];
char UtreeName[256];
char* fDist;
int BUFFERSIZE;
int max_Facilities = 0;
int numQ = 0;
int k;
int max_cars;
int algoNum;
int pNum;
int HilbertOrder;
/* ---------------------- */



/*  GLOBAL VARIABLES  */

CRTree* Ftree;
CRTree* Utree;
string ofname;
ofstream output;

int GLOBAL_DIMENSION=0;
int CGlobal::GRID_WIDTH;
long CGlobal::global_IO = 0;
long CGlobal::global_IO_Buff = 0;
Object** objTable;
CPoint** facilityTable;
Query** queryTable;
CPoint* GLOBAL_POINT = new CPoint();
double angle_range;
set<int> bufferList;
set<CPoint*> RkNN_Points;
int gQID;

/* ---------------------- */




/*  FLAGS  */

bool INITIALIZING = true;
bool INFZONE_ALGO = true;
bool debug = false;
bool FLAG=false;
bool isSLICE = false ;
bool isMono = false;
bool isInitSLICE = false;
bool LOWER=false;
bool DISPLAY = false;
bool cmpIOBound = false;
bool TPL_MONO_POINT_PRUNED = false;
/* -------------------------- */



/*     MEASURING VARIABLES */

int PRUNE_PHASE_IO=0;
int REFINE_PHASE_IO=0;
int VERIFY_PHASE_IO =0;
int VERIFY_BUFFERED_IO=0;
int USER_TREE_IO = 0;
int FACILITY_TREE_IO = 0;
int LOWER_USER_IO = 0 ;
int LOWER_FACILITY_IO = 0 ;

CUtility* p1_util = new CUtility();
CUtility* p2_util = new CUtility();
CUtility* verify_util = new CUtility();
CUtility* tot_util = new CUtility();




int TPL_OLD_NEW_POINT = 0;  // Point pruned by OLD TPL cannot be pruned by OPT TPL
int TPL_NEW_OLD_POINT = 0;  // Point pruned by OPT TPL cannot be pruned by OLD TPL
int TPL_OLD_NEW_NODE = 0;  // Node pruned by OLD TPL cannot be pruned by OPT TPL
int TPL_NEW_OLD_NODE = 0;  // Node pruned by OPT TPL cannot be pruned by OLD TPL






double p1_time = 0;
double p2_time = 0;
double verify_time = 0;
double tot_time = 0;
double UCandidates=0;

double num_of_points_minList = 0;
double EntryPrune_Point_Facility = 0;
double EntryPrune_Node_Facility = 0;
double EntryPrune_Point_User = 0;
double EntryPrune_Node_User = 0;


double num_infObj = 0;
bool BRUTEFORCE = false;


/*     ------------------- */



/*   ALGORITHM NAME   */

const int SIXMONO = 1;
const int FINCHMONO_OPT = 2;
const int TPLMONO =3;
const int InfZoneMONO = 4;
const int SLICEMONO = 5;
const int SIXBRNN = 6;
const int FINCHBRNN_OPT = 7;
const int TPLBRNN = 8 ;   //Version 1
const int InfZoneBRNN = 9;
const int SLICEBRNN = 10;

const int TPLMONO_OPT = 12; 

const int TPLBRNN_OPT = 14; //Version 4
const int TPLBRNN_OPT1 = 15; //Version 2
const int TPLBRNN_OPT2 = 16; //Version 3
const int Theoretical = 20;



/* -----------------------* /



//==========================================//
//************************************************/



using namespace std;

int main(int argc, char *argv[])
{
  
   if(argc != 12)
   {
     cerr << "USAGE: requires 10 parameters\n" <<
          " 1.  #facilities \n"  <<
          " 2.  facility distribution \n" << 
          " 3.  Facility RTree Node Size (in K)\n" <<
          " 4.  #queries \n " << 
          " 5.  k \n"
          " 6.  #users \n"
          " 7.  users distribution\n"
          " 8.  User RTree Node Size (in K) \n"
          " 9.  Buffersize (use 1 for no buffer)\n"
          "10.  Algo: SIXMONO = 1; FINCHMONO =2; TPLMONO =3; InfZoneMONO = 4; SLICEMONO = 5; SIXBRNN = 6; FINCHBRNN = 7; TPLBRNN = 8;  InfZoneBRNN = 9; SLICEBRNN = 10;\n"
 	 "11. Number of Partition"<< endl;
           exit(0);      
   } 
  
  
   fDist = argv[2];
      
   max_Facilities = atoi(argv[1]);
   numQ = atoi(argv[4]);
   k = atoi(argv[5]);
   queryTable = new Query*[numQ+1];   
   pNum= atoi(argv[11]);   
   facilityTable = new CPoint*[max_Facilities+1];

   sprintf(treeName,"%s%d%s%d%s%d%s%s", "../RTrees/", atoi(argv[3]), "K_", atoi(argv[1]), "_", DIMENSION, "d_" , argv[2]);  
   cerr << treeName << endl;
   
   Ftree = new CRTree( treeName );
   Ftree->m_file->open(false);
   
   Ftree->loadTree(); 
   
  
   
   BUFFERSIZE = atoi(argv[9]);
   algoNum = atoi(argv[10]);
   
   loadQueries();
   
   max_cars = atoi(argv[6]);

   sprintf(UtreeName,"%s%d%s%d%s%d%s%s", "../RTrees/",atoi(argv[3]), "K_", atoi(argv[6]), "_" ,DIMENSION, "d_", argv[7]);
 
   cerr << UtreeName << endl << flush;
   

   if(algoNum == SIXMONO)
   {
     ofname = "../results/SIX_MONO_"; 
   }
   if(algoNum == FINCHMONO_OPT)
   {
     ofname = "../results/FINCH_MONO_";     
   }
   if(algoNum == TPLMONO)
   {  
     ofname = "../results/TPL_MONO_";           
   }
   if(algoNum == TPLMONO_OPT)
   {
     ofname = "../results/TPL_OPT_MONO_";              
   }
   if(algoNum == InfZoneMONO)
   {
     ofname = "../results/InfZone_MONO_"; 
   }
   if(algoNum == SLICEMONO)
   {
     ofname = "../results/SLICE_MONO_"; 
   }
   if(algoNum == SIXBRNN)
   {
     ofname = "../results/SIX_BRNN_"; 
   }
   if(algoNum == FINCHBRNN_OPT)
   {
     ofname = "../results/FINCH_BRNN_"; 
   }
   if(algoNum == TPLBRNN)
   {
#ifdef TPLEnforce     
     ofname = "../results/TPL_BRNN_"; 
#else 
#ifdef cmpTPLPruning
     ofname = "../results/TPL_BRNN_CMP_";
#else
     ofname ="../results/TPL_BRNN_OLD_";
#endif
#endif
     
   }
   if(algoNum == TPLBRNN_OPT)
   {
     ofname = "../results/TPL_OPT_BRNN_";   //Version 4
   }
   
   if(algoNum == TPLBRNN_OPT1)
   {
     ofname = "../results/TPL_OPT1_BRNN_";   //Version 2
   }
   
   if(algoNum == TPLBRNN_OPT2)
   {
     ofname = "../results/TPL_OPT2_BRNN_";  //Version3 
   }
   
   
   if(algoNum == InfZoneBRNN)
   {
     ofname = "../results/InfZone_BRNN_"; 
   } 
   if(algoNum == SLICEBRNN)
   {
     ofname = "../results/SLICE_BRNN_"; 
   }
   if(algoNum == Theoretical )
   {
     ofname = "../results/Theoretical_";
   }
     
   std::ostringstream oss;
   oss << DIMENSION << "d.txt";
  
   ofname = ofname + string(argv[1]) + "f_" + string(argv[2]) + "_" + string(argv[3]) + "K_"+ string(argv[11]) + "Partition" +"_FTree_" + string(argv[4]) + "queries_" + string(argv[5]) + "k_" + string(argv[6]) + "u_" + string(argv[7]) + "_" + string(argv[8]) + "K_UTree_" + string(argv[9]) + "buffer_" ;
   ofname +=  oss.str();
   cerr << ofname << endl;
   
#ifdef NDEBUG
   cout << " ASSERT is disabled" << endl;
#else
   cout << " Assert is enabled" << endl;
#endif

   srand(time(NULL));
   Utree = new CRTree( UtreeName );
   Utree->m_file->open(false);
   Utree->loadTree();
   
   tot_util->startCT();
   
   
   for(int xyz=1; xyz<=numQ; xyz++)
   {
      bufferList.clear();  
      gQID = xyz;
     
      
      if(algoNum == SIXMONO)
      {
        SIX_MONO(xyz);
      }
      
      if( algoNum == FINCHMONO_OPT)
      {
        FINCH_MONO(xyz);
      }
      
      if(algoNum == TPLMONO)
      {
        TPL_MONO(xyz);
      }
      
      if(algoNum == TPLMONO_OPT)
      {
	TPL_MONO_OPT(xyz);
      }

      if(algoNum == InfZoneMONO)
      {
         INFZONE_MONO(xyz);      
      }
      
      if(algoNum == SLICEMONO)
      {
        SLICE_MONO(xyz , pNum);       
      }
      
      if(algoNum == SIXBRNN)
      {
        SIX_BRNN(xyz);      
      }
      
      if(algoNum == FINCHBRNN_OPT )
      {
        FINCH_BRNN(xyz);      
      }
      
      if(algoNum == TPLBRNN || algoNum == TPLBRNN_OPT1 || algoNum == TPLBRNN_OPT2)
      {
        TPL_BRNN(xyz);      
      }
      
      if(algoNum == TPLBRNN_OPT)
      {
	TPL_BRNN_OPT(xyz);
      }
      
      if(algoNum == InfZoneBRNN)
      {
        INFZONE_BRNN(xyz);   
      }
      
      if(algoNum == SLICEBRNN)
      {
        SLICE_BRNN(xyz, pNum);   
      }   
      
      
      if(algoNum == Theoretical)
      {
	RkNN_Points.clear();
	algoNum = InfZoneBRNN;
	cmpIOBound = true;
	INFZONE_BRNN(xyz);
	algoNum = Theoretical;	
	ComputeIOBound(xyz);
      }
   }
   tot_time += tot_util->endCT(false);

   output.open(ofname.c_str());
   output.precision(15);
   printStats(output);
   printStats(cout);
   
   delete Ftree;
   delete Utree;
}


void resetQuery(int qID)
{
 
 
  list<infObj*>::iterator iList;
  for(iList = queryTable[qID]->infList.begin() ; iList!= queryTable[qID]->infList.end() ; iList++)
  {
     deleteInfObj((*iList));  
  }
  
  queryTable[qID]->infList.clear();
  
  if(algoNum == InfZoneBRNN || algoNum == InfZoneMONO   )
  {  
    

      map<int, CPoint*>::iterator imap;
      for( imap = queryTable[qID]->candidates.begin() ; imap != queryTable[qID]->candidates.end() ; imap++ )
      {
	  delete( (*imap).second );
      }
      queryTable[qID]->candidates.clear();
    
    list<Vertex*>::iterator iList2;
    for(iList2 = queryTable[qID]->vList.begin() ; iList2 != queryTable[qID]->vList.end() ; iList2++)
    {
        deleteVertex((*iList2));
    }
    list< CPoint* >::iterator iList5;
    for( iList5 = queryTable[qID]->convexChain.begin() ; iList5 != queryTable[qID]->convexChain.end() ; iList5++ )
    {
        delete( (*iList5));
    }

    queryTable[qID]->vList.clear();
    queryTable[qID]->vList_Special.clear();
    queryTable[qID]->vPolygon.clear();   
    queryTable[qID]->convexChain.clear();   
  }
  
  
  
  
  if(algoNum == FINCHBRNN_OPT  || algoNum == FINCHMONO_OPT )
  {
      map<int, CPoint*>::iterator imap;
      for( imap = queryTable[qID]->candidates.begin() ; imap != queryTable[qID]->candidates.end() ; imap++ )
      {
	  delete( (*imap).second );
      }
      queryTable[qID]->candidates.clear();
      list< CPoint* >::iterator iList5;
      for( iList5 = queryTable[qID]->convexChain.begin() ; iList5 != queryTable[qID]->convexChain.end() ; iList5++ )
      {
	  delete( (*iList5));
      }
      
      vector< Vertex* >::iterator iList7;
      for( iList7 = queryTable[qID]->vPolygon.begin() ; iList7 != queryTable[qID]->vPolygon.end() ; iList7++ )
      {
	  delete( (*iList7));
      }
      
      queryTable[qID]->vList_Special.clear();
      queryTable[qID]->vPolygon.clear();   
      queryTable[qID]->convexChain.clear();   
      queryTable[qID]->Prfn.clear();
    
  }
  
  
    
  if( algoNum == TPLBRNN || algoNum == TPLBRNN_OPT || algoNum == TPLMONO )
  {
        list<TPLCand*>::iterator iList8;
	for( iList8 = queryTable[qID]->candList.begin() ; iList8 != queryTable[qID]->candList.end() ; iList8++)
	{
	  deleteTPLCands( (*iList8));
	}	
	queryTable[qID]->candList.clear();	
  }

  if(algoNum == SLICEBRNN || algoNum == SLICEMONO || algoNum == SIXBRNN || algoNum == SIXMONO )
  {
      
      if( algoNum != SLICEMONO )
      {
	map<int, CPoint*>::iterator imap;
	for( imap = queryTable[qID]->candidates.begin() ; imap != queryTable[qID]->candidates.end() ; imap++ )
	{
	    delete( (*imap).second );
	}
      }
     queryTable[qID]->candidates.clear();
    
     for( int i = 0 ; i<pNum ; i++)
     {
       list<minObject*>::iterator iListMin;
       for( iListMin = queryTable[qID]->partitions[i].minList.begin() ; iListMin != queryTable[qID]->partitions[i].minList.end() ; iListMin++)
       {
	   deleteminObj( (*iListMin) );
       }
	
	queryTable[qID]->partitions[i].minList.clear();
	queryTable[qID]->partitions[i].upperArc.clear();
	queryTable[qID]->partitions[i].lowerArc.clear();
     }
     delete [] queryTable[qID]->partitions;
  }
      
    
}


void deleteTPLCands( TPLCand* x)
{
  delete x->p;
  list<CNode*>::iterator iList;
  if(algoNum != TPLBRNN_OPT &&  algoNum != FINCHBRNN_OPT && algoNum != TPLBRNN)
  {
    for( iList = x->toVisit.begin() ; iList != x->toVisit.end() ; iList++)
    {
      delete ((*iList));
    }
    x->toVisit.clear();  
  }
  delete x;
}


void deleteminObj( minObject* x)
{
   delete x->value ;
   delete x ;
}

void deleteInfObj(infObj* x)
{  
    delete x->loc;
    delete  x->eq;
    
    if(algoNum == InfZoneBRNN || algoNum == InfZoneMONO ||  algoNum == FINCHBRNN_OPT || algoNum == FINCHMONO_OPT)
    delete x->cent;
    
    delete x;
}


bool sortOnPolarAngle (CPoint* a, CPoint* b)
{
    double polarA = GLOBAL_POINT->getAngle(a);
    double polarB = GLOBAL_POINT->getAngle(b);
    
    double distA = GLOBAL_POINT->getDist(a);
    double distB = GLOBAL_POINT->getDist(b);
    cout.precision(20);
   if(polarA - polarB < -0.00000001)
    {
      return true;
    }
    else
    {
	if(fabs(polarA - polarB) < 0.00000001)
	{
	    if(distA < distB)
	    {
	      return true;
	    }
	}
    }
    return false;
  
}


double getExpectedSize()
{
    double t = 2*k*max_cars;
    double angle = 2*PI/pNum;
    double c = cos((angle+PI/4)/2);
    double d = max_Facilities*c*c;
    
    return t/d;
}




double getTheoreticalIOs(int TYPE)
{
    double N,f,r;
    if(TYPE == 1)  
    {
	N = max_Facilities;
	f = (double) Ftree->getFanout();
      if(algoNum==6)
      {
	double angle = 2*PI/pNum;
	double alpha = (angle + PI/4)/2;
	double denom;
	if(LOWER==false)
	  denom = 2*cos(alpha);
	else
	  denom =  2*cos(angle/2);
	
	r = sqrt(2*k/(N*(2*alpha-angle)))/denom;
	r = 2*r;
      }
      else
      {     
	r = 2*sqrt(k/(PI*N));
      }
    }
    else
    {   
      N = max_cars;
      f = (double) Utree->getFanout();
      if(algoNum==6)
      {
	double angle = 2*PI/pNum;
	double alpha = (angle + PI/4)/2;      
	double denom;
	if(LOWER==false)
	  denom = 2*cos(alpha);
	else
	  denom =  2*cos(angle/2);
	
	r = sqrt(2*k/(N*(2*alpha-angle)))/denom;     
      }
      else
      {         
	r = sqrt(k/(PI*N));
      }
    }

    cout.precision(10);
    
    bool STOP = false;
    int level=1;
    
    double IOs = 1; // root will be counted
    while(N>f)
    {
      double R = (double)N/f; 
      double half_diagonal = sqrt(1/(2*R));
      double numOfIntersectingRecs = PI*(r + half_diagonal)*(r + half_diagonal)*R; 
      IOs += numOfIntersectingRecs;  
      N = N/f;  
    }
    
    return IOs;
}



double getTheoreticalRkNN()
{
      double num;
      if(algoNum == 1 || algoNum == 6)
      {
	// MONO
	num= k;
      }
      if(algoNum == 3 || algoNum == 7)
      {
	// BICHROMATIC
	// N facilities and M be users
	// BRkNN = k*M/N
	double m = k*max_cars;
	
	num = m/max_Facilities;	
      }
      return num;
}




void printStats(ostream& os)
{
    os.precision(3);
    os << fixed ;
    os << "Phase 1 time (Pruning the space using FTree):  " << p1_time/1000 << endl;
    os << "Phase 2 time (Refinement using Ftree or UTree and verification of candidates):  " << p2_time/1000 << endl;
    os << "Verification time (verification of candidates):  " << verify_time/1000 << endl;
    os << "Total time:  " << tot_time/1000 << endl;
    os << endl;
    os << "Prune Phase IO:  " << PRUNE_PHASE_IO << endl;
    os << "REFINE Phase IO:  " << REFINE_PHASE_IO << endl;
    os << "Verifying Phase IO: " << VERIFY_PHASE_IO << endl;
    os << "Buffered IOs in Verifying phase: " << VERIFY_BUFFERED_IO << endl;
    os << "Total IO:  " << PRUNE_PHASE_IO + REFINE_PHASE_IO + VERIFY_PHASE_IO << endl;
    os << "Total Buffered IO:  " << PRUNE_PHASE_IO + REFINE_PHASE_IO + VERIFY_BUFFERED_IO << endl;
    os << endl;
    os << "Avg number of filtering facilities:  " << num_infObj/numQ<< endl;
    os << "Avg number of RkNN:  " << getAvgRkNN() << endl;
    os << "Avg number of Candidates (for BRkNN):  " << UCandidates/numQ << endl;
    os << "User Tree IO: " << USER_TREE_IO << endl;
    os << "Facility Tree IO: " << FACILITY_TREE_IO <<endl;
    os << "User + Facility: " << USER_TREE_IO + FACILITY_TREE_IO << endl;   
    os << "Entry Prune Node Facility: " << EntryPrune_Node_Facility/numQ << endl; 
    os << "Entry Prune Point Facility: " << EntryPrune_Point_Facility/numQ << endl; 
    os << "Entry Prune Node User: " << EntryPrune_Node_User/numQ << endl; 
    os << "Entry Prune Point User: " << EntryPrune_Point_User/numQ << endl; 
    
    
#ifdef cmpTPLPruning
    if(algoNum == TPLBRNN)
    {   
      os << "TPL OLD PRUNE NEW NOT NODE: " << TPL_OLD_NEW_NODE << endl;
      os << "TPL OLD PRUNE NEW NOT POINT: " << TPL_OLD_NEW_POINT << endl;
      os << "TPL NEW PRUNE OLD NOT NODE: " << TPL_NEW_OLD_NODE << endl;
      os << "TPL NEW PRUNE OLD NOT POINT: " << TPL_NEW_OLD_POINT << endl;
    }
#endif    

}


void ComputeIOBound( int qID)
{
   REFINE_PHASE_IO = getLowerUserIO(qID);
   PRUNE_PHASE_IO =  getLowerFacilityIO(qID); 
}


int getLowerUserIO( int qID)
{  
    return LOWER_USER_IO;
}



int getLowerFacilityIO( int qID)
{
    CStack< CNode* > stk;
    long count = 0;
    Ftree->getRoot()->getRect()->blowHuge();
    stk.push(Ftree->getRoot());
    CNode* node;
    CNode* child = NULL;
    set<CPoint*>::iterator it;
    
    while ( !stk.empty() )
    {
        node = stk.top();
        stk.pop(); 
	bool overlap = false;
        
	if( node->m_id < 0)
	  continue;
	
	for( it = RkNN_Points.begin() ; it != RkNN_Points.end() ; it++ )
	{
	 
	    double dist = node->getRect()->getMinDist( (*it)); 
	    double radius =  (*it)->getDist( queryTable[qID]->loc);
	    if( dist < radius )
	    {
		overlap = true;
		break;
	    }
	}
	
	if( overlap )
	{
	    LOWER_FACILITY_IO++;
	    FACILITY_TREE_IO++;
	    node->load();
	    for ( int i = 0 ; i < node->getNumChild() ; i++ )
	    {	
	      child = (CNode*) node->getEntry(i);
	      stk.push(child);
	    }
	}
      
    }  
    return LOWER_FACILITY_IO;
}







double getAvgInfObj()
{
    double tot=0;
    for(int qID=1; qID<=numQ; qID++)
    {
      tot += queryTable[qID]->infList.size();
    }
    return tot/numQ;
}

double getAvgRkNN()
{
    double tot=0;
    for(int qID=1; qID<=numQ; qID++)
    {
      tot += queryTable[qID]->RkNN.size();
    }
    return tot/numQ;
}


void SLICE_MONO( int qID, int pNum)
{ 
	
   isSLICE = true;	
   isMono = true; 
   queryTable[qID]->partitions = new Partition[pNum];	
   k = k+1;	
   
   angle_range= 360.00/pNum;	
	
   isInitSLICE = true;	
   initQuerySLICE(qID);		
   isInitSLICE = false;
		
   if( qID%100 == 0 )
     cout<<"query num"<<qID<<endl;

   p1_util->startCT();
   SLICE_filtering(qID);
   p1_time += p1_util->endCT(false);
				
   p2_util->startCT();
   SLICE_MONO_Verification(qID);
   p2_time += p2_util->endCT(false);
		
   k = k-1;
		
   isSLICE = false;
   isMono = false;
   
   resetQuery(qID);
}



void SLICE_MONO_Verification( int qID)
{
   CPoint* cp;
   map<int , CPoint* >::iterator ilist;
   ilist = queryTable[qID]->candidates.begin();
  
   for( ilist ; ilist != queryTable[qID]->candidates.end() ; ilist++ )
   {   
      cp = ilist->second;      
      if( ilist->first == qID)
      {
	continue;
      }    
      if( isRkNN( cp , qID ))
      {
	queryTable[qID]->RkNN.insert( ilist->first);
      }      
   }
   
   for( int i = 0 ; i < pNum ; i++ )
   {
     list<minObject*>::iterator ilist2;
     for(ilist2= queryTable[qID]->partitions[i].minList.begin(); ilist2!=queryTable[qID]->partitions[i].minList.end(); ilist2++)
     {
       delete (*ilist2)->value;
       delete (*ilist2);
     }
     queryTable[qID]->partitions[i].minList.clear();
   }
}


void SLICE_BRNN( int qID ,  int pNum)
{  	
   isSLICE = true;
   queryTable[qID]->partitions = new Partition[pNum];	
   angle_range= 360.00/pNum;
		
   if( qID%100 == 0 )
     cout<<"query num"<<qID<<endl;

   isInitSLICE = true;	
   initQuerySLICE(qID);		
   isInitSLICE = false;
   
	
   p1_util->startCT();
   SLICE_filtering(qID);
   p1_time += p1_util->endCT(false);
				
   p2_util->startCT();
   SLICE_BRNN_Verification(qID);
   p2_time += p2_util->endCT(false);
		
   isSLICE = false;  
   resetQuery(qID);
}

void initQuerySLICE( int qID)
{

   CPoint* ld = new CPoint(0,0);
   CPoint* lu = new CPoint(0,1);
   CPoint* rd = new CPoint(1,0);
   CPoint* ru = new CPoint(1,1);
   
   double angleLD = queryTable[qID]->loc->getAngle(ld);
   double angleLU = queryTable[qID]->loc->getAngle(lu);
   double angleRD = queryTable[qID]->loc->getAngle(rd);
   double angleRU = queryTable[qID]->loc->getAngle(ru);
   
   
   for( int i = 0 ; i < pNum ; i++)
   {
     
     if( i*angle_range < angleLD && (i+1)*angle_range>angleLD )
     {
        queryTable[qID]->partitions[i].bouderyArc = queryTable[qID]->loc->getDist(ld);
	queryTable[qID]->partitions[i].upperArc.insert(queryTable[qID]->partitions[i].bouderyArc );
     }else 
     if( i*angle_range < angleLU&& (i+1)*angle_range>angleLU )
     {
       queryTable[qID]->partitions[i].bouderyArc = queryTable[qID]->loc->getDist(lu);
       queryTable[qID]->partitions[i].upperArc.insert(queryTable[qID]->partitions[i].bouderyArc );
     }else
     if(i*angle_range < angleRD && (i+1)*angle_range>angleRD)
     {
	queryTable[qID]->partitions[i].bouderyArc = queryTable[qID]->loc->getDist(rd);
	queryTable[qID]->partitions[i].upperArc.insert(queryTable[qID]->partitions[i].bouderyArc );
     }else
     if(i*angle_range < angleRU && (i+1)*angle_range>angleRU)
     {
       queryTable[qID]->partitions[i].bouderyArc = queryTable[qID]->loc->getDist(ru);
       queryTable[qID]->partitions[i].upperArc.insert(queryTable[qID]->partitions[i].bouderyArc );
     }else    
     queryTable[qID]->partitions[i].bouderyArc = 10;
    
   }
   
   
   int x;  
  // left
   x = -1;
   CPoint* left = new CPoint(queryTable[qID]->loc->coord[0]*-1,queryTable[qID]->loc->coord[1]); 
   updateZone( left , qID ,  x);
   delete left;
	
   // up:  x is same, y is (1+(1-y))
   x = -2;
   CPoint* up = new CPoint(queryTable[qID]->loc->coord[0],2-queryTable[qID]->loc->coord[1]);
   updateZone( up , qID ,  x);
   delete up;

   // right:  y is same, x is 2-x
   x = -3;
   CPoint* right = new CPoint(2-queryTable[qID]->loc->coord[0],queryTable[qID]->loc->coord[1]);	
   updateZone( right , qID ,  x);
   delete right;
	
   // down:  x is same, y is -y
   x = -4;
   CPoint* down = new CPoint(queryTable[qID]->loc->coord[0],queryTable[qID]->loc->coord[1]*-1);
   updateZone( down , qID ,  x);
   delete down;
}

void SLICE_filtering( int qID )
{
   CHeap heap;
   Ftree->getRoot()->getRect()->blowHuge();
   heap.insertNode( 0 , Ftree->getRoot());

   CNode* curNode;
   CHeapNode* heapNode;
   int numChildren;
   CEntry* child;
   while( heap.getHeapSize()!=0 )
   {
     heapNode = heap.top();
     curNode = (CNode*) heapNode->getData();	
     heap.pop();
     
      if(curNode->m_id < 0)
      {
	EntryPrune_Point_Facility++;
      }else{
	EntryPrune_Node_Facility++;
      }
      
     
     if( !FacilityPruned( curNode->getRect() , qID , curNode->m_id ) )
     {
       if( curNode->m_id < 0 )
       {
	 CPoint* cp = new CPoint( curNode->getRect()->m_low );	    
	 if ( !cp->isSame( queryTable[qID]->loc ) )
	 {
	   updateZone(cp , qID , curNode->m_id*-1 );
	 }
         delete cp; 
       }
       else{
	 curNode->load();
	 numChildren = curNode->getNumChild();
	 PRUNE_PHASE_IO++;
	 FACILITY_TREE_IO++;
	 for( int i = 0 ; i < numChildren ; ++i )
	 {
	   child = curNode->getEntry(i);
	   heap.insertNode(child->getRect()->getMinDist(queryTable[qID]->loc), (CNode*) child);
	 }	 
       }       
     }	 
   }
	
   updateMinList( qID );    
   for(int i = 0 ; i < pNum ; i++ )
   {
     queryTable[qID]->partitions[i].k_UpperArc = (double)*queryTable[qID]->partitions[i].upperArc.rbegin(); 	  	  
     queryTable[qID]->partitions[i].upperArc.clear();	 
   }
}

bool FacilityPruned(CRectangle* rect, int qID , int node_ID)
{ 
   if(rect->contains(queryTable[qID]->loc))
     return false;
      
   double minDist = rect->getMinDist(queryTable[qID]->loc);
	    
   if(minDist> 4*queryTable[qID]->maxUpper)              // if mindist >2*Maxupper Prune safely
     return true;
    
   if(minDist< 4*queryTable[qID]->minUpper)              // if mindist <2*Minupper  can not prune
     return false;
      
   if( node_ID < 0)     
     return false;	    
	    
   double maxDist = rect->getMaxDist(queryTable[qID]->loc);
   CPoint* closestPoint = rect->getClosestPoint(queryTable[qID]->loc);    //Get the cloest point;
	    
   double closestAngle =queryTable[qID]->loc->getAngle(closestPoint);	  
   delete closestPoint; 
	    
   double minAngle ;
   double maxAngle ;
   rect->getMaxMinAngle( &minAngle , &maxAngle , queryTable[qID]->loc);
  
   if( maxAngle-minAngle > 180)  // IF the difference is greater than 180, this means the range is actually something like 320->0->40 
   {
     int rmax = (minAngle+89.99999)/angle_range;     
     for( int pNo1 = 0 ; pNo1<= rmax ; pNo1++ )
     {
       if( affectPartitions(rect , qID , pNo1 , closestAngle , 0 , minAngle ) )  // check lies in the partition
	 return false;
     }
     int rmin = (maxAngle-90)/angle_range;
     for( int pNo2 = rmin ; pNo2 < pNum ; pNo2++ )
     {			
       if( affectPartitions(rect , qID , pNo2 , closestAngle , maxAngle , 360 ) )  // check lies in the partition		
	 return false;	 	         
     }
   }
   else{	   
     float sAngle = minAngle-90;
     float eAngle = maxAngle+89.9999;		
     if (sAngle < 0 )
     {
       int pNo3 = 0;	  
       int imax = eAngle/angle_range;
       for( pNo3 ; pNo3 <= imax ; pNo3++ )
       {
	 if( affectPartitions(rect , qID , pNo3 , closestAngle , minAngle , maxAngle ) )
	   return false;	 
       }    
       float ssAngle = 360+sAngle;
       int pNo4 = ssAngle/angle_range;	    		      
       for(pNo4 ; pNo4 < pNum ; pNo4++)
       {
	 if( affectPartitions(rect , qID , pNo4 , closestAngle , minAngle, maxAngle ) )
	   return false;
       }		           
     }
     
     if(eAngle > 360)
     {
       int i = sAngle/angle_range;
       for( i ; i < pNum ; i++ )
       {
	 if( affectPartitions(rect , qID , i , closestAngle , minAngle , maxAngle ) )	
	   return false;	 
       }
       float eeAngle = eAngle-360 ;
       int  imax = eeAngle/angle_range ;
       for( int ii = 0 ; ii <= imax ; ii++ )
       { 
	 if( affectPartitions( rect , qID , ii , closestAngle , minAngle , maxAngle ) )
	   return false; 
       }       
     }
     if( sAngle >= 0 && eAngle <= 360 )
     {
       int i =  sAngle/angle_range;
       int imax = eAngle/angle_range;		       
       for( i ; i <= imax ; i++ )
       {
	 if( affectPartitions( rect , qID , i , closestAngle , minAngle , maxAngle ) )
	   return false;  
       }	
     }	   
   }
   return true;
}



bool affectPartitions(CRectangle* rect, int qID, int pNo, double cangle, double minAngle, double maxAngle)
{
   float sAngle = pNo*angle_range;
   float eAngle = sAngle+angle_range;
   bool overlap = true;
      
   double upper = (double)*queryTable[qID]->partitions[pNo].upperArc.rbegin();
   float minDist;
      
   if( upper > 10000 )
     return true;
   
   if( maxAngle < sAngle || minAngle > eAngle )
     overlap =false;
    
   if(overlap)
   {	    
     minDist=rect->getMinDist(queryTable[qID]->loc);
	    
     if( cangle < sAngle || cangle > eAngle )
     {
       float dist1 = rect->MinDistAtAngle(queryTable[qID]->loc,sAngle);
       float dist2 = rect->MinDistAtAngle(queryTable[qID]->loc,eAngle);
       if(dist1<dist2)
       {
	 minDist=dist1;
       }
       else
       {
	 minDist=dist2;
       }       
     }
     
     if(minDist<4*upper&& fabs(minDist-4*upper)>EPSILON)
       return true;    
   }
    
   CPoint* a = new CPoint(queryTable[qID]->loc, sAngle, sqrt(upper));
   CPoint* b = new CPoint(queryTable[qID]->loc, eAngle, sqrt(upper));	
   double dista = rect->getMinDist(a);
   double distb = rect->getMinDist(b);	
   delete(a);
   delete(b);
	
   if(dista<upper && fabs(dista-upper)>EPSILON)
     return true;
	  
   if(distb<upper &&fabs(distb-upper)>EPSILON)
     return true;    
      
   return false;
}


double getMinAngle( CPoint* cp , int pNon , int qID)
{
   double pAngle = queryTable[qID]->loc->getAngle(cp);   
   int parition = pAngle/angle_range;
	
   if( pNon == parition)
     return 0;
        
   double sangle = fabs( pAngle - pNon*angle_range );
   double eangle = fabs (pAngle - (pNon+1)*angle_range);    
      
   if( sangle > 180)
     sangle = 360 - sangle;
  
   if( eangle > 180)
     eangle = 360 - eangle;
   
   return min( sangle , eangle );  
}


double getMaxAnge( CPoint* cp , int pNon , int qID)
{
   double pAngle = queryTable[qID]->loc->getAngle(cp);    
   int partion = pAngle/angle_range;
   double sangle = fabs( pAngle - pNon*angle_range );
   double eangle = fabs (pAngle - (pNon+1)*angle_range );
      
   if( sangle > 180)
     sangle = 360 - sangle;
  
   if( eangle > 180)
     eangle = 360 - eangle;
       
   return max ( sangle , eangle );  
}





void updateZone( CPoint* cp , int qID , int mid)
{ 
   FLAG =false ;    
   double minAngle = 0;
   double maxAngle;
   double pangle = queryTable[qID]->loc->getAngle( cp);   
   int p = pangle / angle_range;  
   double lowerArc;
   double upperArc;    
   double tempmin;
   double tempmax;

   minAngle = getMinAngle(  cp , p  , qID );
    
   if( minAngle == 0 )
     maxAngle = getMaxAnge( cp  , p  , qID );
   else
     maxAngle = minAngle+angle_range;
      
   while( minAngle < 90 )
   {
     lowerArc = cp->getDist( queryTable[qID]->loc)/( 4 * pow( cos( minAngle*PI /180.0) , 2.0 ) );
     if( maxAngle >= 90)
       upperArc = 50000;
     else
       upperArc = cp->getDist( queryTable[qID]->loc)/( 4 * pow( cos( maxAngle*PI /180.0) , 2.0 ) );
	  
     updateArc( lowerArc , upperArc , p , qID , cp , mid);        
     p--;
	  
     if( p < 0)
       p = pNum+p;
	  
     if( minAngle == 0 )
     {
       minAngle = getMinAngle( cp , p , qID);
       maxAngle = minAngle + angle_range ;       
     }
     else{
       minAngle = maxAngle;
       maxAngle += angle_range;       
     }      
   }
  
   p =  pangle / angle_range + 1 ;
    
   if(p>=pNum)
     p=p-pNum;
    
   minAngle = getMinAngle( cp ,  p , qID ) ;
   maxAngle = minAngle + angle_range ;
  
   while( minAngle < 90 )
   {  
     lowerArc = cp->getDist( queryTable[qID]->loc)/( 4 * pow( cos( minAngle*PI /180.0) , 2.0 ) );
	  
     if( maxAngle >= 90)
       upperArc = 50000;
     else
       upperArc = cp->getDist( queryTable[qID]->loc)/( 4 * pow( cos( maxAngle*PI /180.0) , 2.0 ) );
	     
     updateArc( lowerArc , upperArc , p , qID , cp , mid);   
     p++;
	    
     if( p >= pNum)
       p = p - pNum;
	   
     minAngle = maxAngle;
     maxAngle += angle_range;
   }
      
   if( FLAG )
   { 
     double maxUpper =0;
     double minUpper =50000;
     for(int i = 0 ; i< pNum ; i++ )
     {
       if( (double)*queryTable[qID]->partitions[i].upperArc.rbegin() > maxUpper )
	 maxUpper = (double)*queryTable[qID]->partitions[i].upperArc.rbegin();
       
       if( (double)*queryTable[qID]->partitions[i].upperArc.rbegin() < minUpper )
	 minUpper = (double)*queryTable[qID]->partitions[i].upperArc.rbegin();
     }
     queryTable[qID]->maxUpper = maxUpper;
     queryTable[qID]->minUpper = minUpper;	  	  
   } 
}


void updateArc( double lowerArc , double upperArc  , int pNo , int qID , CPoint* cp , int mID )
{
  
  if( isInitSLICE)
  {
    if(upperArc <  queryTable[qID]->partitions[pNo].bouderyArc )
    {
      queryTable[qID]->partitions[pNo].bouderyArc = upperArc;
      queryTable[qID]->partitions[pNo].upperArc.clear();
      queryTable[qID]->partitions[pNo].upperArc.insert(upperArc);	
      
      return;
    }
    if( upperArc > queryTable[qID]->partitions[pNo].bouderyArc )
      return;
  }
  
  
  
  if( lowerArc > queryTable[qID]->partitions[pNo].bouderyArc  && fabs(upperArc-  queryTable[qID]->partitions[pNo].bouderyArc ) > EPSILON)
    return;
  
  
   double ruk = (double)*queryTable[qID]->partitions[pNo].upperArc.rbegin();				
   if( upperArc < ruk )
   {			  
     if( queryTable[qID]->maxUpper == 50000 || queryTable[qID]->minUpper == 0 )
       FLAG = true;    
     if( ruk == queryTable[qID]->maxUpper || ruk == queryTable[qID]->minUpper )
       FLAG = true;   
     if( queryTable[qID]->partitions[pNo].upperArc.size() == k || (double)*queryTable[qID]->partitions[pNo].upperArc.rbegin() == 50000 )
       queryTable[qID]->partitions[pNo].upperArc.erase( --queryTable[qID]->partitions[pNo].upperArc.end() );
		  
     queryTable[qID]->partitions[pNo].upperArc.insert(upperArc);			  			  
   }
		

   if(isSLICE)
   { 
     if( lowerArc <= (double)*queryTable[qID]->partitions[pNo].upperArc.rbegin() )
     {
       minObject* obj;			
       obj = new minObject;
       obj->key=lowerArc;			
       obj->value = new CPoint(cp->coord);	
       obj->objID = mID;
       
       num_infObj++;
       
       queryTable[qID]->partitions[pNo].minList.push_back(obj);
       if( isMono )
	 queryTable[qID]->candidates.insert( map<int, CPoint*>::value_type(mID , obj->value) );
       
       num_of_points_minList++;       
     }			  
   }	
}



void updateMinList(int qID)
{
   double minLower=500000;  
   list<minObject*>::iterator iList;
      
   for( int i = 0 ; i < pNum ; i++ )
   {
     queryTable[qID]->partitions[i].minList.sort(cmp);
     iList= queryTable[qID]->partitions[i].minList.begin();
     double temp=0;
     
     for(int j = 0; j < k && iList != queryTable[qID]->partitions[i].minList.end() ; iList++ , j++ )
       temp = (*iList)->key;
	      
     queryTable[qID]->partitions[i].k_LowerArc = temp;
     
     if(minLower>queryTable[qID]->partitions[i].k_LowerArc)
       minLower=queryTable[qID]->partitions[i].k_LowerArc;
    }   
    queryTable[qID]->minLower=minLower;    
}



bool cmp( const minObject* a, const minObject* b)
{
   if( a->key<b->key)
     return true;
   else
     return false;
}


void  SLICE_BRNN_Verification(int qID)
{
   CStack< CNode* > stk;
   long count = 0;
   Utree->getRoot()->getRect()->blowHuge();
   stk.push(Utree->getRoot());
   CNode* node;

   CIndexNode* index = NULL;
   CLeafNode* leaf = NULL;
   CNode* child = NULL;

   while ( !stk.empty() )
   {
     node = stk.top();
     stk.pop();   
     
      if(node->m_id < 0)
      {
	EntryPrune_Point_User++;
      }else{
	EntryPrune_Node_User++;
      }

     if( ! UserFiltering( node->getRect() , qID , node->m_id) )
     {
       if ( node->m_id < 0 )
       {    
	 UCandidates++;
	 verify_util->startCT();
	 CPoint* p = new CPoint( node->getRect()->m_low );      
	 
	 if( isRkNN( p , qID ) )
	   queryTable[qID]->RkNN.insert( node->m_id*-1 );
			  
	 delete p;
	 verify_time += verify_util->endCT(false);	 
       }	 
       else{	
	 node->load();
	 REFINE_PHASE_IO++;	
	 USER_TREE_IO++;
	 for ( int i = 0 ; i < node->getNumChild() ; i++ )
	 {			    
	   child = (CNode*) node->getEntry(i);
	   stk.push(child);	   
	 }
       } 			         
     }     
   }
}



bool UserFiltering( CRectangle* rect, int qID , int mID )
{  
   if(rect->contains(queryTable[qID]->loc))
     return false;      	
    
   double mindist = rect->getMinDist(queryTable[qID]->loc);
    
   if(mindist>queryTable[qID]->maxUpper)
     return true;
    
   if( mindist < queryTable[qID]->minUpper && fabs( mindist-queryTable[qID]->minUpper ) > EPSILON )
     return false;

   if(	mID < 0 )
   {  
     CPoint* cp =new CPoint( rect->m_low);
     double angle = queryTable[qID]->loc->getAngle(cp);
     int pNo = angle/angle_range;	  
     double dist = queryTable[qID]->loc->getDist(cp);
     double upper = queryTable[qID]->partitions[pNo].k_UpperArc;	  
     delete cp;
     if( dist < upper || fabs(dist-upper ) < EPSILON )
       return false;
     else
       return true;
   }
   double minAngle;
   double maxAngle;
    
    
   rect->getMaxMinAngle( &minAngle , &maxAngle , queryTable[qID]->loc );
    
   CPoint* closetPoint = rect->getClosestPoint(queryTable[qID]->loc);    //Get the cloest point;
   double closetAngle =queryTable[qID]->loc->getAngle(closetPoint);	
   delete closetPoint;
    			
   if(maxAngle-minAngle>180)  // IF the difference is greater than 180, this means the range is actually something like 320->0->40 
   {
     int rmax = minAngle/angle_range;
     for( int pNo1 = 0 ; pNo1<= rmax ; pNo1++ )
     {	    
       if( liesInPartition(rect , qID , pNo1 , closetAngle ))  // check lies in the partition
	 return false;	
     }

     int rmin = maxAngle/angle_range;
     for(int pNo2 = rmin ; pNo2 <pNum ; pNo2++ )
     {
       if(liesInPartition(rect , qID , pNo2 , closetAngle ))  // check lies in the partition
	 return false;       
     }
   }
   else{
     int pNo3 =  minAngle/angle_range;
     int imax =maxAngle/angle_range;
     for( pNo3 ; pNo3 <= imax ; pNo3++)
     {
       if( liesInPartition( rect , qID , pNo3 , closetAngle) )	
	 return false;
     }     
   }    
   return true;		
}


bool liesInPartition(CRectangle* rect, int qID, int pNo, double cangle)
{	
   double minDist;
   double s_angle = pNo*angle_range;
   double e_angle = s_angle + angle_range;
   double upper = queryTable[qID]->partitions[pNo].k_UpperArc;
    
   if( e_angle-cangle == 360 )
   {
     minDist=rect->getMinDist(queryTable[qID]->loc);
   }
   else{
     if( cangle >= s_angle && cangle <= e_angle )
     {
       minDist = rect->getMinDist(queryTable[qID]->loc);	
     }	
     else{	
       double dist1 = rect->MinDistAtAngle(queryTable[qID]->loc,s_angle);
       double dist2 = rect->MinDistAtAngle(queryTable[qID]->loc,e_angle);				
 
       if( dist1 < dist2 )
	 minDist=dist1;
       else
	 minDist=dist2;
     }	  
   }

   if( minDist < upper )   
     return true;
	
   return false;
}

bool isRkNN(CPoint* cp, int qID)
{
   double dist = cp->getDist(queryTable[qID]->loc);
   if( dist < queryTable[qID]->minLower && fabs(dist- queryTable[qID]->minLower) > EPSILON)
     return true;
    
   double angle = queryTable[qID]->loc->getAngle(cp);
   int pNo = angle/angle_range;
    
   if( dist < queryTable[qID]->partitions[pNo].k_LowerArc && fabs( dist-queryTable[qID]->partitions[pNo].k_LowerArc ) > EPSILON )  
     return true;
    
   int counter =0;
   list<minObject*>::iterator iList;
   iList = queryTable[qID]->partitions[pNo].minList.begin();
 
    
   for( iList = queryTable[qID]->partitions[pNo].minList.begin() ; iList!=queryTable[qID]->partitions[pNo].minList.end() && dist > (*iList)->key ; iList++ )    
   {
     double distoo = (*iList)->value->getDist(cp);
     double key = (*iList)->key;
	
     if( isMono )
     {
       if( distoo >= dist || fabs( distoo-dist ) < EPSILON ) 
	 continue;
       else{        
	 counter++;
	 if( counter >= k )	
	   return false;    
       }    	
     }
     else
       if( distoo < dist )
       {
	 counter++;
	 if( counter >= k )
	   return false;  	 
       }       
   } 
   return true;  
}




void SIX_MONO( int qID)
{ 
	
   isMono= true;
   pNum=6;
   angle_range = 60;  
   k=k+1;     
   queryTable[qID]->partitions = new Partition[6];
   if(qID%100==0)
     cout<<"query num "<<qID<<endl;
	
   for(int i = 0; i<6; i++)
   {
     queryTable[qID]->partitions[i].upperArc.insert(100);
   }
   p1_util->startCT();
   computeSixRegion(qID);
   p1_time += p1_util->endCT(false);
	
   p2_util->startCT();
   SIX_MONO_Refinement(qID);
   p2_time += p2_util->endCT(false);	
   k=k-1;	

   isMono = false;  
   resetQuery(qID);
}

void computeSixRegion( int qID)
{
   CHeap heap;
   Ftree->getRoot()->getRect()->blowHuge();
   heap.insertNode(0,Ftree->getRoot());
   CNode* curNode;
   CHeapNode* heapNode;
   int numChildren;
   CEntry* child;
	
   while(heap.getHeapSize()!=0)
   {
     heapNode = heap.top();
     curNode = (CNode*) heapNode->getData();		
     heap.pop();
	  
     if(curNode->m_id < 0)
     {
       EntryPrune_Point_Facility++;
    }else{
      EntryPrune_Node_Facility++;
    }
     
     if(SixRegionPrune(curNode->getRect(),qID))
       continue;
		
     if( curNode->m_id < 0 )
     {		
       CPoint* cp = new CPoint(curNode->getRect()->m_low);
       if (!cp->isSame(queryTable[qID]->loc))
       {
	 num_infObj++;
	 updateSixRegion(cp, qID , curNode->m_id*-1);
       }
       delete cp;       
     }
     else{
       curNode->load();
       numChildren = curNode->getNumChild();
       
      if(!foundInBuffer(curNode->m_id))
      {
	PRUNE_PHASE_IO++;
	FACILITY_TREE_IO++;	
      }   
      
       for(int i = 0 ; i< numChildren; ++i)
       {
	 child = curNode->getEntry(i);
	 heap.insertNode(child->getRect()->getMinDist(queryTable[qID]->loc), (CNode*) child);	 
       }
     }
   }
}



bool SixRegionPrune(CRectangle* rect, int qID )
{
   if(rect->contains(queryTable[qID]->loc))
     return false;
	
   double minDist = rect->getMinDist(queryTable[qID]->loc);
   
   if(minDist> queryTable[qID]->maxUpper)              // if mindist >Maxupper Prune safely
     return true;
   
   
   CPoint* closetPoint = rect->getClosestPoint(queryTable[qID]->loc);    //Get the cloest point;
   double closetAngle = queryTable[qID]->loc->getAngle(closetPoint);	
   delete closetPoint; 
   double minAngle;
   double maxAngle;
    
   rect->getMaxMinAngle(&minAngle , &maxAngle , queryTable[qID]->loc );
	
   if(maxAngle-minAngle>180)  // If the difference is greater than 180, this means the range is actually something like 320->0->40 
   {
     int rmax = minAngle/angle_range;
     for( int pNo1 = 0 ; pNo1 <= rmax ; pNo1++)
     {
       if( checkInPartitions( rect, qID , pNo1 , closetAngle ) )  // check lies in the partition
	 return false;	
     }
     int rmin = maxAngle/angle_range;
     for(int pNo2 = rmin ; pNo2 < pNum ; pNo2++ )
     {
       if( checkInPartitions( rect , qID , pNo2 , closetAngle ) )  // check lies in the partition
	 return false;       
     }
   }
   else{
     int pNo3 =  (int)(minAngle/angle_range)%6;
     int imax =  (int)(maxAngle/angle_range)%6;
     for(pNo3 ; pNo3 <= imax ; pNo3++)
     {
       if( checkInPartitions(rect , qID , pNo3 , closetAngle) )
	 return false;
     }
   }
  
   return true;
}


bool checkInPartitions( CRectangle* rect, int qID, int pNo, double cangle )    //For six region algorithm
{
   double minDist;
   double s_angle = pNo*angle_range;
   double e_angle = s_angle + angle_range;
 
   double upper = (double)*queryTable[qID]->partitions[pNo].upperArc.rbegin();

   if( e_angle - cangle == 360 )
     minDist=rect->getMinDist(queryTable[qID]->loc);
   else{
     if(cangle >= s_angle && cangle <= e_angle )
       minDist = rect->getMinDist(queryTable[qID]->loc);
     else{		
       double dist1 = rect->MinDistAtAngle(queryTable[qID]->loc,s_angle);
       double dist2 = rect->MinDistAtAngle(queryTable[qID]->loc,e_angle);				
       if( dist1 < dist2 )
	 minDist = dist1;
       else
	 minDist = dist2;       
     }     
   }
     
   if( minDist < upper || fabs(minDist-upper) < EPSILON )   
     return true;
	
   return false;
}



void updateSixRegion(CPoint* cp, int qID , int mID)
{
   FLAG = false;
   double angle_range = 60;	
   double angle = queryTable[qID]->loc->getAngle(cp);	
   int pNo = angle/angle_range;	
   double dist = queryTable[qID]->loc->getDist(cp);	
   double ruk = (double)*queryTable[qID]->partitions[pNo].upperArc.rbegin();
    
   if( dist < ruk  &&  fabs(dist-ruk) > EPSILON )
   {
     
     if( queryTable[qID]->maxUpper == 50000 || queryTable[qID]->minUpper == 0 )
       FLAG = true;    
     if( ruk == queryTable[qID]->maxUpper || ruk == queryTable[qID]->minUpper )
       FLAG = true;      
     
     if(queryTable[qID]->partitions[pNo].upperArc.size()==k)
     {
       FLAG = true;
       double temp = (double)*queryTable[qID]->partitions[pNo].upperArc.rbegin();
       queryTable[qID]->partitions[pNo].upperArc.erase(queryTable[qID]->partitions[pNo].upperArc.find(temp));
     }      
     queryTable[qID]->partitions[pNo].upperArc.insert(dist);    
     
   if( FLAG )
   { 
     double maxUpper =0;
     double minUpper =50000;
     for(int i = 0 ; i< pNum ; i++ )
     {
       if( (double)*queryTable[qID]->partitions[i].upperArc.rbegin() > maxUpper )
	 maxUpper = (double)*queryTable[qID]->partitions[i].upperArc.rbegin();
       
       if( (double)*queryTable[qID]->partitions[i].upperArc.rbegin() < minUpper )
	 minUpper = (double)*queryTable[qID]->partitions[i].upperArc.rbegin();
     }
     queryTable[qID]->maxUpper = maxUpper;
     queryTable[qID]->minUpper = minUpper;	  	  
   } 
     
	
     if( isMono )
     {
       CPoint* cpp = new CPoint(cp->coord);
       queryTable[qID]->candidates.insert(map<int, CPoint*>::value_type( mID , cpp) ); 
       UCandidates++;
     }
   }
}


void SIX_MONO_Refinement( int qID)
{
   CPoint* cp;
   map<int , CPoint* >::iterator ilist;
   ilist = queryTable[qID]->candidates.begin();
  
   verify_util->startCT();
   for(ilist ; ilist != queryTable[qID]->candidates.end() ; ilist++ )
   {
     cp = ilist->second;        
     if( ilist->first == qID )
       continue;
	
     if( lessThanKInRange(cp , cp->getDist( queryTable[qID]->loc ),qID )  )
       queryTable[qID]->RkNN.insert( ilist->first );
    }
     verify_time += verify_util->endCT(false);    
}



void SIX_BRNN( int qID)
{ 	
   pNum = 6;	
   angle_range = 60;

   if(qID%100==0)
     cout<<"query num "<<qID<<endl;

   queryTable[qID]->partitions = new Partition[6];

   for(int i = 0; i < 6 ; i++)
     queryTable[qID]->partitions[i].upperArc.insert(100);  
	 
   p1_util->startCT();
   computeSixRegion(qID);
   p1_time += p1_util->endCT(false);

   for(int i = 0 ; i< 6 ;i++)
     queryTable[qID]->partitions[i].k_UpperArc = (double)*queryTable[qID]->partitions[i].upperArc.rbegin(); 	   

  
   p2_util->startCT();
   SIX_BRNN_Refinement(qID);
   p2_time += p2_util->endCT(false);
   
    resetQuery(qID);
}


void SIX_BRNN_Refinement( int qID)
{
   CStack< CNode* > stk;
   long count = 0;
   Utree->getRoot()->getRect()->blowHuge();  
   stk.push(Utree->getRoot());
   CNode* node;
   CIndexNode* index = NULL;
   CLeafNode* leaf = NULL;
   CNode* child = NULL;
	
   while (!stk.empty())
   {
     node = stk.top();
     stk.pop();	
     
     
     if(node->m_id < 0)
     {
       EntryPrune_Point_User++;
    }else{
      EntryPrune_Node_User++;
    }
    
     if(SixRegionPrune(node->getRect(),qID))
       continue;
		 
     if( node->m_id < 0 )
     {
       UCandidates++;
       verify_util->startCT();
       CPoint* p = new CPoint(node->getRect()->m_low);

       if( lessThanKInRange( p , p->getDist(queryTable[qID]->loc),qID ) )
	 queryTable[qID]->RkNN.insert(node->m_id*-1);
       
       DISPLAY = false;
       delete p;
       verify_time += verify_util->endCT(false);    
     }
     else{		  
       node->load();
       REFINE_PHASE_IO++;
       USER_TREE_IO++;
       for (int i = 0; i < node->getNumChild(); i++)
       {		    
	 child = (CNode*) node->getEntry(i);
	 stk.push(child);
       }
     }
   }
}


void TPL_MONO(int qID)
{    
    
   if(qID%100==0)
     cout <<"query num " << qID << endl ; 
  
   isMono = true; 
    
   p1_util->startCT();  

   TPL_Pruning_Original(qID);    
   
   p1_time += p1_util->endCT(false);
  
    
   k=k+1; // since every object is dominated by itself, refinement is based on k+1
  
   p2_util->startCT();
   TPL_MONO_refinement(qID);
   p2_time += p2_util->endCT(false);
   k=k-1;   
    
   isMono = false;
   resetQuery(qID);
}



void TPL_MONO_OPT(int qID)
{    	
   if(qID%100==0)
     cout <<"query num " << qID << endl ; 
  
   isMono = true;   
   p1_util->startCT();  
   TPL_MONO_pruningPhase(qID);   
   p1_time += p1_util->endCT(false);
      
   k=k+1; // since every object is dominated by itself, refinement is based on k+1
  
   p2_util->startCT();
   TPL_MONO_refinement(qID);
   p2_time += p2_util->endCT(false);
   k=k-1;   
       
   isMono = false;

    resetQuery(qID);
}


void TPL_Pruning_Original( int qID)
{
   CHeap heap;
   Ftree->getRoot()->getRect()->blowHuge();
   heap.insertNode(0,Ftree->getRoot());

   CNode* curNode;
   CHeapNode* heapNode;
   int numChildren;
   CEntry* child;
   double minDist;
   int infListSize = 0; //
   while(heap.getHeapSize() != 0)
   {
     heapNode = heap.top();
     curNode = (CNode*)heapNode->getData(); 
     heap.pop();   
     DISPLAY = false;

   
      if(curNode->m_id < 0)
      {
	if(algoNum == TPLBRNN_OPT2)
	{
	  CPoint* x = new CPoint(curNode->getRect()->m_low);
	  if(!x->isSame(queryTable[qID]->loc))
	  {
	    num_infObj++;
	    MD_addInfObj(curNode->m_id*-1, x, qID);   
	  }
	  delete x;
	  continue;
	  
	}else
	EntryPrune_Point_Facility++;
      }
      else{
	EntryPrune_Node_Facility++;
      }
     
     double dist = K_Trim_TPL(curNode->getRect(),qID);
     DISPLAY = false;
       
#ifdef cmpTPLPruning
         
     int tmpCounter = getCounterOfRec( curNode->getRect(), qID);    
     
     if(curNode->m_id > 0)
     {
       if( dist >= MAX_DATA_VALUE )
       {
	 if( tmpCounter < k )
	 {
	   TPL_OLD_NEW_NODE++;
	 }	 
       }
    
       if( tmpCounter >= k )
       {
	 if( dist < MAX_DATA_VALUE)
	 {
	   TPL_NEW_OLD_NODE++;		   
	 } 
       }     
       
     }else
     { 
	if( dist >= MAX_DATA_VALUE )
	{
	  if( tmpCounter < k)
	  {
	    TPL_OLD_NEW_POINT++;
	  }
	}
	if( tmpCounter >= k )
	{	
	  if( dist < MAX_DATA_VALUE )
	  {
	    TPL_NEW_OLD_POINT++;
	  }
	}
    
     }    
#endif
	
     if( dist >= MAX_DATA_VALUE )
     {      
	
        if(algoNum == TPLBRNN_OPT2)
	{
	    if( curNode->m_id > 0)
	    {
	      queryTable[qID]->Nrfn.push_back( curNode );
	      continue; 
	    }
	}
	else
	{
	  if( curNode->m_id > 0)
	  {
	    queryTable[qID]->Nrfn.push_back( curNode );
	    continue; 
	  }
	  else
	  {
	    queryTable[qID]->Prfn.push_back( curNode );
	    continue;	 
	  }	  
        }
     }
     if (curNode->m_id < 0)   //data entry i.e; mbr of object
     { 	      
       CPoint* x = new CPoint(curNode->getRect()->m_low);
       if(!x->isSame(queryTable[qID]->loc))
       {
	 num_infObj++;
	 MD_addInfObj(curNode->m_id*-1, x, qID);   
       }
       delete x;
      }
      else
      {  	  
	curNode->load();
	numChildren = curNode->getNumChild();      
	PRUNE_PHASE_IO++;
	FACILITY_TREE_IO++;
	
	for (int i = 0; i < numChildren; ++i){
	  child = curNode->getEntry(i);
	  
	  if(child->m_id < 0)
	  {
	    if(algoNum == TPLBRNN_OPT2 )
	    {
	      minDist = child->getRect()->getMinDist( queryTable[qID]->loc);
	      heap.insertNode(minDist , (CNode*) child);
	      continue;
	    }else
	      EntryPrune_Point_Facility++;
	  }else{
	    EntryPrune_Node_Facility++;
	  }
	 minDist = K_Trim_TPL( child->getRect(), qID);
	
#ifdef cmpTPLPruning
	int tmpCounter2 = getCounterOfRec( child->getRect(), qID);
	if(child->m_id > 0)
	{
	  if( minDist >= MAX_DATA_VALUE )
	  {	  
	    if(tmpCounter2 < k)
	    {
	      TPL_OLD_NEW_NODE++;
	    }	    	    
	  }
	  
	  if( tmpCounter2 >= k )
	  {	    
	    if(minDist < MAX_DATA_VALUE)
	    {
	      TPL_NEW_OLD_NODE++;
	    }
	  }	  
	}
	else
	{ 
	  if( minDist >= MAX_DATA_VALUE )
	  {	 
	    if( tmpCounter2 < k)
	    {
	      TPL_OLD_NEW_POINT++;
	    }
	  }	  
	  if( tmpCounter2 >= k )
	  {	   
	    if( minDist < MAX_DATA_VALUE)
	    {
	     TPL_NEW_OLD_POINT++;
	    }
	  }	  	
	}     
#endif
	  
	  if( minDist >= MAX_DATA_VALUE )
	  {
	    if(algoNum == TPLBRNN || algoNum == TPLMONO )
	    {
	       if( child->m_id > 0)
		{
		  queryTable[qID]->Nrfn.push_back( (CNode*)child );
		  continue;
		}
		else
		{
		  queryTable[qID]->Prfn.push_back( (CNode*)child );
		  continue;
		}	      
	     }
	     if( algoNum == TPLBRNN_OPT2)
	     {
	        if( child->m_id > 0)
		{
		  queryTable[qID]->Nrfn.push_back( (CNode*)child );
		  continue;
		}
	     }
	  }
	   else
	    heap.insertNode(minDist , (CNode*) child);
	 }
       }
    }
}


double K_Trim_TPL( CRectangle* rec , int qID)
{
  
   int n = queryTable[qID]->infList.size();
    
   if( n < k )
   {     
     return rec->getMinDist( queryTable[qID]->loc);
   }
   list<infObj*>::iterator iList;
   int i,j;
   double result=0;
   	
 
   
   double* q = new double[DIMENSION];	
   for( int d = 0 ; d< DIMENSION ; d++ )
   {
     q[d]= queryTable[qID]->loc->coord[d];
   }
   infObj** list = new infObj*[n] ;
   double * p=new double [DIMENSION];	
   
  
   
   double * bounces = new double [2*DIMENSION];
   for (int d=0 ; d<DIMENSION ; d++)
   {						
     bounces[2*d]=rec->m_low[d];
     bounces[2*d+1]=rec->m_high[d];
   }
   
   int ti = 0;
   for( iList = queryTable[qID]->infList.begin() ; iList != queryTable[qID]->infList.end() ; iList++)
   {	
     list[ti]=  (*iList) ;		
     ti++;
   }	  
   

   double *newBounces=new double[2*DIMENSION];
   double *resBounces=new double[2*DIMENSION];
   double *tmpBounces=new double[2*DIMENSION];
   double *cmpBounces=new double[2*DIMENSION];
   
   memcpy(resBounces,bounces,2*DIMENSION*sizeof(double));

   double *z=new double[DIMENSION];

   double *A=new double[DIMENSION];

   double d,c;

   int counter;
   bool firsttime;
   
#ifdef TPLEnforce		
   while(true)
#endif
   {     
     memcpy(cmpBounces,resBounces,2*DIMENSION*sizeof(double));

     for (int u=0;u<n;u++)
     {
       counter=0;
       firsttime=true;
       for (int v=0;v<k;v++)
       {
	 int index=(u+v)%n;
	 
	
	 
	for (i=0;i<DIMENSION;i++)
	{
	  p[i] = list[index]->loc->coord[i];	  
	}
	 
	for (i=0;i<DIMENSION;i++)
	{
	  A[i]=2*(q[i]-p[i]);
	}
	
	c=0;
	for (i=0;i<DIMENSION;i++)
	{
	  c+=q[i]*q[i]-p[i]*p[i];	  
	}
	memcpy(newBounces,resBounces,2*DIMENSION*sizeof(double));
	for (i=0;i<DIMENSION;i++)
	{
	  if (A[i]>0)
	    z[i]=newBounces[2*i+1];
	  else
	    z[i]=newBounces[2*i];
	}
	d=c;
	for (i=0;i<DIMENSION;i++)
	{
	  d-=A[i]*z[i];
	}
	if (d>0)
	{	  
	  counter++;
	  continue;
	}

	for (i=0;i<DIMENSION;i++)
	{
	  
	  if (A[i]>0)
	  {
	    newBounces[2*i]=max(newBounces[2*i],newBounces[2*i+1]+d/A[i]);
	  }
	  if (A[i]<0)
	  {
	    newBounces[2*i+1]=min(newBounces[2*i+1],newBounces[2*i]+d/A[i]);
	  }
	}		
	if (firsttime==true)		
	{
	  firsttime=false;
	  for (i=0;i<DIMENSION;i++)
	  {						
	    tmpBounces[2*i]=newBounces[2*i];
	    tmpBounces[2*i+1]=newBounces[2*i+1];
	  }
	}
	else
	{
	  for (i=0;i<DIMENSION;i++)
	  {
	    {				
	      tmpBounces[2*i+1]=max(newBounces[2*i+1],tmpBounces[2*i+1]);				tmpBounces[2*i]=min(tmpBounces[2*i],newBounces[2*i]);  
	    }
	  }	
	} 
      }
      if (counter==k)
      {
	result=MAX_DATA_VALUE;
	delete []list;
	delete []p;
	delete []z;
	delete []A;
	delete []newBounces;
	delete []resBounces;
	delete []tmpBounces;
	delete []cmpBounces;
	delete [] bounces;
	delete [] q;
	return result;
      }
      memcpy(resBounces,tmpBounces,2*DIMENSION*sizeof(double));
    }
     
    double areaRes = (resBounces[1]-resBounces[0])*(resBounces[3]-resBounces[2]);
    double areaCmp = (cmpBounces[1]-cmpBounces[0])*(cmpBounces[3]-cmpBounces[2]);
    
#ifdef TPLEnforce 
    if( areaRes/areaCmp > 0.99 || areaRes==areaCmp )
      break;  
#endif     
   
   }
   
   
   if (result != MAX_DATA_VALUE )
   {
     CRectangle* tmpRec = new CRectangle;	
     for (int d=0 ; d<DIMENSION ; d++)
     {						
       tmpRec->m_low[d] = bounces[2*d];
       tmpRec->m_high[d] = bounces[2*d+1];
    }
    result = tmpRec->getMinDist( queryTable[qID]->loc);	
    delete tmpRec;
  }

  delete []list;
  delete []p;
  delete []z;
  delete []A;
  delete []newBounces;
  delete []resBounces;
  delete []tmpBounces;
  delete []cmpBounces;
  delete [] bounces;
  delete [] q;

  return result;
	
}


bool cmpHilbertValue( const infObj* s1, const infObj* s2)
{ 
   return s1->hilbertValue < s2->hilbertValue ;
}


bool cmpObjID(const infObj* s1, const infobj* s2) 
{
   return s1->objID < s2->objID ;
}

void TPL_MONO_pruningPhase(int qID)
{

   CHeap heap;
   Ftree->getRoot()->getRect()->blowHuge();
   heap.insertNode(0,Ftree->getRoot());

   CNode* curNode;
   CHeapNode* heapNode;
   int numChildren;
   CEntry* child;

   int infListSize = 0; //
   while(heap.getHeapSize() != 0)
   {
     heapNode = heap.top();
     curNode = (CNode*)heapNode->getData(); 

     heap.pop();
     
     TPL_MONO_POINT_PRUNED = false;
     
     if(getCounterOfRec(curNode->getRect(),qID)>=k)
     {    
       if( curNode->m_id > 0)
       {
	 queryTable[qID]->Nrfn.push_back( curNode );
	 continue;
       }
       else{
	 TPL_MONO_POINT_PRUNED = true;
	 queryTable[qID]->Prfn.push_back(curNode);
      }
     }
        
     if (curNode->m_id < 0){ //data entry i.e; mbr of object
       CPoint* x = new CPoint(curNode->getRect()->m_low);
       if(!x->isSame(queryTable[qID]->loc))
       {         
	
	 MD_addInfObj(curNode->m_id*-1, x, qID);
	 num_infObj++;   
       }
       delete x;
     }
     else{     
       
       curNode->load();
       numChildren = curNode->getNumChild(); 
       PRUNE_PHASE_IO++;
       FACILITY_TREE_IO++;
       for (int i = 0; i < numChildren; ++i){
	 child = curNode->getEntry(i);
	 heap.insertNode(child->getRect()->getMinDist(queryTable[qID]->loc),(CNode*) child);
       }
     }
   }   
}



void MD_addInfObj(int objID, CPoint* point, int qID)
{
   infObj* x = new infObj(); 
   x->objID = objID;
   x->loc = new CPoint(point->coord);
   x->eq = new CEquation(queryTable[qID]->loc, x->loc);
   x->distFromQ = queryTable[qID]->loc->getDist(x->loc);
    
   if(algoNum==TPLBRNN || algoNum == TPLMONO || algoNum == TPLBRNN_OPT2)
   {
     x->hilbertValue = point->cal_hilbert(15, 2, 1.0);
     list<infObj*>::iterator it;
     
     it = std::upper_bound(queryTable[qID]->infList.begin() , queryTable[qID]->infList.end() , x , cmpHilbertValue);        
     queryTable[qID]->infList.insert(it,x);    
     
   }
   else   
     queryTable[qID]->infList.push_back(x);
   
   if( isMono )
   {
     if(algoNum == TPLMONO_OPT)
     {
       if(!TPL_MONO_POINT_PRUNED)
       {
	  UCandidates++;
	  TPLCand* y = new TPLCand();
	  y->counter = k;
	  y->candID = objID;
	  y->p = new CPoint(point->coord);
	  queryTable[qID]->candList.push_back( y );  
       }
       else
       {
	 
       }
     }
     else 
     { 
        UCandidates++;
	TPLCand* y = new TPLCand();
	y->counter = k;
	y->candID = objID;
	y->p = new CPoint(point->coord);
	queryTable[qID]->candList.push_back( y );         
    }
   }
}


void MD_addCands(int objID, CPoint* point, int qID)
{
   TPLCand* y = new TPLCand();
   y->counter = k;
   y->candID = objID;
   y->p = point;
   queryTable[qID]->candList.push_back( y );
}



void TPL_MONO_refinement( int qID)
{
   list<TPLCand*>::iterator iList;
   list<TPLCand*>::iterator iList2;   
   
   verify_util->startCT();
   for(iList = queryTable[qID]->candList.begin(); iList != queryTable[qID]->candList.end(); )
   {
     if((*iList)->candID <0)
       continue;
	
     for( iList2 = iList; iList2 != queryTable[qID]->candList.end() ; iList2++)
     {
       if(  (*iList)->candID != (*iList2)->candID )
       {
	 if( (*iList)->p->getDist((*iList2)->p) < (*iList)->p->getDist(queryTable[qID]->loc) )
	 {
	   (*iList)->counter--;
	 }
	 if( (*iList)->p->getDist((*iList2)->p) < (*iList2)->p->getDist(queryTable[qID]->loc) )
	 {
	   (*iList2)->counter--;	   
	 }
       }       
     }
     if( (*iList)->counter <= 0 )
     {
       iList = queryTable[qID]->candList.erase( iList );    
     }
     else
       iList++;
   }
    
   
   int numofChild;
   CNode* nextNode;
   
   while( !queryTable[qID]->candList.empty() )
   {
     
     
     TPL_Refinement_Round( qID);
    
     if( queryTable[qID]->candList.empty() )
     {
       return ;
     }
     queryTable[qID]->Prfn.clear();
     queryTable[qID]->Nrfn.clear();
     nextNode = findNext( qID) ;
	
     numofChild = nextNode->getNumChild();
     
     nextNode->load();
     
     VERIFY_PHASE_IO++;
     VERIFY_BUFFERED_IO++;
     FACILITY_TREE_IO++;
   
     if( nextNode->isLeaf())   //Leaf Node
     {  
       for( int i = 0 ; i < numofChild ; ++i)  
       {
	 queryTable[qID]->Prfn.push_back((CNode*) nextNode->getEntry(i));
       }
     }
     else{     //Intermediate Node
       for( int j = 0 ; j < numofChild ; ++j)
       {
	 queryTable[qID]->Nrfn.push_back((CNode*) nextNode->getEntry(j));
       }
     }	
   }  
   verify_time+= verify_util->endCT(false);
}


void TPL_Refinement_Round( int qID)
{
   list<TPLCand*>::iterator iList;
   list<CNode*>::iterator pList;
   list<CNode*>::iterator nList;
   CPoint* p;
  
    
   for( iList = queryTable[qID]->candList.begin() ; iList != queryTable[qID]->candList.end() ; )
   {       
     for( pList = queryTable[qID]->Prfn.begin() ; pList != queryTable[qID]->Prfn.end() ; pList++)
     {
       p = new CPoint( (*pList)->getRect()->m_low );
       if( (*iList)->p->getDist(p) < (*iList)->p->getDist( queryTable[qID]->loc) )
       {
	 (*iList)->counter--;	
       }
       if( (*iList)->counter <= 0 )
       {
	
	 iList = queryTable[qID]->candList.erase( iList );
	 delete p;
	 goto next;
       }	
       delete p;
     }
     
     for( nList = queryTable[qID]->Nrfn.begin() ; nList != queryTable[qID]->Nrfn.end() ; nList++)
     {
       if( (*nList)->getRect()->getMaxDist((*iList)->p) < (*iList)->p->getDist(queryTable[qID]->loc) && (*nList)->m_dataCnt >= (*iList)->counter )
       {
	 iList = queryTable[qID]->candList.erase( iList );
	 goto next;
       }
     }
     
     for( nList = queryTable[qID]->Nrfn.begin() ; nList != queryTable[qID]->Nrfn.end() ; nList++)
     {
       if( (*nList)->getRect()->getMinDist( (*iList)->p) < (*iList)->p->getDist( queryTable[qID]->loc ) )
       {
	 (*iList)->toVisit.push_back(  (*nList)  );
       }
     }
     if( (*iList)->toVisit.size() == 0 )
     {
       queryTable[qID]->RkNN.insert( (*iList)->candID );
       iList = queryTable[qID]->candList.erase( iList );
       goto next;
     }
     iList++;		  
     next:  ;
   }
}


CNode* findNext( int qID)
{
   CNode* next ;
   int counter = 0;
   int level = 500000;
   list<TPLCand*>::iterator iList;
   list<CNode*>::iterator iList2;
   list<counterObject*> nextList;
   list<counterObject*>::iterator iNext;
    
   for( iList = queryTable[qID]->candList.begin() ; iList != queryTable[qID]->candList.end() ; iList++)
   {
     for( iList2 = (*iList)->toVisit.begin() ; iList2 != (*iList)->toVisit.end() ; iList2++)
     {
       if( (*iList2)->m_level < level )
       {	     
	 level = (*iList2)->m_level;
	 nextList.clear();
	 counterObject* nextObject = new counterObject;
	 nextObject->counter = 1;
	 nextObject->node = (*iList2);
	 nextList.push_back( nextObject );	     
       }
       else
	 if( (*iList2)->m_level == level )
	 {      
	   bool isSame = false;
	   for( iNext = nextList.begin() ; iNext != nextList.end() ; iNext++)
	   {
	     if( (*iList2)->m_id == (*iNext)->node->m_id )
	     {		      
	       (*iNext)->counter++;
	       isSame = true;
 	     }
	   }
	   if( !isSame)
	   {		  
	     counterObject* nextObject2 = new counterObject;
	     nextObject2->counter = 1;
	     nextObject2->node = (*iList2);	    
	     nextList.push_back( nextObject2 );		 
	   }
	 }
      }
   }
    
   for( iNext = nextList.begin() ; iNext != nextList.end() ; iNext++)
   {
     if( (*iNext)->counter > counter )
     {
       next = (*iNext)->node;
       counter = (*iNext)->counter;
     }
   }
    
    
   for( iList = queryTable[qID]->candList.begin() ; iList != queryTable[qID]->candList.end() ; iList++ )
   {
     if( (*iList)->toVisit.size()>0 )
     {
       for( iList2 = (*iList)->toVisit.begin() ; iList2 != (*iList)->toVisit.end() ; )
       {
	 if( (*iList2)->m_id == next->m_id )
	 {
	   iList2 = (*iList)->toVisit.erase( iList2 );
	   counter --;
	 }
	 else{
	   iList2++;
 	 }	
	if( counter == 0)
	{
	  return next;
	}
       }
     }     
   }
   return next;  
}


void TPL_BRNN(int qID)
{	
   if(qID%100==0)
     cout <<"query num " << qID << endl; 
 
   if( algoNum == TPLBRNN_OPT1)
   {
      p1_util->startCT();
      TPL_BRNN_pruningPhase(qID);
      p1_time += p1_util->endCT(false);
	
      p2_util->startCT();    
      TPL_BRNN_refinement(qID);
      p2_time += p2_util->endCT(false);
   }
   else{
    p1_util->startCT();
    TPL_Pruning_Original(qID); 
    p1_time += p1_util->endCT(false);
      
    p2_util->startCT();    
    TPL_BRNN_refinement_OLD(qID);
    p2_time += p2_util->endCT(false);
   }
	
    resetQuery(qID);
    
}


void TPL_BRNN_OPT(int qID)
{

   if(qID%100==0)
     cout <<"query num " << qID << endl; 

    
   p1_util->startCT();
   TPL_BRNN_pruningPhase(qID);
   p1_time += p1_util->endCT(false);
        
   p2_util->startCT();    
   TPL_BRNN_refinement(qID);
   p2_time += p2_util->endCT(false);

    resetQuery(qID);
    
}



void TPL_BRNN_pruningPhase(int qID)
{
   CHeap heap;
   Ftree->getRoot()->getRect()->blowHuge();
   heap.insertNode(0,Ftree->getRoot());

   CNode* curNode;
   CHeapNode* heapNode;
   int numChildren;
   CEntry* child;

   while(heap.getHeapSize() != 0)
   {
     heapNode = heap.top();        
     curNode = (CNode*)heapNode->getData(); 
     heap.pop();
      
     //Only for node
     if( curNode->m_id >= 0)
     {
          EntryPrune_Node_Facility++;
	  if(getCounterOfRec(curNode->getRect(),qID)>=k)
	  {
	
		    if( curNode->m_id > 0)
		    {
		      queryTable[qID]->Nrfn.push_back( curNode );
		      continue;	 
		    }   
	    
	  }	    
    }else{  //For points 
      
      if(algoNum == TPLBRNN_OPT1 )
      {
	EntryPrune_Point_Facility++;
	if(getCounterOfRec(curNode->getRect(),qID)>=k)
	{
	  queryTable[qID]->Prfn.push_back(curNode);
	  continue;	  
	}
      }            
    }
           
     
     if (curNode->m_id < 0){ 
       CPoint* x = new CPoint(curNode->getRect()->m_low);
       if(!x->isSame(queryTable[qID]->loc))
       {    
	 num_infObj++;
	 MD_addInfObj(curNode->m_id*-1, x, qID);		
       }
       delete x;
     }
     else{	  
       curNode->load();
       numChildren = curNode->getNumChild();
       
       PRUNE_PHASE_IO++;
       FACILITY_TREE_IO++;

       for (int i = 0; i < numChildren; ++i){
	 child = curNode->getEntry(i);
	 heap.insertNode(child->getRect()->getMinDist(queryTable[qID]->loc),(CNode*) child);
       }
     }
   }  
}




void TPL_BRNN_refinement_OLD( int qID)
{
   CStack< CNode* > stk;
   long count = 0;

   Utree->getRoot()->getRect()->blowHuge();
  
   stk.push(Utree->getRoot());
   CNode* node;
   
   CIndexNode* index = NULL;
   CLeafNode* leaf = NULL;
   CNode* child = NULL;

   while (!stk.empty())
   {
       node = stk.top();
       stk.pop();
       
        if(node->m_id < 0)
	{
	  EntryPrune_Point_User++;
	}else{
	  EntryPrune_Node_User++;
	}
    
       if( K_Trim_TPL(node->getRect(),qID) >= MAX_DATA_VALUE)
       {
           continue;
       }
       
       if (node->m_id<0)
       {        
	 CPoint* p = new CPoint( node->getRect()->m_low );
	 UCandidates++;
	 MD_addCands(node->m_id*(-1), p , qID );	        
       }
       else
       {       

	 REFINE_PHASE_IO++;  
	 USER_TREE_IO++;

	 for (int i = 0 ; i < node->getNumChild() ; i++ ) 
	 {
	   child = (CNode*) node->getEntry(i);
	   stk.push(child);
	 }	    
       }
   }
    
   verify_util->startCT(); 
   TPL_BRNN_Verification( qID );
   verify_time += verify_util->endCT(false);
   
}



void TPL_BRNN_refinement(int qID)
{
   CStack< CNode* > stk;
   long count = 0;

   Utree->getRoot()->getRect()->blowHuge();
  
   stk.push(Utree->getRoot());
   CNode* node;

   CIndexNode* index = NULL;
   CLeafNode* leaf = NULL;
   CNode* child = NULL;

   while (!stk.empty())
   {
     node = stk.top();
     stk.pop();
     
      if(node->m_id < 0)
      {
	EntryPrune_Point_User++;
      }else{
	EntryPrune_Node_User++;
      }
    
     if( getCounterOfRec(node->getRect(),qID) >= k)
     {
       continue;
     }
        
     if (node->m_id<0)
     {
       CPoint* p = new CPoint( node->getRect()->m_low );	  	        
       UCandidates++;	
       MD_addCands(node->m_id*(-1), p , qID );    
     }
     else
     {       
       node->load();  
       REFINE_PHASE_IO++;  
       USER_TREE_IO++;
       for (int i = 0 ; i < node->getNumChild() ; i++ ) 
       {
	 child = (CNode*) node->getEntry(i);
	 stk.push(child);
       }    
     }
   }
   
    verify_util->startCT(); 
    TPL_BRNN_Verification( qID );
    verify_time += verify_util->endCT(false);
  
}



void TPL_BRNN_Verification( int qID)
{
   list<TPLCand*>::iterator iList;
   list<infobj*>::iterator iList2;

   for(iList = queryTable[qID]->candList.begin(); iList != queryTable[qID]->candList.end(); )
   {
     for( iList2 = queryTable[qID]->infList.begin() ; iList2 != queryTable[qID]->infList.end() ; iList2++)
     {
       if( (*iList)->p->getDist( (*iList2)->loc ) < (*iList)->p->getDist( queryTable[qID]->loc ))
       {
	 (*iList)->counter--;	 
       }
     }
     if( (*iList)->counter <= 0 )
     {
       deleteTPLCands( (*iList));
       iList = queryTable[qID]->candList.erase( iList );   
     }
     else
       iList++;
   }
   
   int numofChild;
   CNode* nextNode;

   while( !queryTable[qID]->candList.empty() )
   {
     TPL_Refinement_Round( qID);
     if( queryTable[qID]->candList.empty() )
     {
       return ;
     }

     queryTable[qID]->Prfn.clear();
     queryTable[qID]->Nrfn.clear();

     nextNode = findNext( qID) ;
     nextNode->load();
     numofChild = nextNode->getNumChild();

     VERIFY_PHASE_IO++;
     VERIFY_BUFFERED_IO++;
     FACILITY_TREE_IO++;
     
     if( nextNode->isLeaf())   //Leaf Node
     {  
       for( int i = 0 ; i < numofChild ; ++i)  
       {
	 queryTable[qID]->Prfn.push_back((CNode*) nextNode->getEntry(i));
       }
     }
     else{     //Intermediate Node
       for( int j = 0 ; j < numofChild ; ++j)
       {
	 queryTable[qID]->Nrfn.push_back((CNode*) nextNode->getEntry(j));
       }
     }	
   }     
}



void INFZONE_MONO(int qID)
{
   if(qID%100==0)
     cout <<"query num " << qID << endl; 

   k=k+1;
    
   p1_util->startCT();
   computeInfluenceZone(qID);
   
   constructPolygon(qID);


   p1_time += p1_util->endCT(false); 
   
   p2_util->startCT();
   INFZONE_MONO_refinement(qID);
   p2_time += p2_util->endCT(false);
    
   k=k-1;

   resetQuery(qID);
}

bool liesOutOfSR(CPoint* p, int qID)
{
   if(queryTable[qID]->vPolygon.size()==0)
     return false;
 
   if( algoNum == FINCHMONO_OPT || algoNum == FINCHBRNN_OPT)  
   {
     if(p->getDist(queryTable[qID]->loc) > queryTable[qID]->maxRadius)
     {
      
       return true;
     }
     
     if(!queryTable[qID]->approxRegion->isIntersected(p))
     {
       return true;
       
    }
   }

   vector<Vertex*>::iterator iList;
   vector<Vertex*>::iterator iFirst;
   vector<Vertex*>::iterator iSecond;
   Vertex* v = new Vertex();
   v->angleFromQ = queryTable[qID]->loc->getAngle(p);
    
   iSecond = std::upper_bound(queryTable[qID]->vPolygon.begin(), queryTable[qID]->vPolygon.end(), v, sortVerticesOnAngle);
   
   delete v;
   
    
    // set the iterators to proper elements
   if(iSecond == queryTable[qID]->vPolygon.end())
   {
     iFirst = queryTable[qID]->vPolygon.begin();
     iSecond--;
   }
   else
   {
     if(iSecond == queryTable[qID]->vPolygon.begin())
     {
       iFirst = queryTable[qID]->vPolygon.begin();
       iList = queryTable[qID]->vPolygon.end();
       iList--;
       iSecond = iList;
     }
     else
     {
       iList = iSecond;
       iList -- ;
       iFirst = iList;
     }
   }

   if(edgesIntersect(queryTable[qID]->loc, p, (*iFirst)->loc, (*iSecond)->loc))
   {
     return true;
   }
   else
   {
     return false;
   }
}


bool liesOutOfPolygon(CPoint* p, int qID)
{
   vector<Vertex*>::iterator iList;
   vector<Vertex*>::iterator iFirst;
   vector<Vertex*>::iterator iSecond;
   Vertex* v = new Vertex();
   v->angleFromQ = queryTable[qID]->loc->getAngle(p);
   
   iSecond = std::upper_bound(queryTable[qID]->vPolygon.begin(), queryTable[qID]->vPolygon.end(), v, sortVerticesOnAngle);
    
    // set the iterators to proper elements
   if(iSecond == queryTable[qID]->vPolygon.end())
   {
     iFirst = queryTable[qID]->vPolygon.begin();
     iSecond--;
   }
   else{
     if(iSecond == queryTable[qID]->vPolygon.begin())
     {
       iFirst = queryTable[qID]->vPolygon.begin();
       iList = queryTable[qID]->vPolygon.end();
       iList--;
       iSecond = iList;
     }
     else{
       iList = iSecond;
       iList -- ;
       iFirst = iList;
     }
   }
     
   delete v;
   infobj* obj = getCommonObj((*iFirst), (*iSecond));
  
    
   if( p->pliesInSpace(queryTable[qID]->loc,obj->loc) )
     return true;
   else
     return false;
}

infobj* getCommonObj(Vertex* v1, Vertex* v2)
{
  // check if the first object matches any of the two objects in v2
   if(v1->obj1->objID == v2->obj1->objID || v1->obj1->objID == v2->obj2->objID)
     return v1->obj1;
    
    
    // check if the second object matches any of the two objects in v2
   if(v1->obj2->objID == v2->obj1->objID || v1->obj2->objID == v2->obj2->objID)
     return v1->obj2;
    
   return v1->obj2;
}


void constructPolygon(int qID)
{
   list<Vertex*>::iterator iList;

    // insert the vertices that lie on polygon in vPolygon
   for(iList = queryTable[qID]->vList.begin() ; iList != queryTable[qID]->vList.end() ; iList++)
   {
     if( (*iList)->counter >= k-2 )
     {
       (*iList)->angleFromQ = queryTable[qID]->loc->getAngle((*iList)->loc);
       queryTable[qID]->vPolygon.push_back((*iList));
     }
     else
     {
       if(liesAtDataSpaceBoundary((*iList)->loc))
       {
	 (*iList)->angleFromQ = queryTable[qID]->loc->getAngle((*iList)->loc);
	 queryTable[qID]->vPolygon.push_back((*iList));
       }
     }
  }

// vPolygon is stored in a vector instead of a list so that a binary search can be done in logarithmic time
   std::sort(queryTable[qID]->vPolygon.begin(), queryTable[qID]->vPolygon.end(), sortVerticesOnAngle); 

}

bool correctPolygon(int qID)
{  
   vector<Vertex*>::iterator iList;
   iList = queryTable[qID]->vPolygon.end();
   iList--;
   int obj1=(*iList)->obj1->objID;
   int obj2=(*iList)->obj2->objID;
   bool matched=false;
   bool rValue = true;
   int i=1;
   for(iList = queryTable[qID]->vPolygon.begin(); iList!= queryTable[qID]->vPolygon.end(); iList++)
   {
     if(obj1== (*iList)->obj1->objID || obj1==(*iList)->obj2->objID)
     {
       matched=true;
     }
     if(obj2== (*iList)->obj1->objID || obj2==(*iList)->obj2->objID)
     {
       matched=true;
     }
     if(matched==false)
     {
       cout << "Problem at vertex: " << i << endl;
       rValue = false;       
     }
     matched=false;
     obj1=  (*iList)->obj1->objID;
     obj2=  (*iList)->obj2->objID;
     i++;     
   }
    
   if(rValue == false)
     printPolygon(qID);
   return rValue;
}

void printPolygon(int qID)
{
   vector<Vertex*>::iterator iList;
   int i =1;
   for(iList = queryTable[qID]->vPolygon.begin(); iList!= queryTable[qID]->vPolygon.end(); iList++)
   {
     cout << "v " << i << " : ";
     (*iList)->loc->print();
     cout << "  counter: " << (*iList)->counter << "  obj1: " << (*iList)->obj1->objID << " obj2: " << (*iList)->obj2->objID <<  " distFromQ: " << (*iList)->distFromQ << " angleFromQ: " << (*iList)->angleFromQ <<  endl;
     i++; 
   }
}

bool correctCounters(int qID)
{
   list<Vertex*>::iterator iList;
   int i = 1;
    
   for(iList = queryTable[qID]->vList.begin(); iList!= queryTable[qID]->vList.end(); iList++)
   {
     int num = computeCounter( (*iList)->loc, qID );
     if( (*iList)->counter != num )
     {
       cout << "v " << i << " : ";
       (*iList)->loc->print();
       cout << "  counter: " << (*iList)->counter << "  obj1: " << (*iList)->obj1->objID << " obj2: " << (*iList)->obj2->objID <<  " distFromQ: " << (*iList)->distFromQ << endl;
       cout << " ComputeCounter returned: " << num << endl;
       return false;       
     }
     i++;     
   }
   return true; 
}


void INFZONE_BRNN(int qID)
{  
   if(qID%100==0)
     cout <<"query num " << qID << endl; 

   p1_util->startCT();
   computeInfluenceZone(qID);
   constructPolygon(qID);


   p1_time += p1_util->endCT(false);    
   
   p2_util->startCT();
   INFZONE_BRNN_refinement(qID);
   p2_time += p2_util->endCT(false);
    
  
   resetQuery(qID);
}


void  INFZONE_BRNN_refinement(int qID)
{
   CStack< CNode* > stk;
   long count = 0;
   Utree->getRoot()->getRect()->blowHuge();  
   stk.push(Utree->getRoot());    
   CNode* node;
   CIndexNode* index = NULL;
   CLeafNode* leaf = NULL;
   CNode* child = NULL;

   while ( !stk.empty() )
   {
     node = stk.top();
     stk.pop();

     if ( node->m_id < 0 )
     {
       verify_util->startCT();
       CPoint* p = new CPoint( node->getRect()->m_low );
       
       EntryPrune_Point_User++;
       if( outsideInfZone(p,qID) ==  0)
       {
	 // if not outside, then it is definitely inside
	 UCandidates++;
	 queryTable[qID]->RkNN.insert( node->m_id*-1 );
	 
	 if(cmpIOBound)
	 { 
	   CPoint* tmp = new CPoint( p->coord );	
	   RkNN_Points.insert(tmp);		  
   	 }
       }
       delete p;
       verify_time += verify_util->endCT(false);
     }
     else{
       
       EntryPrune_Node_User++;
       if( !outsideInfZone( node->getRect() , qID ) )
       {
	 node->load();  
	 REFINE_PHASE_IO++;  
	 USER_TREE_IO++;
	 for (int i = 0 ; i < node->getNumChild() ; i++ ) 
	 {
	   child = (CNode*) node->getEntry(i);
	   stk.push(child);	   
	 }	 
       }       
     }
   }
}


void  INFZONE_MONO_refinement(int qID)
{
   list<infObj*>::iterator iList;
   for(iList = queryTable[qID]->infList.begin(); iList!= queryTable[qID]->infList.end(); iList++)
   {
     if((*iList)->objID<0)
       continue;
	
     verify_util->startCT();
     int status = outsideInfZone((*iList)->loc,qID);
     if( status == 0)
       queryTable[qID]->RkNN.insert((*iList)->objID);
     else{
       if( status == 2 )
	 break;	
     }
     verify_time += verify_util->endCT(false);    
   }
}


int outsideInfZone(CPoint* p, int qID)
{
  // 0 means does not lie outsideInfZone
  // 1 means lies outside safe zone
  // 2 means lies outside and was returned by the first simple check

   double minD = p->getDist(queryTable[qID]->loc);
   //num_UserPruned++;

    // if further than maxRadius then outside
   if(minD > queryTable[qID]->maxRadius && fabs(minD-queryTable[qID]->maxRadius) > 1)
   {
     return 2;
   }   
   if(minD < queryTable[qID]->minRadius && fabs(minD -queryTable[qID]->minRadius)>0.00001 )
   {
     return 0;
   }

   if(liesOutOfPolygon(p,qID)) // 
   {
     return 1;
   }
   else
   {
     return 0;
   }
}


bool outsideInfZone(CRectangle* rec, int qID)
{
   double minD = rec->getMinDist(queryTable[qID]->loc);
    // if further than maxRadius then outside
   //num_UserPruned++;
   if(minD > queryTable[qID]->maxRadius) 
     return true;  
    
   if(minD < queryTable[qID]->minRadius)
     return false;
    
#ifdef InfZone_opt
   if(overlapsInfZone(rec, qID))
     return false;
   return true;
#else
   int count = 0;
   list<infObj*>::iterator iList;
   for( iList = queryTable[qID]->infList.begin() ; iList != queryTable[qID]->infList.end() ; iList++ )
   { 
     if(rec->prunedByBisector((*iList)->eq))
       count ++;
     if( count >= k )
       return true;	
   }
   return false;
#endif   
}

bool overlapsInfZone(CRectangle* R, int qID)
{
  
    CPoint* points[4];
    R->getCorners(points);
    
    
    
    // first check if a point of rectangle is within the 
    for(int i=0; i<4; i++)
    {
      if(!liesOutOfPolygon(points[i],qID))
      {
	deletePoints(points);
  	return true;
      }
    }
  
    
    //check if a point of InfZone is within the rectanlge
    vector<Vertex*>::iterator iP;
    for(iP = queryTable[qID]->vPolygon.begin(); iP != queryTable[qID]->vPolygon.end(); iP++)
    {
      if(R->contains((*iP)->loc))
      {
	deletePoints(points);
	return true;
      } 
    }

    
    // now check if any edge of rectangle intersects with any edge of polygon
    for(int r1=0; r1<4; r1++)
    {
      for(int r2 = r1+1; r2<4; r2++)
      {

	    // if the edge intersects with any edge of polygon, the rec cannot be pruned        
	if(edgeIntersectsInfZone(points[r1], points[r2], qID))
	{
	  deletePoints(points);
	  return true;
	}
      }
    }
    
    deletePoints(points);
    return false;

}


bool edgeIntersectsInfZone(CPoint* a1, CPoint* a2,int qID)
{
    vector<Vertex*>::iterator iP;
    vector<Vertex*>::iterator iP1;
    vector<Vertex*>::iterator temp;
    for(iP = queryTable[qID]->vPolygon.begin(); iP != queryTable[qID]->vPolygon.end(); iP++)
    {    
	temp = iP;
	temp++;
	if(temp == queryTable[qID]->vPolygon.end())
	{
	  iP1 = queryTable[qID]->vPolygon.begin();
	}
	else
	  iP1 = temp;
		
	if(edgesIntersect2(a1,a2,(*iP)->loc,(*iP1)->loc))
	{
	  
	  return true;
	}
    }
    return false;
}




bool lessThanKInRange(CPoint* p, double range , int qID )
{
   int in_range=0;
   CHeap heap; 
   Ftree->getRoot()->getRect()->blowHuge();
   heap.insertNode(0,Ftree->getRoot());

   CNode* curNode;
   CHeapNode* heapNode;
   int numChildren;
   CEntry* child; 
    
   while(heap.getHeapSize() != 0)
   {
     heapNode = heap.top();
     curNode = (CNode*)heapNode->getData(); 
     
     heap.pop();
     double dist = curNode->getRect()->getMinDist(p);
     
     if(dist >= range || curNode->m_id == qID*(-1) )
     {     
       continue;
     } 
     
     if( curNode->m_id < 0)
     { 
       //data entry i.e; mbr of object       
       in_range++;    
       
       if( in_range >= k )
       {	 
	 return false;
       }  
     }
     else{
       curNode->load();
       numChildren = curNode->getNumChild();
        
       if( !foundInBuffer(curNode->m_id) )
       {
	 VERIFY_BUFFERED_IO++;
	 FACILITY_TREE_IO++;	 
       }
       
       VERIFY_PHASE_IO++;
       for (int i = 0; i < numChildren; ++i){
	 child = curNode->getEntry(i);
	 heap.insertNode( child->getRect()->getMinDist(p) , (CNode*) child );	 
       }       
     }     
   }
   return true;
}

bool foundInBuffer(int ID)
{ 
   if( bufferList.find(ID) == bufferList.end() )
   {
     if( bufferList.size() == BUFFERSIZE )
     { 
       int eviction = rand()%BUFFERSIZE ; // if buffsize is 3, generate number 0 to 2
       set<int>::iterator it;
       it = bufferList.begin();
       // increment counter by eviction
       for( int i = 0 ; i < eviction ; i++ )
       {
	 it++;	
       }
       assert(it!=bufferList.end());
       bufferList.erase(it);
       bufferList.insert(ID);
     }
     else{
       assert(bufferList.size()<BUFFERSIZE);
       bufferList.insert(ID);
     }
     return false;    
   }
   return true;
}
    
bool sortOnDimensionalValue (CPoint* a, CPoint* b)
{
   if(a->coord[GLOBAL_DIMENSION] > b->coord[GLOBAL_DIMENSION])
     return false;
   else
     return true; 
}

bool sortVerticesOnAngle (Vertex* a, Vertex* b)
{
   if(a->angleFromQ > b->angleFromQ)
     return false;
   else
     return true;
}

bool edgeIntersectsPolygon2(CPoint* a1, CPoint* a2,int qID)
{
   list<CPoint*>::iterator iP;
   list<CPoint*>::iterator iP1;
   list<CPoint*>::iterator temp;
   for(iP = queryTable[qID]->convexChain.begin(); iP != queryTable[qID]->convexChain.end(); iP++)
   {    
     temp = iP;
     temp++;
     if(temp == queryTable[qID]->convexChain.end())
     {
       iP1 = queryTable[qID]->convexChain.begin();
     }
     else
       iP1 = temp;
		
     if(edgesIntersect2(a1,a2,(*iP),(*iP1)))
     {  
       return true;
     }
   }
   return false;
}

bool edgesIntersect(CPoint* a1, CPoint* a2, CPoint* b1, CPoint* b2)
{
   CEquation* eq2 = new CEquation((b1),(b2),true);
   CEquation* eq1 = new CEquation(a1,a2, true);
    
   CPoint* intersection = eq1->IntersectionPoint(eq2);
   if(intersectionExists(intersection, a1,a2))
   {
     delete eq2;
     delete eq1;
     delete intersection;
     return true;
   }
     
   delete eq2;
   delete eq1;
   delete intersection;
   return false;
}

bool edgesIntersect2(CPoint* a1, CPoint* a2, CPoint* b1, CPoint* b2)
{
   CEquation* eq2 = new CEquation((b1),(b2),true);
   CEquation* eq1 = new CEquation(a1,a2, true);
   CPoint* intersection = eq1->IntersectionPoint(eq2);
     
   if( intersectionExists2(intersection, a1,a2) && intersectionExists2(intersection, b1,b2) )
   {
     delete eq2;
     delete eq1;
     delete intersection;
     return true;
   }
    
   delete eq2;
   delete eq1;
   delete intersection;
   return false;
}

bool intersectionExists2(CPoint* c, CPoint* p1, CPoint* p2)
{
   if( !withinDataSpace(c) )
   {
     return false;
   }
    
   for( int i = 0 ; i < DIMENSION ; i++ )
   {
     if(p1->coord[i] < p2->coord[i])
     {
       if(c->coord[i] >= p2->coord[i] || fabs(c->coord[i] - p2->coord[i]) < EPSILON2 )
       {
	 return false;	 
       }
       if(c->coord[i] <= p1->coord[i] || fabs(c->coord[i] - p1->coord[i]) < EPSILON2 )
       {
	 return false;	 
       }
     }
     else
     {
       // p1 is larger
       if(c->coord[i] >= p1->coord[i] || fabs(c->coord[i] - p1->coord[i]) < EPSILON2 )
       {
	 return false;	 	 
       }
       if(c->coord[i] <= p2->coord[i] || fabs(c->coord[i] - p2->coord[i]) < EPSILON2 )
       {
	 return false;	 
       }       
     }     
   } 
   return true;
}


bool intersectionExists(CPoint* c, CPoint* p1, CPoint* p2)
{
   if(!withinDataSpace(c))
     return false;
    
   for(int i=0; i<DIMENSION; i++)
   {
     if(p1->coord[i] < p2->coord[i])
     {
       if(c->coord[i] >  p2->coord[i])
       {
	 return false;	 
       } 
       if(c->coord[i] < p1->coord[i])
       {
	 return false;	 
       }
     }
     else
     {
       // p1 is larger
       if(c->coord[i] > p1->coord[i])
       {
	 return false;	 
       }
       
       if(c->coord[i] < p2->coord[i] )
       {
	 return false;	 
       }       
     }     
   }
    
   if(c->getDist(p1)<EPSILON)
   {
     return false;
   }
   if(c->getDist(p2)<EPSILON)
   {
     return false;
   }
   return true;
}


bool pointIsInSR(CPoint* p, int qID)
{
   if(liesOutOfSR(p,qID))
   {
     return false;
   }
   else
   {
     return true;
   } 
}


bool prunedBySR(CRectangle* R, int qID)
{
   if( algoNum == FINCHBRNN_OPT || algoNum == FINCHMONO_OPT)
   {
     if(R->getMinDist(queryTable[qID]->loc) > queryTable[qID]->maxRadius)
       return true;

     if(!R->isIntersected(queryTable[qID]->approxRegion))
       return true;
   }

   
   
   CPoint* points[4];
   R->getCorners(points);
 
    // first check if a point of rectangle is within the SR
   for(int i=0; i<4; i++)
   {
     if(pointIsInSR(points[i],qID))
     {
       deletePoints(points);
       return false;
     }
   }
  
    
      //check if a point of polygon is within the rectanlge
   list<CPoint*>::iterator iP;
   for(iP = queryTable[qID]->convexChain.begin(); iP != queryTable[qID]->convexChain.end(); iP++)
   {
     if(R->contains((*iP)))
     {
       deletePoints(points);		
       return false;
     } 
   } 
    
    // now check if any edge of rectangle intersects with any edge of polygon
   for(int r1=0; r1<4; r1++)
   {
     for(int r2 = r1+1; r2<4; r2++)
     {
       // if the edge intersects with any edge of polygon, the rec cannot be pruned        
       if(edgeIntersectsPolygon2(points[r1], points[r2], qID))
       {
	 deletePoints(points);
	 return false;	 
       }       
     }     
   }

   deletePoints(points);
   return true;
}



void deletePoints(CPoint** points)
{
   for(int i=0; i<4; i++)
   {
     delete points[i];
   }
}

double crossProduct(stack<CPoint*> *mystack, CPoint* p3)
{
   CPoint* p2 = mystack->top();
   mystack->pop();
    
   CPoint* p1 = mystack->top();
   mystack->push(p2);
        
   double val = (p2->coord[0] - p1->coord[0])*(p3->coord[1] - p1->coord[1]) - (p3->coord[0] - p1->coord[0])*(p2->coord[1] - p1->coord[1]);
    
   return val;
}

void shortlistPoints(int qID, list<CPoint*> *pList)
{
   list<Vertex*>::iterator iList;
   list<CPoint*> boundaryPoints;
    
   for(iList = queryTable[qID]->vList.begin(); iList!= queryTable[qID]->vList.end(); iList++)
   {     
     if((*iList)->counter == k-1 )//|| (*iList)->counter == k-2 )
     {
       pList->push_back((*iList)->loc);
     }
     else
     {
       assert((*iList)->counter<k);
       if(liesAtDataSpaceBoundary((*iList)->loc))
       {
	 boundaryPoints.push_back((*iList)->loc);	  
       }       
     }	         
   }

   keepExtremeBoundaryPoints(&boundaryPoints);
   list<CPoint*>::iterator iP;
    
   for(iP= boundaryPoints.begin(); iP!=boundaryPoints.end(); iP++)
   {
     pList->push_back((*iP));
   }
}

void printArray(extremes* *pArray)
{
   for(int i=0; i<4; i++)
   {
     cout << "index: " << i << " min: " << pArray[i]->min <<  " max: " << pArray[i]->max << endl;
   }
}

void keepExtremeBoundaryPoints(list<CPoint*> *boundaryPoints)
{
   extremes* pArray[4];
   for(int i=0; i<4; i++)
   {
     pArray[i] = new extremes;
   }
   // 0 lower horizontal line, 1 upper horizontal line, 2 left vertical line, 3 right vertical line
  
   list<CPoint*>::iterator iP;
   list<CPoint*>::iterator tmpiP;
   for(iP= boundaryPoints->begin(); iP!=boundaryPoints->end(); iP++)
   {
      // lower horizontal (e.g., 0)     
      if((*iP)->coord[1] == 0)
      {
        updateExtremes(pArray,(*iP)->coord[0],0);
      }
      
      // upper horizontal : 1
      if((*iP)->coord[1] == 1)
      {
        updateExtremes(pArray,(*iP)->coord[0],1);
      }
      
      // left vertical : 2
      if((*iP)->coord[0] == 0)
      {
        updateExtremes(pArray,(*iP)->coord[1],2);
      }
      
      // right vertical : 3
      if((*iP)->coord[0] == 1)
      {
        updateExtremes(pArray,(*iP)->coord[1],3);
      }
      
   }
   
   for(iP= boundaryPoints->begin(); iP!=boundaryPoints->end(); )
   {
     tmpiP= iP;
     tmpiP++;
     
     bool deleteIt = true;
     
     if((*iP)->coord[1] == 0)
     {
       if((*iP)->coord[0] == pArray[0]->min || (*iP)->coord[0] == pArray[0]->max )
       {
	 deleteIt = false;	 
       }
     }
      
      // upper horizontal : 1
      if((*iP)->coord[1] == 1)
      {
	if((*iP)->coord[0] == pArray[1]->min || (*iP)->coord[0] == pArray[1]->max )
	{
	  deleteIt = false;	  
	}
      }
      
      // left vertical : 2
      if((*iP)->coord[0] == 0)
      {
	if((*iP)->coord[1] == pArray[2]->min || (*iP)->coord[1] == pArray[2]->max )
	{
	  deleteIt = false;	  
	}
      }
      
      // right vertical : 3
      if((*iP)->coord[0] == 1)
      {
	if((*iP)->coord[1] == pArray[3]->min || (*iP)->coord[1] == pArray[3]->max )
	{
	  deleteIt = false;
        }
      }
      
      if(deleteIt)
      {
	iP=boundaryPoints->erase(iP);
      }
      iP = tmpiP;
   }
  

   for(int i=0; i<4; i++)
   {
     delete pArray[i];     
   }  
}

void printPoints(list<CPoint*> *boundaryPoints)
{
   list<CPoint*>::iterator iP;
  
   cout << "Global Point : ";
   GLOBAL_POINT->print();
   cout << endl;
  
   int count=1;
   for(iP= boundaryPoints->begin(); iP!=boundaryPoints->end(); iP++)
   {
     cout << "Point " << count << " :  ";
     (*iP)->print();
     cout << " angle: " << GLOBAL_POINT->getAngle((*iP))  << " distance: " << GLOBAL_POINT->getDist((*iP));
     cout << endl;
     count++;
  }
}

void updateExtremes(extremes** pArray, double val, int index)
{
   if(pArray[index]->min > val)
   {
     pArray[index]->min = val;
   }
   if(pArray[index]->max < val)
   {
     pArray[index]->max = val;     
   }
}

bool liesAtDataSpaceBoundary(CPoint* p)
{
   for(int i=0; i<DIMENSION; i++)
   {
     if(fabs(p->coord[i] - 0)<EPSILON)
     {
       p->coord[i] = 0;
       return true;       
     }
     if(fabs(p->coord[i]-1)<EPSILON)
     {
       p->coord[i] = 1;
       return true;
     }
   }
   return false;
}

double computeAreaBySampling(int qID)
{
   srand(time(NULL));
   CPoint* p = new CPoint();
   double sampleSize = 1000000;
  
   double inside = 0;
   for(int i=0; i<sampleSize; i++)
   {
     for(int dim=0; dim<DIMENSION; dim++)
     {
       p->coord[dim] = (double)rand()/RAND_MAX;
     }
     
     if(!outsideInfZone(p,qID))
     {
       inside ++;
     }       
   }
   double ratio = inside/sampleSize;
   return ratio;
}

void computeNumOfBisectors(int qID)
{
   set<int> bisecs;  
   getBisecsOfInfZone(qID,&bisecs);
   set<int>::iterator it;
   int i=1;
   for(it = bisecs.begin(); it!=bisecs.end(); it++)
   {
     cout << i << ".   " << (*it) <<endl;
     i++;
   }
   cout << "total bisectors " << bisecs.size() << endl;
  
}

void computeAreaOfInfZone(int qID)
{
   double totArea = 0;
   set<int>bisecs;
   getBisecsOfInfZone(qID,&bisecs);
  
   set<int>::iterator iBisec;
  
   for(iBisec = bisecs.begin(); iBisec!=bisecs.end(); iBisec++)
   {
     list<Vertex*>::iterator iList;
    
     if((*iBisec)<0)
       continue;
    
     list<CPoint*> pointList;
     pointList.clear();
     for(iList = queryTable[qID]->vList.begin(); iList!= queryTable[qID]->vList.end(); iList++)
     {
       if((*iList)->counter>=k-2)
       {
	 if((*iList)->objID1 == (*iBisec) || (*iList)->objID2 == (*iBisec))
	 {
	   pointList.push_back((*iList)->loc);
	   assert(fabs(queryTable[qID]->loc->getDist((*iList)->loc)-facilityTable[(*iList)->objID1]->getDist((*iList)->loc))<EPSILON);
         }        
       } 
     }
     double area = getAreaOfEdges(qID,(*iBisec),&pointList);
     totArea += area;   
   }
   cout << "exact InfZone Area: " << totArea << endl;  
}

void printPList(list<CPoint*> *pList)
{
   list<CPoint*>::iterator iP;

   int count = 0;
   for(iP = pList->begin(); iP!=pList->end(); iP++)
   {
     cout << count << "  :  ";
     (*iP)->print();
     cout << endl;
     count++;
   }
}
   

double getAreaOfEdges(int qID, int objID, list<CPoint*> *pList)
{
   double totArea = 0;
   GLOBAL_DIMENSION = 0;
   pList->sort(sortOnDimensionalValue);
  
   if((*pList->begin())->coord[0]== ((*(--pList->end()))->coord[0]))
   {
     GLOBAL_DIMENSION = 1;
     pList->sort(sortOnDimensionalValue);
   }

   list<CPoint*>::iterator iP;
   list<CPoint*>::iterator iP1;
   list<CPoint*>::iterator temp;
   for(iP = pList->begin(); iP!=pList->end(); iP++)
   {
     temp = iP;
     temp++;
     iP1 = temp;
     if (iP1== pList->end())
     {
      break;       
     }
      
     double area = getAreaOfOneEdge(qID,objID, (*iP), (*iP1));
     totArea += area;     
   }
  
   return totArea;
}

double getAreaOfOneEdge(int qID, int objID, CPoint* p1, CPoint* p2)
{
   assert(fabs(queryTable[qID]->loc->getDist(p1)-facilityTable[objID]->getDist(p1))<EPSILON);
   assert(fabs(queryTable[qID]->loc->getDist(p2)-facilityTable[objID]->getDist(p2))<EPSILON);
   CPoint* cent = p1->getCenter(p2);
   assert(fabs(queryTable[qID]->loc->getDist(cent)-facilityTable[objID]->getDist(cent))<EPSILON);
   int counter = computeCounter(cent,qID);
 
   delete cent;
   if(counter<k-2 || counter>=k)
   {
     return 0;  
   }
   else
   {
     return getTriangleArea(p1,p2,queryTable[qID]->loc);
   }
}

double getTriangleArea(CPoint* p1, CPoint* p2, CPoint* p3)
{
// using Heron's formula  (e.g., let a, b and c be side lengths)
// s be semiperimieter; s = (a+b+c)/2
// area = sqrt(s*(s-a)(s-b)(s-c))
   
   double a = sqrt(p1->getDist(p2));
   double b = sqrt(p2->getDist(p3));
   double c = sqrt(p3->getDist(p1));
   double s = (a+b+c)/2;
   double area = sqrt(s*(s-a)*(s-b)*(s-c));
   return area;
}

void getBisecsOfInfZone(int qID,set<int> *bisecs)
{
   list<Vertex*>::iterator iList;
  
   for(iList = queryTable[qID]->vList.begin(); iList!= queryTable[qID]->vList.end(); iList++)
   {
     if((*iList)->counter>=k-2)
     {
       bisecs->insert((*iList)->objID1);
       bisecs->insert((*iList)->objID2);
     } 
   }
}

void estimateInfZoneArea(int qID)
{ 
   double totArea = 0;
   double count = 0;
   double maxArea = 0;
   list<Vertex*>::iterator iList;
   for(iList = queryTable[qID]->vList.begin(); iList!= queryTable[qID]->vList.end(); iList++)
   { 
     if((*iList)->counter>=k-2)
     {
       double area = PI*(*iList)->distFromQ;
       if(area>maxArea)
	 maxArea = area;
       totArea += area;
       count ++;
     }        
   }
  
   cout << "Mean Area " << totArea/count << endl;
   cout << "Max Circular Area " << maxArea << endl;
   cout << "Theoretical Area " << (double)k/max_Facilities << endl;

}



void computeInfluenceZone(int qID)
{
   initializeQuery(qID);
   CHeap heap;
   Ftree->getRoot()->getRect()->blowHuge();
   heap.insertNode(0,Ftree->getRoot());

   CNode* curNode;
   CHeapNode* heapNode;
   int numChildren;
   CEntry* child;
  

   while(heap.getHeapSize() != 0)
   {
     heapNode = heap.top();
     curNode = (CNode*)heapNode->getData(); 
    
     heap.pop();
     
      if(curNode->m_id < 0)
      {
	EntryPrune_Point_Facility++;
      }else{
	EntryPrune_Node_Facility++;
      }
      
     if(prunedByCircles(curNode->getRect(),qID))
     {
       continue;
     }
     if (curNode->m_id < 0){ //data entry i.e; mbr of object
       CPoint* x = new CPoint(curNode->getRect()->m_low);
       if(!x->isSame(queryTable[qID]->loc))
       {
	 num_infObj ++;
	 addInfObj(curNode->m_id*-1, x, qID);     
       }
       delete x;
     }
     else{    
       curNode->load(); // Make sure load()
       numChildren = curNode->getNumChild();
       PRUNE_PHASE_IO++;
      
       if(!cmpIOBound)
         FACILITY_TREE_IO++;

       for (int i = 0; i < numChildren; ++i){
	 child = curNode->getEntry(i);
	 double myKey=child->getRect()->getMinDist(queryTable[qID]->loc);
	 
#ifdef InfZone_opt2
	 if(child->m_id < 0)
	   myKey = -100 + myKey;
#endif
        heap.insertNode(myKey,(CNode*) child);
      }
    }
  }
}

bool prunedByCircles(CRectangle* rec,int qID)
{
  
  // if it lies 2*R_{max} then it is definitely pruned
   double minD = rec->getMinDist(queryTable[qID]->loc);

  // since distances are in square, we need to double the distance
   if(minD > 4*queryTable[qID]->maxRadius)
   {
     return true;
   }
  
   if(minD < 4*queryTable[qID]->minRadius)
   {
     return false;
   }
 
   list<CPoint*>::iterator iList;
   for(iList = queryTable[qID]->vList_Special.begin(); iList!= queryTable[qID]->vList_Special.end(); iList++)
   {
     if(rec->getMinDist((*iList)) <= (*iList)->getDist(queryTable[qID]->loc))
     {
       return false;
     }
   }
   return true;
}

bool FINCH_prunedByCircles(CRectangle* rec, int qID)
{
   vector<Vertex*>::iterator iList;

   for(iList = queryTable[qID]->vPolygon.begin(); iList!= queryTable[qID]->vPolygon.end(); iList++)
   {
     double recDist = rec->getMinDist((*iList)->loc);
     double qDist = (*iList)->loc->getDist(queryTable[qID]->loc);
     if( recDist <= qDist || fabs(recDist-qDist)<EPSILON )
     {     
       return false;
     }
   }
  
   return true; 
}

void initializeQuery(int qID)
{
  // Left = 1; Up = 2; right = 3; down = 4;
  //  x coordinate is -x and y is same
   INITIALIZING = true;
  
   queryTable[qID]->approxRegion = new CRectangle(queryTable[qID]->loc->coord);

   queryTable[qID]->approxRegion->blowHuge();
  
  
   CPoint* x = new CPoint(queryTable[qID]->loc->coord[0]*-1,queryTable[qID]->loc->coord[1]);
   addInfObj(-1,x,qID);
  
   delete x;
   // up:  x is same, y is (1+(1-y))
   x = new CPoint(queryTable[qID]->loc->coord[0],2-queryTable[qID]->loc->coord[1]);
   addInfObj(-2,x,qID);
  
   delete x;
  // right:  y is same, x is 2-x
   x = new CPoint(2-queryTable[qID]->loc->coord[0],queryTable[qID]->loc->coord[1]);
   addInfObj(-3,x,qID);
  
   delete x;
   // down:  x is same, y is -y
   x = new CPoint(queryTable[qID]->loc->coord[0],queryTable[qID]->loc->coord[1]*-1);
   addInfObj(-4,x,qID);

   delete x;
  
   INITIALIZING = false;
  
}

void printInfObj(int qID)
{
   cout << "InfList of query: " << qID << endl;
   list<infObj*>::iterator iList;
   int i = 1;
   for(iList = queryTable[qID]->infList.begin(); iList!= queryTable[qID]->infList.end(); iList++)
   {
     cout << i << ".  objID: " << (*iList)->objID << " ";
     (*iList)->loc->print();
     if((*iList)->objID>0)
     {
       assert(facilityTable[(*iList)->objID]->isSame((*iList)->loc));
     }
     cout << "  numVertices: " << (*iList)->numVertices;

     cout<< endl;
     i++;    
  }
}

void printVertices(int qID)
{
   int count = 0;
   list<Vertex*>::iterator iList;
   int i = 1;

   for(iList = queryTable[qID]->vList.begin(); iList!= queryTable[qID]->vList.end(); iList++)
   {
     cout << "v " << i << " : ";
     (*iList)->loc->print();
     cout << "  counter: " << (*iList)->counter << "  obj1: " << (*iList)->obj1->objID << " obj2: " << (*iList)->obj2->objID <<  " distFromQ: " << (*iList)->distFromQ << endl;
    
     if((*iList)->counter>=k-2)
     { 
       count++;
     }
     i++;   
  }
  
  cout << "Number of vertices on polygon: " << count << endl;
}

void addInfObj(int objID, CPoint* point, int qID)
{
   bool affectsInfZone = false;
   list<infObj*>::iterator iList;
   infObj* x = new infObj(); 
   x->objID = objID;
   x->loc = new CPoint(point->coord);
   x->eq = new CEquation(queryTable[qID]->loc, x->loc);
   x->distFromQ = queryTable[qID]->loc->getDist(x->loc);
   x->cent = queryTable[qID]->loc->getCenter(x->loc);
   x->checkCenter = true;
   x->numVertices = 0;
  
  
   // compute intersection with each existing influence object
   for(iList = queryTable[qID]->infList.begin(); iList!= queryTable[qID]->infList.end(); iList++)
   {
     bool counterNeeded = false;
     CPoint* intersection = x->eq->IntersectionPoint((*iList)->eq);
     if( algoNum == FINCHBRNN_OPT || algoNum == FINCHMONO_OPT)
     {
       if( algoNum == FINCHBRNN_OPT || algoNum == FINCHMONO_OPT)
       {
	 if(withinDataSpace(intersection)  && intersection->getDist(queryTable[qID]->loc) < queryTable[qID]->maxRadius)
	 {
	   counterNeeded = true;
	 }
       }
       else
       {
	 if(withinDataSpace(intersection))
	 {
	   counterNeeded = true;
	 }	
       }
    }
    if(algoNum == InfZoneBRNN || algoNum == InfZoneMONO || BRUTEFORCE == true)
    {
      if(withinDataSpace(intersection) && intersection->getDist(queryTable[qID]->loc) < queryTable[qID]->maxRadius)
      {
	counterNeeded = true;
      } 
    }
    
    if(counterNeeded)
    {
      assert(fabs(intersection->getDist(queryTable[qID]->loc) -intersection->getDist(x->loc)) < EPSILON);
      assert(fabs(intersection->getDist(queryTable[qID]->loc) -  intersection->getDist((*iList)->loc) < EPSILON));     
      int counter = computeCounter(intersection, qID);
       
       // if pruned by less than k, then add it
      if(counter < k )
      {
	Vertex* v = new Vertex();
	v->loc = intersection;
	v->counter = counter;
	v->obj1 = x;
	v->obj2 = (*iList);
        v->objID1 = x->objID;
        v->objID2 = (*iList)->objID;
        v->distFromQ = intersection->getDist(queryTable[qID]->loc);
        queryTable[qID]->vList.push_back(v);
         
	x->numVertices ++;
        (*iList)->numVertices++;
        affectsInfZone = true;
     }
     else
     {
       delete intersection;
     }      
    }
    else
    {
      delete intersection;
    }
  
  }
 
  queryTable[qID]->infList.push_back(x);
  updateCountersAndRemoveIntersections(point,qID);
}

void updateCountersAndRemoveIntersections(CPoint* point, int qID)
{ 
   double minRadius = 10;
   double maxRadius = 0;
   list<CPoint*> boundaryPoints;
   queryTable[qID]->vList_Special.clear();    
   int validV = 0;
  
   list<Vertex*>::iterator iList;
   list<Vertex*>::iterator tmpList;
   for(iList = queryTable[qID]->vList.begin(); iList!= queryTable[qID]->vList.end();)
   {
     tmpList = iList;
     tmpList++;
     if((*iList)->loc->pliesInSpace(queryTable[qID]->loc,point))
     {
       (*iList)->counter ++;
     } 
    
     if((*iList)->counter >= k)
     {
       (*iList)->obj1->numVertices --;
       (*iList)->obj2->numVertices --;
       assert((*iList)->objID1 != (*iList)->objID2);
       assert((*iList)->objID1!=0);
       assert((*iList)->objID2!=0);
      
       deleteVertex(*iList);
       iList=queryTable[qID]->vList.erase(iList);      
     }
     else
     { 
       if(INFZONE_ALGO)
       {
	 if((*iList)->counter >= k-2 )
	 {
	   if((*iList)->counter== k-1)
	   {
	     queryTable[qID]->vList_Special.push_back((*iList)->loc);	     
	   }
           validV++;
           if((*iList)->distFromQ < minRadius )
           {
	     minRadius = (*iList)->distFromQ;
           }
          
           if((*iList)->distFromQ > maxRadius )
           {
	     maxRadius = (*iList)->distFromQ;
           }
         }
         if(liesAtDataSpaceBoundary((*iList)->loc))
         {
	   boundaryPoints.push_back((*iList)->loc);
              
           if((*iList)->distFromQ > maxRadius )
           {
	     maxRadius = (*iList)->distFromQ;
           }
         }
       }          
    }
    iList=tmpList;
   }
  
   if(INFZONE_ALGO)
   {
     if(validV==0 || INITIALIZING)
     {
       queryTable[qID]->maxRadius = 10;
       queryTable[qID]->minRadius = 0;      
     }
     else
     { 
       assert(queryTable[qID]->maxRadius>=maxRadius);
       queryTable[qID]->maxRadius = maxRadius;
       queryTable[qID]->minRadius = 0;
     }
    
     keepExtremeBoundaryPoints(&boundaryPoints);  
     list<CPoint*>::iterator iP;
     for(iP= boundaryPoints.begin(); iP!=boundaryPoints.end(); iP++)
     {
       queryTable[qID]->vList_Special.push_back((*iP));
     }
   }
}


void deleteVertex(Vertex* v)
{
   delete v->loc;
   delete v;
}


int computeCounter(CPoint* p, int qID)
{  
   int count = 0;
   list<infObj*>::iterator iList;
   for(iList = queryTable[qID]->infList.begin(); iList!= queryTable[qID]->infList.end(); iList++)
   {
     if(p->pliesInSpace(queryTable[qID]->loc,(*iList)->loc))
     {
       count ++;
     }
   }
   return count;
}

bool withinDataSpace(CPoint* p)
{
   for(int i=0; i<DIMENSION; i++)
   {    
     if(p->coord[i] <0 || p->coord[i] > 1)
     {
       return false;
     }
   }
   return true;
}

void printQuery(int i)
{
   cout << "QID: " << i ;
   queryTable[i]->loc->print();
   cout << endl;
}

void loadQueries()
{
   CStack< CNode* > stk;
   CNode* node = Ftree->getRoot();
   CIndexNode * index = NULL;
   CLeafNode * leaf = NULL;
   CNode * child = NULL;
   stk.push( node );

   while (!stk.empty())
   {
     node = stk.top();
     stk.pop();
     if ( node->isLeaf() )
     {
       leaf = ( CLeafNode* ) node;
       for ( int i=0; i < leaf->getNumChild() ; i++ )
       {
         CNode *pEnt = (CNode*) leaf->getEntry( i ); 
         facilityTable[pEnt->m_id*-1] = new CPoint(pEnt->m_mbr->m_low);
       }
     }
     else
     {
       index = ( CIndexNode* ) node ;
       for ( int i=0; i< index->getNumChild(); i ++ )
       {
         CNode *pEnt = (CNode*) index->getEntry( i );
         pEnt->load(); 
         stk.push( pEnt );        
       }
     }
   }  
  
  if(strcmp( fDist , "out") == 0)
  {
     ifstream inputquery ;
     int id ;
     inputquery.open("./query");
     
     if( ! inputquery)
     {
         cout << "Can't open query " << endl;  
	 exit(0);
     }
  
     for(int i=1; i<=numQ; i++)
     {
	queryTable[i] = new Query();
        
	double coord [2] ;
	double x , y;
        inputquery >> id >> x >> y ;
	
	coord[0]=x;
	coord[1]=y;
	queryTable[i]->loc = new CPoint(coord);       
     }      
  }
  else
  if(strcmp( fDist, "onlap" ) == 0)
  {
     int m = 1;
     int i = 1;
     while(  m <= numQ )
     {
       if( facilityTable[i]->coord[0] > 0.2 && facilityTable[i]->coord[0]<0.3 
	 && facilityTable[i]->coord[1]>0.2 && facilityTable[i]->coord[1]<0.3 )
       {
	  queryTable[m]= new Query();
	  queryTable[m]->loc = new CPoint(facilityTable[i]->coord);
	  m++;
	  i++;
       }
       else{
	 i++;
       }
     }   
  }
  else
  {
   
      for(int i=1; i<=numQ; i++)
      {	
	queryTable[i] = new Query();
	queryTable[i]->loc = new CPoint(facilityTable[i]->coord);
      }
   }
  
}

float RandomEqual(float min, float max)
{
	float x = (float)rand()/RAND_MAX;
	return x*(max-min)+min;
}



int getCounterOfRec(CRectangle* rec, int qID)
{
   list<infObj*>::iterator iList;
   int count = 0;
   
   double area = rec->getArea();
   bool isPoint = false;
   if(area == 0 )
   {
     isPoint = true;
   }
   
   
   
   list<infObj*> tmpList;  
   for(iList = queryTable[qID]->infList.begin(); iList!= queryTable[qID]->infList.end(); iList++)
   {     
     if( isPoint)  // rec is point
     {
       if(rec->getMinDist( (*iList)->loc) < rec->getMinDist( queryTable[qID]->loc))
       {    
	 count++;	 
       }
       else
       {
	 tmpList.push_back((*iList));   	 
       }
     }
     else
     {                //rec is rectangle
       if( rec->prunedByBisector((*iList)->eq) )
       {   
	 count++;	 
       }
       else
       {
	 tmpList.push_back((*iList));   	 
       }
     }	
     if(count>=k)
     {
       return count;
     }     
   }
   
   if(isPoint)
     return count;
    

    
    // now see how many can be pruned by trimming
   CRectangle* tmpRec1 =  new CRectangle();
   CRectangle* tmpRec2 =  new CRectangle();
   tmpRec1->clone(rec);
   tmpRec2->clone(rec);
     
   for(iList = tmpList.begin(); iList!= tmpList.end(); iList++)
   {    	
     
     if( tmpRec1->bisectorTrim((*iList)->eq,tmpRec2) )
     {	  
       count ++;     
       // the temp rec is initialized 
       tmpRec1->clone(rec);	  	   	          
     }
     else
     {
       // update it to clipped rectangle
       tmpRec1->clone(tmpRec2);   	    
     }
     if(count>=k)
     {
       delete tmpRec1;
       delete tmpRec2;
       return count;   
     }
   }
   delete tmpRec1;
   delete tmpRec2;
   return count;
}



void FINCH_MONO(int qID)
{ 
  
   if(qID%100==0)
     cout <<"query num " << qID << endl; 
 
   k=k+1;
  
   p1_util->startCT();
   FINCH_computeSR(qID);
   p1_time += p1_util->endCT(false);
  
   p2_util->startCT();
   FINCH_MONO_refinement(qID);
   p2_time += p2_util->endCT(false);
  
   k=k-1;

   resetQuery(qID);
}


void FINCH_computeSR(int qID)
{
  // initialize the influence objects with data space boundary
   initializeQuery(qID);

   CHeap heap;
   Ftree->getRoot()->getRect()->blowHuge();
   heap.insertNode(0,Ftree->getRoot());

   CNode* curNode;
   CHeapNode* heapNode;
   int numChildren;
   CEntry* child;

   int checked=0;
   while(heap.getHeapSize() != 0)
   {
     heapNode = heap.top();
     curNode = (CNode*)heapNode->getData(); 

     heap.pop();
     
     if(curNode->m_id < 0)
     {
       EntryPrune_Point_Facility++;
    }else{
      EntryPrune_Node_Facility++;
    }
    

     if(curNode->m_id>=0)
     {
       if(prunedBySR(curNode->getRect(),qID))
       {
	    continue; 
       }
     }

     if (curNode->m_id < 0){ //data entry i.e; mbr of object
       CPoint* x = new CPoint(curNode->getRect()->m_low);
       if(!x->isSame(queryTable[qID]->loc))
       {
	 if(pointIsInSR(x,qID))
	 {
           num_infObj++;
	   addInfObj(curNode->m_id*-1, x, qID);
	   FINCH_computeConvexHull(qID);
	 }
       }
       delete x;
    }
    else{	
      curNode->load();
      numChildren = curNode->getNumChild();     

      if(!foundInBuffer(curNode->m_id))
      {
	PRUNE_PHASE_IO++;
	FACILITY_TREE_IO++;	
      }    
      
      for (int i = 0; i < numChildren; ++i){      
	child = curNode->getEntry(i);      
	if(child->m_id >= 0)
	{
	  if(prunedBySR(child->getRect(),qID))
	  {
	    continue; 
	  }
	  
	}
	else
	{
	  CPoint* x = new CPoint(child->getRect()->m_low);
          if(!pointIsInSR(x,qID))
	  {
	    delete x; 
	    continue;
	  }
	  delete x; 
	}
        heap.insertNode(child->getRect()->getMinDist(queryTable[qID]->loc),(CNode*) child);
      
	
	
	
      }      
    }
  } 
}

void  FINCH_MONO_refinement(int qID)
{
   verify_util->startCT(); 
   if(algoNum == FINCHBRNN_OPT || algoNum == FINCHMONO_OPT)
   { 
     
     list<infObj*>::iterator iList;
     for(iList = queryTable[qID]->infList.begin(); iList!= queryTable[qID]->infList.end(); iList++)
     {
       if((*iList)->objID<0)
	 continue;

       if(pointIsInSR((*iList)->loc,qID))
       {
	 if( lessThanKInRange( (*iList)->loc, queryTable[qID]->loc->getDist((*iList)->loc) , qID)  )
	 {
	   queryTable[qID]->RkNN.insert((*iList)->objID);
	 }
       }
     }     
    
   }

   verify_time += verify_util->endCT(false);
}


void FINCH_computeConvexHull(int qID)
{
  // list of points on which we will compute the convex hull
   list<CPoint*> pList;
   list<CPoint*>::iterator iP;
   list<CPoint*>::iterator minP;
  
   shortlistPoints(qID,&pList);

  // assert(pList.size()>2);
  // http://www.bookrags.com/wiki/Graham_scan
   CPoint* minPoint = new CPoint(10,10);
  
   int minIndex =0;

   minP = pList.begin();
   for(iP = pList.begin(); iP !=  pList.end(); iP++)
   {  
     if((*iP)->coord[1] <= minPoint->coord[1] )
     {
       // if y values are same && x value of new point is smaller than update minPoint
       if( (*iP)->coord[1] == minPoint->coord[1] &&  (*iP)->coord[0] > minPoint->coord[0] )
       {
    
       }
       else
       {
	 minPoint->coord[0] = (*iP)->coord[0];
         minPoint->coord[1] = (*iP)->coord[1];
         minP = iP;
       }
     }
   }
    
   GLOBAL_POINT->coord[0]=minPoint->coord[0];
   GLOBAL_POINT->coord[1]=minPoint->coord[1];

   pList.sort(sortOnPolarAngle);
  
   stack<CPoint*> mystack;
   iP = pList.begin();
  
   mystack.push((*iP));
   iP++;
  
   mystack.push((*iP));
   iP++;
   
   for(iP=iP;iP!=pList.end();iP++)
   {
     while(mystack.size()>1 && crossProduct(&mystack, (*iP)) <= 0)
     {    
       mystack.pop();
     }
     mystack.push((*iP));      
  }
  
  
   queryTable[qID]->convexChain.clear();
   emptyVPolygon(qID);

  // the points will be sorted from their angle to query
   GLOBAL_POINT->coord[0] = queryTable[qID]->loc->coord[0];
   GLOBAL_POINT->coord[1] = queryTable[qID]->loc->coord[1];

   if(algoNum == FINCHBRNN_OPT || algoNum == FINCHMONO_OPT)
   {
     queryTable[qID]->maxRadius= -1;
     queryTable[qID]->approxRegion->setEmpty();  
   }
  
   while(!mystack.empty())
   {
     Vertex* v = new Vertex();
     v->loc = mystack.top();
     v->angleFromQ = queryTable[qID]->loc->getAngle(v->loc);
     queryTable[qID]->vPolygon.push_back(v);
     queryTable[qID]->convexChain.push_back(mystack.top()); 
     
     if( algoNum == FINCHBRNN_OPT || algoNum == FINCHMONO_OPT )
     {
       double tmp_dist =  queryTable[qID]->loc->getDist(v->loc);
       if(tmp_dist > queryTable[qID]->maxRadius)
       {
	 queryTable[qID]->maxRadius = tmp_dist;
       }
       queryTable[qID]->approxRegion->combine(v->loc);
    }
    mystack.pop();
  }

   std::sort(queryTable[qID]->vPolygon.begin(), queryTable[qID]->vPolygon.end(), sortVerticesOnAngle); 

  delete minPoint;

}

void emptyVPolygon(int qID)
{
   vector<Vertex*>::iterator iList;
   for(iList = queryTable[qID]->vPolygon.begin(); iList!= queryTable[qID]->vPolygon.end(); iList++)
   {
     delete (*iList); 
   }
   queryTable[qID]->vPolygon.clear();
}


void FINCH_BRNN(int qID)
{	
  
  if(qID%100==0)
    cout <<"query num " << qID << endl; 

   p1_util->startCT();
   FINCH_computeSR(qID);
   p1_time += p1_util->endCT(false);
  
  
  
   p2_util->startCT();
   FINCH_BRNN_refinement(qID);
   p2_time += p2_util->endCT(false);

   resetQuery(qID);
}



void FINCH_BRNN_refinement(int qID)
{
   CStack< CNode* > stk;
   long count = 0;

   Utree->getRoot()->getRect()->blowHuge();
  
   stk.push(Utree->getRoot());
   CNode* node;

   CIndexNode* index = NULL;
   CLeafNode* leaf = NULL;
   CNode* child = NULL;

   while (!stk.empty())
   {
     node = stk.top();
     stk.pop();
     if(node->m_id >= 0)
     {  
       EntryPrune_Node_User++;
       if(prunedBySR(node->getRect(),qID))
       {
	 continue;
       }
    
       node->load();
       REFINE_PHASE_IO++;
       USER_TREE_IO++;
       for (int i = 0; i < node->getNumChild(); i++)
       {
         child = (CNode*) node->getEntry(i);
	 stk.push(child);
       }
    }
    else
    { // it is a data point
        EntryPrune_Point_User++;
        CPoint* p = new CPoint(node->getRect()->m_low);
	
	if(!pointIsInSR(p,qID))
	{
	  delete p ;
	  continue;
	}
        UCandidates++;
	if(algoNum == FINCHBRNN_OPT || algoNum == FINCHMONO_OPT)
	{
	  verify_util->startCT();
	  if( lessThanKInRange( p, p->getDist(queryTable[qID]->loc), qID )  )
	  {
	    queryTable[qID]->RkNN.insert(node->m_id*-1);
	  }
	  delete p;
	  verify_time += verify_util->endCT(false);	
	}
        else
	{	
	    MD_addCands(node->m_id*(-1), p , qID );  
	}  
    }
  }

}
