#include "node.h"

#include "indexnode.h"
#include "leafnode.h"
#include "math.h"

long CNode::CSplitEntry::m_sortDim;

CNode::CNode(long id)
    : CEntry(id, NULL, 0, 0.0),
    m_level(-1),
    m_rtree(NULL),
    m_size(-1)
{
  // m_mbr = new CRectangle();
  // m_fm = new CFMData();
}

CNode::CNode(long level, CRTree* rtree)
    : CEntry(0, NULL, 0, 0.0),
    m_level(level),
    m_rtree(rtree),
    m_size(0)
{
  m_mbr = new CRectangle();
}

CNode::~CNode(void)
{
}

bool CNode::isLeaf()
{
  return(m_level == 0);
}

bool CNode::insertData(CEntry* e, CStack<CNode*>& path, bool* overflow)
{
  /*
    insert algorithm:
    1. if(there is space in this node) then
         // case 1: normal insertion
    2.   do insert
    3. else if(reinsertion is not done in this level)
         // case 2: reinsertion
    4.   choose reinsertion nodes of this node
    5.   reinsert all reinsertion nodes
    6. else
         // case 3: split
      7.   split the node to two
    8.   insert the new node (two nodes when the root is splited)
  */

  if (m_size < m_rtree->m_nodeCapacity) // case 1: normal insertion
  {
    bool adjusted = false;

    CRectangle* oldMBR = new CRectangle(m_mbr);

    insertEntry(e);

    if (!path.empty())
    {
      // adjustTree(this, oldMBR, path, e, false);
      adjustTree(this, oldMBR, path, e, true);
      adjusted = true;
    }

    delete oldMBR;

    return adjusted;
  }
  else if (!path.empty() && overflow[m_level] == false) // case 2: reinsert
  {
    overflow[m_level] = true;

    vector<long> reinsert;
    vector<long> remain;

    chooseReinsertNode(e->m_mbr, reinsert, remain);

    //xxxxxxxxxxxxxxxxxx
    /*
    for ( int i=0; i< reinsert.size(); i++ )
      cout << reinsert[i] << " " ;

    cout << endl;

    for ( int i=0; i< remain.size(); i++ )
    cout << remain[i] << " " ;

    cout << endl;
    */
    //xxxxxxxxxxxxxxxxx

    vector<CEntry*> entries;

    for (long i = 0; i < m_rtree->m_nodeCapacity; ++ i)
    {
      entries.push_back(getEntry(i));
    }
    entries.push_back(e);

    if (isLeaf())
    {
      for (long i = 0; i < remain.size(); ++ i)
      {
        ((CLeafNode*)this)->m_entries[i] = entries[remain[i]];
      }
    }
    else
    {
      for (long i = 0; i < remain.size(); ++ i)
      {
        ((CIndexNode*)this)->m_entries[i] = (CNode*)entries[remain[i]];
      }
    }

    m_size = remain.size();


    CRectangle* oldMBR = new CRectangle(m_mbr);

    // update mbr, the number of data points, and fm sketch
    m_mbr->setEmpty();
    m_dataCnt = 0;
    
    //added by Bin
    m_p = 0.0;

    for (long i = 0; i < m_size; ++ i)
    {
      m_mbr->combine(getEntry(i)->m_mbr);

      m_dataCnt += getEntry(i)->m_dataCnt;
      
      //added by Bin
      m_p += getEntry(i)->m_p;
    }

    adjustTree(this, oldMBR, path, this, true);

    delete oldMBR;

    for (long i = 0; i < reinsert.size(); ++ i)
    {
      m_rtree->insertData(entries[reinsert[i]], m_level, overflow);

      //xxxxxxxxxxxxxxxx
      // cout << "reinsert dump ................................. " << endl;
      // m_rtree->dump( cout , m_rtree->getRoot() );
      //xxxxxxxxxxxxxxxx

    }

    return true;
  }
  else  // case 3: split
  {
    CNode* right = split(e);

    vector<CEntry*> entries;

    if (path.empty()) // root is split
    {
      CIndexNode* newRoot = new CIndexNode(m_level + 1, m_rtree);

      newRoot->insertEntry(this);
      newRoot->insertEntry(right);

      m_rtree->m_root = newRoot;
    }
    else
    {
      CNode* parent = path.top();
      path.pop();

      parent->adjustTree(this, right, path, overflow);
    }

    return true;
  }
}

void CNode::insertEntry(CEntry* e)
{
  if (isLeaf())
  {
    ((CLeafNode*)this)->m_entries[m_size] = e;
  }
  else
  {
    ((CIndexNode*)this)->m_entries[m_size] = (CNode*)e;
  }
  m_size ++;

  // update mbr
  m_mbr->combine(e->m_mbr);

  // update the number of data points
  m_dataCnt += e->m_dataCnt;
  
  // added by Bin
  m_p += e->m_p;
}

void CNode::adjustTree(CNode* node, CRectangle* oldMBR, CStack<CNode*>& path,
                       CEntry* e, bool recompute)
{
  bool doUpdateMBR = true;
  CIndexNode* parent;
  CNode* current = node;

  while (!path.empty())
  {
    parent = (CIndexNode*)(path.top());
    path.pop();

    /*
    if(doUpdateMBR)
    {
    if(parent->m_mbr->isContain(current->m_mbr))
    {
    doUpdateMBR = false;
    }
    //      else if(!(parent->m_mbr->isTouch(oldMBR)))
    //      {
    //        doUpdateMBR = false;
    //      }
      
    if(doUpdateMBR)
    {
    oldMBR->clone(parent->m_mbr);
    parent->m_mbr->setEmpty();

    for(long i = 0; i < parent->m_size; ++ i)
    {
    parent->m_mbr->combine(parent->m_entries[i]->m_mbr);
    }
    }
    }
    */

    // always update parent mbr, it is not final one
    oldMBR->clone(parent->m_mbr);
    parent->m_mbr->setEmpty();
    for (long i = 0; i < parent->m_size; ++ i)
    {
      parent->m_mbr->combine(parent->m_entries[i]->m_mbr);
    }


    if (recompute)
    {
      parent->m_dataCnt = 0;
      
      //added by Bin
      parent->m_p = 0.0;

      for (long i = 0; i < parent->m_size; ++ i)
      {
        parent->m_dataCnt += parent->m_entries[i]->m_dataCnt;
        
        // added by Bin
        parent->m_p += parent->m_entries[i]->m_p;
      }
    }
    else
    {
      // update the number of data points
      parent->m_dataCnt += e->m_dataCnt;
      
      //added by Bin
      parent->m_p += e->m_p;
    }

    current = parent;
  }
}

void CNode::adjustTree(CNode* curNode, CNode* newNode, CStack<CNode*>& path,
                       bool* overflow)
{
  m_mbr->setEmpty();
  m_dataCnt = 0;
  
  //added by Bin
  m_p = 0.0;

  for (long i = 0; i < m_size; ++ i)
  {
    m_mbr->combine(getEntry(i)->m_mbr);
    m_dataCnt += ((CNode*)getEntry(i))->m_dataCnt;
    
    //added by Bin
    m_p += ((CNode*)getEntry(i))->m_p;
  }

  insertData(newNode, path, overflow);
}

void CNode::chooseReinsertNode(CRectangle* mbr,
                               vector<long>& reinsert, vector<long>& remain)
{
  CRectangle* center = new CRectangle();
  m_mbr->getCenter(center);

  //xxxxxxxxxxxxxxxxxxxxxxx
  /*
  cout << "center print" << endl;
  center->dump(cout);
  cout << endl;
  */

  CReinsertEntry** v = new CReinsertEntry*[m_rtree->m_nodeCapacity + 1];

  CRectangle* ctr = new CRectangle();
  for (long i = 0; i < m_rtree->m_nodeCapacity; ++ i)
  {
    if (isLeaf())
    {
      ((CLeafNode*)this)->m_entries[i]->m_mbr->getCenter(ctr);
    }
    else
    {
      ((CIndexNode*)this)->m_entries[i]->m_mbr->getCenter(ctr);
    }

		//xxxxxxxxxxxxxxxxx
		/*    cout.precision(10);
			ctr->dump(cout);
			cout << endl;
			cout << center->getDist(ctr) << endl;
		*/

    v[i] = new CReinsertEntry(i, center->getDist(ctr));
  }

  mbr->getCenter(ctr);
  v[m_rtree->m_nodeCapacity] = new CReinsertEntry(m_rtree->m_nodeCapacity,
                               center->getDist(ctr));

  qsort(v, m_rtree->m_nodeCapacity + 1, sizeof(CReinsertEntry*),
        CReinsertEntry::compare);

  long reinsertCnt = (long)ceil((m_rtree->m_nodeCapacity + 1.0) *
                                m_rtree->m_reinsertFactor);

  for ( int i = m_rtree->m_nodeCapacity ; i >  m_rtree->m_nodeCapacity -
reinsertCnt; i--  )
  {
    reinsert.push_back(v[i]->m_index);
    delete v[i];
  }
  for ( int i=m_rtree->m_nodeCapacity - reinsertCnt; i >=0 ; i--  )
  {
    remain.push_back(v[i]->m_index);
    delete v[i];
  }

  /*
  for(long i = 0; i < reinsertCnt; ++ i)
  // for(long i =reinsertCnt-1  ; i >= 0  ;  i-- )
  {
    reinsert.push_back(v[i]->m_index);

    delete v[i];
  }

  for(long i = reinsertCnt; i < m_rtree->m_nodeCapacity + 1; ++ i)
  {
    remain.push_back(v[i]->m_index);
    delete v[i];
  }
  */
  // yingz ++

  delete center;
  delete ctr;

  delete[] v;
}

CNode* CNode::split(CEntry* e)
{
  long capacity = m_rtree->m_nodeCapacity;

  long spf = long((capacity + 1.0) * m_rtree->m_splitFactor);
  long splitF = (capacity + 1) - (2 * spf) + 2;

  CSplitEntry** dataLow = new CSplitEntry*[capacity + 1];
  CSplitEntry** dataHigh = new CSplitEntry*[capacity + 1];

  CNode* child;
  vector<CNode*> nodes;

  for (long i = 0; i < capacity; ++ i)
  {
    dataLow[i] = new CSplitEntry(i, getEntry(i)->m_mbr);
    dataHigh[i] = dataLow[i];
  }
  dataLow[capacity] = new CSplitEntry(capacity, e->m_mbr);
  dataHigh[capacity] = dataLow[capacity];

  double minMargin = numeric_limits<double>::max();
  double marginLow = 0.0;
  double marginHigh = 0.0;

  long splitAxis = -1;
  long sortOrder = -1;

  CRectangle* mbrLow1 = new CRectangle();
  CRectangle* mbrLow2 = new CRectangle();
  CRectangle* mbrHigh1 = new CRectangle();
  CRectangle* mbrHigh2 = new CRectangle();

  for (long i = 0; i < DIMENSION; ++ i)
  {

    marginLow = 0.0;
    marginHigh = 0.0;
    CSplitEntry::m_sortDim = i;

    qsort(dataLow, capacity + 1, sizeof(CSplitEntry*),
          CSplitEntry::compareLow);
    qsort(dataHigh, capacity + 1, sizeof(CSplitEntry*),
          CSplitEntry::compareHigh);

    for (long j = 1; j <= splitF; ++ j)
    {

      long k = spf - 1 + j;
      mbrLow1->setEmpty();
      mbrLow2->setEmpty();
      mbrHigh1->setEmpty();
      mbrHigh2->setEmpty();

      for (long l = 0; l < k; ++ l)
      {
        mbrLow1->combine(dataLow[l]->m_mbr);
        mbrHigh1->combine(dataHigh[l]->m_mbr);
      }

      for (long l = k; l < capacity + 1; ++ l)
      {
        mbrLow2->combine(dataLow[l]->m_mbr);
        mbrHigh2->combine(dataHigh[l]->m_mbr);
      }

      marginLow += mbrLow1->getPerimeter() + mbrLow2->getPerimeter();
      marginHigh += mbrHigh1->getPerimeter() + mbrHigh2->getPerimeter();
    }

    double margin = min(marginLow, marginHigh);

    if (margin < minMargin)
    {
      minMargin = margin;
      splitAxis = i;
      sortOrder = (marginLow < marginHigh) ? 0 : 1;
    }
  }

  CSplitEntry::m_sortDim = splitAxis;

  if (sortOrder == 0)
  {
    qsort(dataLow, capacity + 1, sizeof(CSplitEntry*),
          CSplitEntry::compareLow);
  }
  else
  {
    qsort(dataLow, capacity + 1, sizeof(CSplitEntry*),
          CSplitEntry::compareHigh);
  }

  double minArea = numeric_limits<double>::max();
  double minOverlap = numeric_limits<double>::max();

  long splitPoint = -1;

  for (long i = 1; i <= splitF; ++ i)
  {
    mbrLow1->setEmpty();
    mbrLow2->setEmpty();

    long j = spf - 1 + i;

    for (long k = 0; k < j; ++ k)
    {
      mbrLow1->combine(dataLow[k]->m_mbr);
    }

    for (long k = j; k < capacity + 1; ++ k)
    {
      mbrLow2->combine(dataLow[k]->m_mbr);
    }

    double overlap = mbrLow1->getIntersectArea(mbrLow2);

    if (overlap < minOverlap)
    {
      minOverlap = overlap;
      minArea = mbrLow1->getArea() + mbrLow2->getArea();
      splitPoint = i;
    }
    else if (overlap == minOverlap)
    {
      double area = mbrLow1->getArea() + mbrLow2->getArea();
      if (area < minArea)
      {
        minArea = area;
        splitPoint = i;
      }
    }
  }
  
  //added by Bin to prevent error when inserting rectangle
  if (splitPoint == capacity + 1)
  {
      splitPoint /= 2;
  }
  
  delete mbrLow1;
  delete mbrLow2;
  delete mbrHigh1;
  delete mbrHigh2;

  vector<long> g1;
  vector<long> g2;

  for (long i = 0; i < spf - 1 + splitPoint; ++ i)
  {
     g1.push_back(dataLow[i]->m_index);
     delete dataLow[i];
  }

  for (long i = spf - 1 + splitPoint; i < capacity + 1; ++ i)
  {
     g2.push_back(dataLow[i]->m_index);
     delete dataLow[i];
  }

  delete[] dataLow;
  delete[] dataHigh;

  vector<CEntry*> entries;

  for (long i = 0; i < m_rtree->m_nodeCapacity; ++ i)
  {
    entries.push_back(getEntry(i));
  }
  entries.push_back(e);

  // prepare left node, here left node is the current node
  if (isLeaf())
  {
    for (long i = 0; i < g1.size(); ++ i)
    {
      ((CLeafNode*)this)->m_entries[i] = entries[g1[i]];
    }
  }
  else
  {
    for (long i = 0; i < g1.size(); ++ i)
    {
      ((CIndexNode*)this)->m_entries[i] = (CNode*)entries[g1[i]];
    }
  }

  m_size = g1.size();
  // update mbr, the number of data points, and fm sketch
  m_mbr->setEmpty();
  m_dataCnt = 0;
  
  //added by Bin
  m_p = 0.0;

  for (long i = 0; i < m_size; ++ i)
  {
    m_mbr->combine(getEntry(i)->m_mbr);
    m_dataCnt += getEntry(i)->m_dataCnt;
    //added by Bin
    m_p += getEntry(i)->m_p;
  }

  // prepare right node, here right node is a new node
  CNode* right;
  if (isLeaf())
  {
		right = new CLeafNode(m_rtree);
		for (long i = 0; i < g2.size(); ++ i)
		{
			((CLeafNode*)right)->m_entries[i] = entries[g2[i]];
		}
  }
  else
  {
    right = new CIndexNode(m_level, m_rtree);
    for (long i = 0; i < g2.size(); ++ i)
    {
        ((CIndexNode*)right)->m_entries[i] = (CNode*)entries[g2[i]];
    }
  }

  right->m_size = g2.size();

  // update mbr, the number of data points, and fm sketch
  for (long i = 0; i < right->m_size; ++ i)
  {
		right->m_mbr->combine(right->getEntry(i)->m_mbr);
		right->m_dataCnt += right->getEntry(i)->m_dataCnt;    
		//added by Bin
		right->m_p += right->getEntry(i)->m_p;
  }

  return right;
}

void CNode::dump(ostream& os)
{
	long id = m_id;
	if ( id < 0 ) id = -id;
	
	os << id << " "
	<< m_level << " ";
	m_mbr->dump(os);
	os << endl;
}
