#include "rectangle.h"

#include <limits>
#include <string>
#include <math.h>
#include <float.h>
#include <time.h>


using namespace std;




double CRectangle::m_minCoord[DIMENSION];
double CRectangle::m_maxCoord[DIMENSION];

CRectangle::CRectangle(void)
{
  memcpy(m_low, m_maxCoord, sizeof(double) * DIMENSION);
  memcpy(m_high, m_minCoord, sizeof(double) * DIMENSION);
}

CRectangle::CRectangle(double* coord)
{
  memcpy(m_low, coord, sizeof(double) * DIMENSION);
  memcpy(m_high, coord, sizeof(double) * DIMENSION);
}

// Works only for two dimensional data
CRectangle::CRectangle(int i, int j)
{
  memcpy(m_low, m_maxCoord, sizeof(double) * DIMENSION);
  memcpy(m_high, m_minCoord, sizeof(double) * DIMENSION);
  
  m_low[0] = (double)i/CGlobal::GRID_WIDTH;
  m_low[1] = (double)j/CGlobal::GRID_WIDTH;
  m_high[0] = (double)(i+1)/CGlobal::GRID_WIDTH;
  m_high[1] = (double)(j+1)/CGlobal::GRID_WIDTH;
        
//   cout<< " i  is " << i << " Gridwidth " << CGlobal::GRID_WIDTH << endl;
//   cout << m_low[0] << endl;

}

CRectangle::CRectangle(double* low, double* high)
{
  memcpy(m_low, low, sizeof(double) * DIMENSION);
  memcpy(m_high, high, sizeof(double) * DIMENSION);
}

CRectangle::CRectangle(CRectangle* mbr)
{
  memcpy(m_low, mbr->m_low, sizeof(double) * DIMENSION);
  memcpy(m_high, mbr->m_high, sizeof(double) * DIMENSION);
}

CRectangle::CRectangle(CPoint* center, double side_length)
{
  memcpy(m_low, m_maxCoord, sizeof(double) * DIMENSION);
  memcpy(m_high, m_minCoord, sizeof(double) * DIMENSION);
  for(int i=0; i < DIMENSION; i++)
  {
    
    m_low[i] = center->coord[i] - side_length/2;
    m_high[i] = center->coord[i] + side_length/2;
  }
}


CRectangle::~CRectangle(void)
{
}

double* CRectangle::getLow(void)
{
  return m_low;
}

double CRectangle::getLow(long index)
{
  return m_low[index];
}

double* CRectangle::getHigh(void)
{
  return m_high;
}

double CRectangle::getHigh(long index)
{
  return m_high[index];
}

void CRectangle::initMinMaxCoord(void)
{
  for (long i = 0; i < DIMENSION; ++ i)
  {
    CRectangle::m_minCoord[i] = -numeric_limits<double>::max();
    CRectangle::m_maxCoord[i] = numeric_limits<double>::max();
  }
}

void CRectangle::combine(CRectangle* mbr)
{
  for (long i = 0; i < DIMENSION; ++ i)
  {
    m_low[i] = min(m_low[i], mbr->m_low[i]);
    m_high[i] = max(m_high[i], mbr->m_high[i]);
  }
}

void CRectangle::combine(CPoint* p)
{
  for (long i = 0; i < DIMENSION; ++ i)
  {
    m_low[i] = min(m_low[i], p->coord[i]);
    m_high[i] = max(m_high[i], p->coord[i]);
  }
}

void CRectangle::combine(CRectangle* mbrA, CRectangle* mbrB)
{
  for (long i = 0; i < DIMENSION; ++ i)
  {
    m_low[i] = min(mbrA->m_low[i], mbrB->m_low[i]);
    m_high[i] = max(mbrA->m_high[i], mbrB->m_high[i]);
  }
}

void CRectangle::setMBR(double* low )
{
    memcpy( m_low, low , sizeof(double)*DIMENSION );
    memcpy( m_high, low , sizeof(double)*DIMENSION );
}

void CRectangle::setMBR(double* low , double* high)
{
    memcpy( m_low, low , sizeof(double)*DIMENSION );
    memcpy( m_high, high , sizeof(double)*DIMENSION );
}



double CRectangle::getArea(void)
{
  double area = 1.0;
  for (long i = 0; i < DIMENSION; ++ i)
  {
    area *= m_high[i] - m_low[i];
  }

  return area;
}

double CRectangle::getIntersectArea(CRectangle* mbr)
{
  double area = 1.0;
  for (long i = 0; i < DIMENSION; ++ i)
  {
    if (m_low[i] >= mbr->m_high[i] ||
            m_high[i] <= mbr->m_low[i])   // do not insert
    {
      return 0.0;
    }
    area *= min(m_high[i], mbr->m_high[i]) -
            max(m_low[i], mbr->m_low[i]);
  }    
  return area;    
  
}

bool CRectangle::isContain(CRectangle* mbr)
{
  for (long i = 0; i < DIMENSION; ++ i)
  {
    if (m_low[i] > mbr->m_low[i] ||
            m_high[i] < mbr->m_high[i])
    {
      return false;
    }
  }
  
  return true;  
}

bool CRectangle::isTouch(CRectangle* mbr)
{
  for (long i = 0; i < DIMENSION; i ++)
  {
    if (m_low[i] == mbr->m_low[i] ||
            m_high[i] == mbr->m_high[i])
    {
      return true;
    }
  }

  return false;
}

void CRectangle::clone(CRectangle* mbr)
{
  memcpy(m_low, mbr->m_low, sizeof(double) * DIMENSION);
  memcpy(m_high, mbr->m_high, sizeof(double) * DIMENSION);
  
  for( int i = 0 ; i <DIMENSION ; i++)
  {
    if( m_low[i] >= 10 || m_high[i] < 0)
    {
        this->isEmpty = true;
	return;
    }      
  }
  this->isEmpty = false;
}

void CRectangle::clone(double* low, double* high)
{
  memcpy(m_low, low, sizeof(double) * DIMENSION);
  memcpy(m_high, high, sizeof(double) * DIMENSION);
}

void CRectangle::setEmpty(void)
{
  memcpy(m_low, CRectangle::m_maxCoord, sizeof(double) * DIMENSION);
  memcpy(m_high, CRectangle::m_minCoord, sizeof(double) * DIMENSION);
}

void CRectangle::getCenter(CRectangle* center)
{
  for (long i = 0; i < DIMENSION; ++ i)
  {
    center->m_low[i] = (m_low[i] + m_high[i]) / 2.0;
//    center->m_high[i] =  center->m_low[i];
  }
}

void CRectangle::getCenter(CPoint* center)
{
  for (long i = 0; i < DIMENSION; ++ i)
  {
    center->coord[i] = (m_low[i] + m_high[i]) / 2.0;
//    center->m_high[i] =  center->m_low[i];
  }
}

double CRectangle::getDist(CRectangle* mbr)
{
  double dist = 0.0;
  
  for (long i = 0; i < DIMENSION; ++ i)
  {
    dist += pow( double (this->getLow(i) - mbr->getLow(i)) , 2.0);
  }
  return dist;
}

double CRectangle::getMinDist(CRectangle* mbr)
{
  double dist = 0.0;
  double min_i = 0.0; // minimum distance in dimension i
  for (long i = 0; i < DIMENSION; ++ i)
  {
    // if min_i is positive it means rectangles intersect in ith dimension
    min_i = min( m_high[i] - mbr->m_low[i] , mbr->m_high[i]-m_low[i] );
    if(min_i >= 0)
    {
      dist += 0;
    }
    else
    {
      dist += pow( double (min_i) , 2.0);
    }
    
  }
  return dist;
}

double CRectangle::getMaxDist(CRectangle* mbr)
{
  double dist = 0.0;
  double max_i = 0.0; // maximum distance in dimension i
  for (long i = 0; i < DIMENSION; ++ i)
  {
    max_i = max( m_high[i] - mbr->m_low[i] , mbr->m_high[i]-m_low[i] );
    dist += pow( double (max_i) , 2.0);
  }
  return dist;
}


double CRectangle::getPerimeter(void)
{
  double mul = pow(2.0, DIMENSION - 1.0);
  double peri = 0.0;

  for (long i = 0; i < DIMENSION; ++ i)
  {
    peri += (m_high[i] - m_low[i]) * mul;
  }

  return peri;
}

void CRectangle::dump(ostream& os)
{
  os << "[";
  for (long i = 0; i < DIMENSION - 1; ++ i)
  {
    os << m_low[i] << ", ";
  }
  os << m_low[DIMENSION - 1] << "] X [";
  for(long i = 0; i < DIMENSION - 1; ++ i)
  {
    os << m_high[i] << ", ";
  }
  os << m_high[DIMENSION - 1] << "]";
}

long CRectangle::relativePos(CRectangle* rect, const double* coord)
{
  switch (relativePos(rect->m_low, coord))
  {
  case 1:
    return(relativePos(rect->m_high, coord));
    break;
  case 2:
    return(relativePos(rect->m_high, coord) + 2);
    break;
  case 3:
    return 6;
    break;
  }

  return -1;  // if return follows this path, something wrong
}

long CRectangle::relativePos(const double* point, const double* coord)
{
  bool isFwDom = true;
  bool isBwDom = true;

  for (long i = 0; i < DIMENSION && (isFwDom || isBwDom); i ++)
  {
    if (point[i] - coord[i] > numeric_limits<double>::epsilon())
    {
      isFwDom = false;
    }
    else if (point[i] - coord[i] < -numeric_limits<double>::epsilon())
    {
      isBwDom = false;
    }
  }

  if (isFwDom)
  {
    return 1;
  }
  if (isBwDom)
  {
    return 3;
  }
  return 2;
}

double CRectangle::getXY()
{
  double ret = 0.0;

  for ( int i=0; i < DIMENSION ; i++ )
    ret+= m_low[i];

  return ret;
}

double CRectangle::getLXY() const
{
  double ret = 0.0;
  for ( int i=0; i < DIMENSION ; i++ )
    ret+= m_low[i];
  return ret;
}

double CRectangle::getL2XY() const
{
  double ret = 0.0;
  for ( int i=0; i < DIMENSION ; i++ )
    ret+= m_low[i]*m_low[i];

  return ret;

}

double CRectangle::getLfacXY( long factor ) const
{
    double ret = 0.0;
    for ( int i=0; i < DIMENSION ; i++ )
         ret+= pow( m_low[i], factor);
    return ret;
}


double CRectangle::getMXY()
{
    double ret = 1.0;
    for ( int i=0; i < DIMENSION ; i++ )
        ret*= m_low[i];
    return ret;
}

double CRectangle::getMaxHigh() const
{
    double max = -MAX_FLOAT;
    for ( int i=0; i< DIMENSION; i++ )
        if ( m_high[i]> max ) max = m_high[i];
    
    return max;    
}

double CRectangle::getMinLow() const
{
    double min = MAX_FLOAT;    
    for ( int i=0; i< DIMENSION; i++ )
    {
        if ( m_low[i] < min ) min=m_low[i];    
    }
    return min;   
}


double CRectangle::getHXY() const
{
  double ret = 0.0;

  for ( int i=0; i < DIMENSION ; i++ )
    ret+= m_high[i];
  return ret;
}

void CRectangle::resetRectangle(void)
{
  for (int i = 0; i < DIMENSION; ++i)
  {
    m_low[i] = -FLT_MAX;
    m_high[i] = FLT_MAX;
  }
}

bool CRectangle::prunedByBisector(CEquation* eqn)
{

  float z[DIMENSION];
  
  for(int i=0; i<DIMENSION; i++)
  {
    if(eqn->getCoeff(i)>0)
      z[i] = this->getHigh(i);
    else
      z[i] = this->getLow(i);
  }
  
  
  // To calculate d = c -Ai.z, first calculate Ai.z
  
  float Aiz = 0;
  for(int i=0; i<DIMENSION; i++)
  {
    Aiz = Aiz + eqn->getCoeff(i)*z[i];
  }
  
  // Now calculate d = c- Ai.z
  float d = eqn->getCoeff(DIMENSION) - Aiz;
  
  if(d>0)
    return true;  // The whole rectangle can be pruned (i.e; Unpruned rectangle is empty )
  else
    return false;
}

bool CRectangle::bisectorTrim(CEquation* eqn, CRectangle* mbr)
{

  double z[DIMENSION];
  
  for(int i=0; i<DIMENSION; i++)
  {
    if(eqn->getCoeff(i)>0)
      z[i] = this->getHigh(i);
    else
      z[i] = this->getLow(i);
  }
  
  
  // To calculate d = c -Ai.z, first calculate Ai.z
  
  double Aiz = 0;
  for(int i=0; i<DIMENSION; i++)
  {
    Aiz = Aiz + eqn->getCoeff(i)*z[i];
  }
  
  // Now calculate d = c- Ai.z
  double d = eqn->getCoeff(DIMENSION) - Aiz;
  
  if(d>0)
  {
//      for(int i=0; i<DIMENSION; i++)
//     {
//       if(eqn->getCoeff(i)==0)
//       {
//         mbr->m_low[i] = this->getLow(i);
//         mbr->m_high[i] = this->getHigh(i);
//       }
//       
//       if(eqn->getCoeff(i)>0)
//       {
//         mbr->m_high[i] = this->getHigh(i);
//         mbr->m_low[i] = max(this->getLow(i), this->getHigh(i) +  d/eqn->getCoeff(i));
//       }
//       
//       if(eqn->getCoeff(i)<0)
//       {
//         mbr->m_low[i] = this->getLow(i);
//         mbr->m_high[i] = min(this->getHigh(i), this->getLow(i) +  d/eqn->getCoeff(i));
//       }
//     }
    
    
    
    
    mbr->isEmpty = true;
     
    return true;  // The whole rectangle can be pruned (i.e; Unpruned rectangle is empty )
  }
  else
  {
    for(int i=0; i<DIMENSION; i++)
    {
      if(eqn->getCoeff(i)==0)
      {
        mbr->m_low[i] = this->getLow(i);
        mbr->m_high[i] = this->getHigh(i);
      }
      
      if(eqn->getCoeff(i)>0)
      {
        mbr->m_high[i] = this->getHigh(i);
        mbr->m_low[i] = max(this->getLow(i), this->getHigh(i) +  d/eqn->getCoeff(i));
      }
      
      if(eqn->getCoeff(i)<0)
      {
        mbr->m_low[i] = this->getLow(i);
        mbr->m_high[i] = min(this->getHigh(i), this->getLow(i) +  d/eqn->getCoeff(i));
      }
    }
    
    mbr->isEmpty = false;    //The rect can not be pruned.
    return false;
  }
  
  
}

void CRectangle::print(void)
{
  for (int i=0; i<DIMENSION; i++)
  {
     cout << "[" << m_low[i] << "," << m_high[i] <<"]";
     if (i+1!=DIMENSION)
       cout << " X ";
  }
  cout << endl;
  
}

// Uses ALL bisectors between all corners of Q and mbr (not optimized)
bool CRectangle::pruneByMBR(CRectangle* Q, CRectangle* mbr)
{
  bool pruned = true;  // pruned is true if this->rectangle is completely pruned
  
  // If one of the bisectors doesn't trim any part of MBR we don't need to test 
  // other bisectors because trimmed part is intersection of all half planes
  bool furtherTrimsNeeded = true; 
  
  if(Q->isIntersected(mbr))
  {
    // Q and mbr intersect each other hence nothing can be pruned
     return false;
  }
  else
  {
    CEquation* equation;
    int numCorners = (int)pow(2,DIMENSION);
    
    CPoint* QCorners[numCorners];
    Q->getCorners(QCorners);
    
    CPoint* mbrCorners[numCorners];
    mbr->getCorners(mbrCorners);
    
    CRectangle* unPruned = new CRectangle(); // stores the unPruned rectangle by each bisector
    CRectangle* final = new CRectangle();   // stores the MBR of all unPruned rectangles
    
    for(int i=0; i<numCorners; i++)
    {
      for(int j=0; j<numCorners; j++)
      {
        
        equation = new CEquation(QCorners[i],mbrCorners[j]);
        
        if(!this->bisectorTrim(equation,unPruned)) // if whole rectangle is pruned do nothing;
        {
          pruned = false;
          
          if(this->equals(unPruned))
          {
             furtherTrimsNeeded = false;
             break;
          }
          
          for(int k=0; k<DIMENSION; k++)
          {
            if(unPruned->getLow(k) < final->getLow(k))
            {
              final->m_low[k] = unPruned->getLow(k);
            }
            if(unPruned->getHigh(k) > final->getHigh(k))
            {
              final->m_high[k] = unPruned->getHigh(k);
            }
          } 
        }
        else
        {
          // Whole rectangle is pruned do nothing;
        }
      }
      if(!furtherTrimsNeeded)
      {
        break;
      }
    }
    // if furtherTrimsNeeded is true this means that some part of rectangle is pruned
    if(furtherTrimsNeeded)
       this->clone(final);
    return pruned;
  }
}

bool CRectangle::pruneByInstance(CPoint* qPoint, CRectangle* mbr)
{
  bool pruned = true;  // pruned is true if this->rectangle is completely pruned
  // If one of the bisectors doesn't trim any part of MBR we don't need to test 
  // other bisectors because trimmed part is intersection of all half planes
  bool furtherTrimsNeeded = true; 
  
  if(mbr->isIntersected(qPoint))
  {
    // q and mbr intersect each other hence nothing can be pruned
    return false;
  }
  else
  {
    
    CEquation* equation;
    int numCorners = (int)pow(2,DIMENSION);
    
    CPoint* mbrCorners[numCorners];
    mbr->getCorners(mbrCorners);
    
    CRectangle* unPruned = new CRectangle(); // stores the unPruned rectangle by each bisector
    CRectangle* final = new CRectangle();   // stores the MBR of all unPruned rectangles
    
    
    for(int j=0; j<numCorners; j++)
    {
      
      equation = new CEquation(qPoint,mbrCorners[j]);
      
      if(!this->bisectorTrim(equation,unPruned)) // if whole rectangle is pruned do nothing;
      {
        pruned = false;
        if(this->equals(unPruned))
        {
          furtherTrimsNeeded = false;
          break;
        }
        
        for(int k=0; k<DIMENSION; k++)
        {
          if(unPruned->getLow(k) < final->getLow(k))
          {
            final->m_low[k] = unPruned->getLow(k);
          }
          if(unPruned->getHigh(k) > final->getHigh(k))
          {
            final->m_high[k] = unPruned->getHigh(k);
          }
        } 
      }
      else
      {
        // Whole rectangle is pruned do nothing;
      }
    }
    if(furtherTrimsNeeded)
      this->clone(final);
    return pruned;
  }
}

bool CRectangle::isPrunedByMBR(CRectangle* Q, CRectangle* mbr)
{
  bool pruned = true;  // pruned is true if this->rectangle is completely pruned
  
  if(Q->isIntersected(mbr))
  {
    // Q and mbr intersect each other hence nothing can be pruned
    return false;
  }
  else
  {
    CEquation* equation;
    int numCorners = (int)pow(2,DIMENSION);
    
    CPoint* QCorners[numCorners];
    Q->getCorners(QCorners);
    
    CPoint* mbrCorners[numCorners];
    mbr->getCorners(mbrCorners);
    
    CRectangle* unPruned = new CRectangle(); // stores the unPruned rectangle by each bisector
    CRectangle* final = new CRectangle();   // stores the MBR of all unPruned rectangles
    
    for(int i=0; i<numCorners; i++)
    {
      for(int j=0; j<numCorners; j++)
      {
        
        equation = new CEquation(QCorners[i],mbrCorners[j]);
        
        if(!this->bisectorTrim(equation,unPruned)) // if some part of rectangle is unpruned
        {
          pruned = false;
          break;
        }
        else
        {
          // Whole rectangle is pruned do nothing;
        }
      }
      if(!pruned)
      {
        break;
      }
    }
    return pruned;
  }
}

void CRectangle::getCorners(CPoint** corners)
{
  double coord[DIMENSION];
  
  for(int i=0; i<pow(2,DIMENSION); i++)
  {
    int bin = i;
    
    // Get coordinate of ith corner
    for(int j=0; j<DIMENSION; j++)
    {
      if(bin%2==0)
      {
        coord[j] = this->getLow(j);
      }
      else
      {
        coord[j] =  this->getHigh(j);
      }
      bin = bin/2;
    }
    
    CPoint* p = new CPoint(coord);
    corners[i] = p;
    
  }
  
}

bool CRectangle::isIntersected(CRectangle* mbr)
{
  for (int i = 0; i < DIMENSION;  i++ )
  {
    if (m_low[i] > mbr->m_high[i] ||
        m_high[i] < mbr->m_low[i])   // do not insert
    {
      return false;
    }
  }
  return true;    
}

bool CRectangle::isIntersected(CPoint* p)
{
  for (int i = 0; i < DIMENSION;  i++ )
  {
    if (m_low[i] > p->coord[i] ||
        m_high[i] < p->coord[i])   // do not insert
    {
      return false;
    }
  }
  return true;    
}


double CRectangle::getDiagonal(void)
{
  double diagonal = 0.0;
  for(int i = 0 ; i < DIMENSION; i++ )
  {
    diagonal += pow((m_high[i] - m_low[i]),2);
  }
    
  return diagonal;
}

double CRectangle::getMaxDist(CPoint* p)
{
  double dist = 0.0;
  double max_i = 0.0; // maximum distance in dimension i
  for (long i = 0; i < DIMENSION; ++ i)
  {
    max_i = max( fabs(m_high[i] - p->coord[i]) , fabs(m_low[i] - p->coord[i]) );
    dist += pow( double (max_i) , 2.0);
  }
  return dist;
}

double CRectangle::getMinDist(CPoint* p)
{
  double dist = 0.0;
  double min_i = 0.0; // minimum distance in dimension i
  for (long i = 0; i < DIMENSION; ++ i)
  {
    // if min_i is positive it means rectangle intersect the point in ith dimension
    min_i = min( m_high[i] - p->coord[i] , p->coord[i]-m_low[i] );
    if(min_i >= 0)
    {
      dist += 0;
    }
    else
    {
      dist += pow( double (min_i) , 2.0);
    }
    
  }
  return dist;
}

bool CRectangle::equals(CRectangle* mbr)
{
  bool equal = true;
  for(int i = 0; i < DIMENSION; i++)
  {
    if(fabs(m_low[i] - mbr->m_low[i]) > EPSILON || fabs(m_high[i] - mbr->m_high[i]) > EPSILON)
    {
      equal = false;
      break;
    }
  }
  return equal;
}

bool CRectangle::pruneByDominance(CRectangle* mbr, CPoint* bitMapMBR, CPoint* domPoint )
{
  bool pruned = true;  // pruned is true if this->rectangle is completely pruned
       
  // if domPoint lies within cnd mbr, pruning cannot be used
  if(this->isIntersected(domPoint))
  {
    return false;
  }
  
  
  for(int i=0; i<DIMENSION; i++)
  {
    if(bitMapMBR->coord[i] == 2)  // ith dimension dominance  is "smaller the better"
    {
      if(this->m_low[i] >= domPoint->coord[i]) // completely dominated by domPoint    
      {
        // do nothing;
      }
      else
      {
        pruned = false;  
        if(this->m_high[i] > domPoint->coord[i]) // part of this->rec is dominated 
        {
          this->m_high[i] = domPoint->coord[i];
        }
      }
    }
    if(bitMapMBR->coord[i] == 1) // ith dimension dominance is "greater the better" 
    {
      if(this->m_high[i] <= domPoint->coord[i]) // completely dominated by domPoint 
      {
        // do nothing 
      }
      else
      {
        pruned = false;
        if(this->m_low[i] < domPoint->coord[i]) // part of this->rec is dominated 
        {
          this->m_low[i] = domPoint->coord[i];
        }
      }
    }
  }
  return pruned;
}

 bool CRectangle::isOverlapped(CRectangle* mbr)
{
  for(int i = 0; i < DIMENSION; i++)
  {
    if (m_high[i] > mbr->m_low[i] && m_high[i] < mbr->m_high[i])   
    {
      return true;
    }
    if (mbr->m_high[i] > m_low[i] && mbr->m_high[i] < m_high[i])
    {
      return true;
    }   
  
  }
  return false;
}

CPoint* CRectangle::getClosestCorner(CRectangle* mbr)
{
  CPoint* closest = new CPoint();
  
  for(int i=0; i < DIMENSION; i++)
  {
    if(fabs(this->m_low[i] - mbr->m_high[i]) < fabs(this->m_high[i] - mbr->m_low[i]))
    {
      closest->coord[i] = this->m_low[i];
    }
    else
    {
      closest->coord[i] = this->m_high[i];
    }
  
  }
  return closest;
}

CPoint* CRectangle::getClosestCorner(CPoint* P)
{
  CPoint* closest = new CPoint();
  
  for(int i=0; i < DIMENSION; i++)
  {
    if(fabs(this->m_low[i] - P->coord[i]) < fabs(this->m_high[i] - P->coord[i]))
    {
      closest->coord[i] = this->m_low[i];
    }
    else
    {
      closest->coord[i] = this->m_high[i];
    }
  
  }
  return closest;
}



CPoint* CRectangle::getFurthestCorner(CPoint* p)
{
  CPoint* furthest = new CPoint();
  
  for(int i=0; i < DIMENSION; i++)
  {
    if(fabs(p->coord[i] - this->m_high[i]) > fabs(p->coord[i] - this->m_low[i]))
    {
      furthest->coord[i] = this->m_high[i];
    }
    else
    {
      furthest->coord[i] = this->m_low[i];
    }
  
  }
  return furthest;
}
    
CPoint* CRectangle::getDominanceBitMap(CRectangle* mbr)    
{
  CPoint* bitMap = new CPoint();
  
  
  
  for (int i=0; i < DIMENSION; i++)
  {
    bitMap->coord[i] = 3; // for the dimension that overlaps
    if(mbr->m_low[i] >= this->m_high[i])
    {
      bitMap->coord[i] = 2;
    }
    if( mbr->m_high[i] <= this->m_low[i])
    {
      bitMap->coord[i] = 1;
    }
  }
  return bitMap;
}

void CRectangle::passList(list<CEquation*> l)
{
  list<CEquation*>::iterator i;
  for(i = l.begin(); i!=l.end(); i++)
  {
    (*i)->print();
  }

}




int CRectangle::getBisectors(CRectangle* Q, CRectangle* mbr, list<CEquation*> *bisectors)
{
  

  int numOfBisectors = 0;
  
  if(Q->isIntersected(mbr))
  {
    // Q and mbr intersect each other hence nothing can be pruned
    return numOfBisectors;
  }
  int numCorners = (int)pow(2,DIMENSION);
  
//   cout << "Q is " << endl;
//   Q->print();
//   
//   cout << "MBR is " << endl;
//   mbr->print();
  
  
  CPoint* QCorners[numCorners];
  Q->getCorners(QCorners);
  
  int num_of_equations;

  int index=0;
  num_of_equations = numCorners;
  CEquation* equations[num_of_equations];
  CPoint* passesThrough = new CPoint();
  CPoint* antipodalCorner = new CPoint();
  for(int i=0; i<numCorners; i++)
  {
    antipodalCorner = mbr->getOppositeCorner(Q,mbr,QCorners[i]);
//     QCorners[i]->print();
//     antipodalCorner->print();
    
    // determine the point through which the bisector passes
    for(int c = 0; c < DIMENSION; c++)
    {
      if(QCorners[i]->coord[c] <= antipodalCorner->coord[c])
        passesThrough->coord[c] = (mbr->m_high[c] + Q->m_high[c])/2;
      else
        passesThrough->coord[c] = (mbr->m_low[c] + Q->m_low[c])/2;
    }
//     passesThrough->print();
//     cout << "\n";

    CEquation* eq = new CEquation(QCorners[i],antipodalCorner,passesThrough);
    bisectors->push_back(eq);
    numOfBisectors++;
    
  }
   
  return numOfBisectors;

}

int CRectangle::getoverlap(CPoint* p)
{
  if(this->m_low[0] < p->coord[0] && this->m_high[0]>p->coord[0])
    return 1;//overlap x
  if(this->m_low[1] < p->coord[1] && this->m_high[1]>p->coord[1])
    return 2; //overlap y
  else
    return 0; //don't overlap
}

CPoint* CRectangle::getClosestPoint(CPoint* p)
{
  CPoint* closest = new CPoint();
  
  for(int i=0; i < DIMENSION; i++)
  {
    
    // if the point is overlapped in dimension i
    if(p->coord[i] < this->m_high[i] && p->coord[i] > this->m_low[i])
    {
      closest->coord[i] = p->coord[i];
    }
    else
    {
      if(p->coord[i] >= this->m_high[i])
      {
        closest->coord[i] = this->m_high[i];
      }
      if(p->coord[i] <= this->m_low[i])
      {
        closest->coord[i] = this->m_low[i];
      }
    }
  }
  return closest;
}


bool CRectangle::pruneByEquations(CRectangle* Q, list<CEquation*> bisectorsList)
{
  bool pruned = true;  // pruned is true if this->rectangle is completely pruned
  
  // If one of the bisectors doesn't trim any part of MBR we don't need to test 
  // other bisectors because trimmed part is intersection of all half planes
  bool furtherTrimsNeeded = true; 
  
  if(bisectorsList.size()==0)
  {
    // Q and mbr intersect each other hence nothing can be pruned
    return false;
  }
  else
  {
    
    
    CRectangle* unPruned = new CRectangle(); // stores the unPruned rectangle by each bisector
    CRectangle* final = new CRectangle();   // stores the MBR of all unPruned rectangles
    
    list<CEquation*>::iterator i_Eq;
    
      for(i_Eq = bisectorsList.begin(); i_Eq != bisectorsList.end(); i_Eq++)
      {
        
//         equation = new CEquation(QCorners[i],mbrCorners[j]);
        
        if(!this->bisectorTrim((*i_Eq),unPruned)) // if whole rectangle is pruned do nothing;
        {
          pruned = false;
          
          if(this->equals(unPruned))
          {
            furtherTrimsNeeded = false;
            break;
          }
          
          for(int k=0; k<DIMENSION; k++)
          {
            if(unPruned->getLow(k) < final->getLow(k))
            {
              final->m_low[k] = unPruned->getLow(k);
            }
            if(unPruned->getHigh(k) > final->getHigh(k))
            {
              final->m_high[k] = unPruned->getHigh(k);
            }
          } 
        }
        else
        {
          // Whole rectangle is pruned do nothing;
        }
      }
     
    
    // if furtherTrimsNeeded is true this means that some part of rectangle is pruned
    if(furtherTrimsNeeded)
      this->clone(final);
    return pruned;
  }
}


void CRectangle::printLH(CPoint* p)
{
  cout << "[";
  for (int i=0; i<DIMENSION; i++)
  {
    if(p->coord[i]==this->m_low[i])
    {
      cout << "L";
    }
    else
    {
      assert(p->coord[i]==this->m_high[i]);
      cout << "H";
      
      
    }
     
  
  }
  cout << "]";
}


bool CRectangle::isOppositeCorner(CRectangle* Q, CRectangle* mbr, CPoint* q, CPoint* p)
{
  for (int i=0; i<DIMENSION; i++)
  {
    if(q->coord[i]==Q->m_high[i])
    {
      if(p->coord[i]!=mbr->m_low[i])
      {
        return false;
      }
    }
    else
    {
      if(p->coord[i]!=mbr->m_high[i])
      {
        return false;
      }
    }
  }
  return true;
}

CPoint* CRectangle::getOppositeCorner(CRectangle* Q, CRectangle* mbr, CPoint* q)
{
  
  CPoint* opposite = new CPoint();
  for (int i=0; i<DIMENSION; i++)
  {
    if(q->coord[i]==Q->m_high[i])
    {
      opposite->coord[i] = mbr->m_low[i];
    }
    else
    {
      assert(q->coord[i]==Q->m_low[i]);
      opposite->coord[i] = mbr->m_high[i];
    }
  }
  return opposite;

}


int CRectangle::getBisectors2(CRectangle* Q, CRectangle* mbr, list<CEquation*> *bisectors)
{
  
//  list<CEquation*> bisectors;
  int numOfBisectors = 0;
  
  if(Q->isIntersected(mbr))
  {
    // Q and mbr intersect each other hence nothing can be pruned
//     cout << "\n\n\n\n\n\n\n*************\n\n\n\n\n";
    return numOfBisectors;
  }
  int numCorners = (int)pow(2,DIMENSION);
  
//   cout << "Q is " << endl;
//   Q->print();
//   
//   cout << "MBR is " << endl;
//   mbr->print();
//   
  
  CPoint* QCorners[numCorners];
  Q->getCorners(QCorners);
    
  CPoint* mbrCorners[numCorners];
  mbr->getCorners(mbrCorners);

  CEquation* equations[numCorners*numCorners];
  
  
  
  int index=0;
  for(int i=0; i<numCorners; i++)
  {
    for(int j=0; j<numCorners; j++)
    {
      equations[index] = new CEquation(QCorners[i],mbrCorners[j]);
//       cout << "Eq # " << index << "  " ;
//       QCorners[i]->print();
//       mbrCorners[j]->print();
      equations[index]->moveToOrigin();
      
      // We need to handle special cases where some coefficient is zero
      // to handle this we change zero to some very small value so that our approach works
      equations[index]->fixZeroes();
//       equations[index]->print();
    
      
      index++;
      
    }
  
  }
  
  
  double solveFor[DIMENSION]; // the coefficients of all dimensions
  
  /*
  For each equations e
  for any pow(2,DIMENSION-1) points on e
  if the point lies in subspace of all other equations
  it must not be deleted
  */
  for(int i=0; i<numCorners*numCorners; i++)
  {
//     cout << "testing " << i <<  "\n\n"<< endl;
    bool canBeDeleted;
   
    
    // For all pow(2,DIMENSION-1) points on the equation (set points by setting 1,-1)
    
    for(int test=1; test<2; test++)
    {
    for(int j=0; j<pow(2,DIMENSION-1); j++)
    {
      int tmp=j;
//       cout << ".... subtest "<<  j << endl;
      
      
//       cout << "solverFor[k] is ";  
          // Set solveFor   
      srand ( time(NULL) );
      double put = 786;
      
      
        for(int k=0; k<DIMENSION-1; k++)
        {
        
//         put = put*1.124;
          if(tmp%2==0)
          {
//           solveFor[k] = 100; // set maximum
            solveFor[k] = ((rand() % 10000 + 1)+1)*0.0001; // set maximum
            assert(solveFor[k]>0);
          
          
          }
          if(tmp%2==1)
          {
//           solveFor[k] = -1000; //set minimum
            solveFor[k] = ((rand() % 10000 + 1)*-1)*0.0001; //set minimum
            assert(solveFor[k]<0);
          }
//         cout << solveFor[k] << "  ";
//         cout << solveFor[k] << "  ";
          tmp = tmp/2;
        }
      // solveFor is set
        canBeDeleted = false;
      
        double unKnownDimension =   equations[i]->solveFor(solveFor);
        solveFor[DIMENSION-1] = unKnownDimension;
//       cout << solveFor[DIMENSION-1] << "  " << endl;
         
      
      // For ANY other equations, if the point doesn't lie it's space it can be deleted
        for(int m=0; m<numCorners*numCorners; m++)
        {
          if(m==i)
            continue;
        
          if(!equations[m]->liesInSpace(solveFor))
          {
//           equations[i]->print();
//           cout << " deleted by " << m << endl;
//           equations[m]->print();
          
            canBeDeleted = true;
//          
            break;
          }
        }
        if(!canBeDeleted)
          break;
      }
      if(!canBeDeleted)
        break;
    }
    if(!canBeDeleted)
    {
//       cout << "Bisectors between these points must be kept" << endl;
      int q = i/numCorners;
      int p = i%numCorners;
      
      numOfBisectors ++;
//       cout << "q is " << q << " p is " << p << endl;      
//       QCorners[q]->print();
//       mbrCorners[p]->print();
//       cout << endl;
//       Q->printLH(QCorners[q]);
//       mbr->printLH(mbrCorners[p]);
//       cout << endl;
      assert(Q->isOppositeCorner(Q,mbr,QCorners[q],mbrCorners[p]));
      
      
      /* Find the point from which this bisector must pass
         
      */
      double point[DIMENSION];
      for(int c = 0; c < DIMENSION; c++)
      {
        if(QCorners[q]->coord[c] < mbrCorners[p]->coord[c])
          point[c] = mbr->m_high[c];
        else
          point[c] = mbr->m_low[c];
      }
      CPoint* A = new CPoint(point);
//       cout << "\nA is ";
//       A->print();
//       cout << endl;
      CPoint* B = Q->getClosestPoint(A);
//       cout << "B is ";
//       B->print();
//       cout << endl;
      CPoint* liesAt = A->getCenter(B);
//       cout << "the bisector will lie at ";
//       liesAt->print();
//       cout << endl;
      CEquation* eq = new CEquation(QCorners[q],mbrCorners[p],liesAt);
//       eq->print();
//       Q->print();
//       mbr->print();
      bisectors->push_back(eq);
    
    }
    else
    {
      
    }
    
  }
  
//   cout << "num of bise " << numOfBisectors << endl;
  if(bisectors->size() != numOfBisectors)
  {
    cout << "\n\\n\n\n\n\nBANG!!!!!\n\n\n\n\n\n\n\n" << endl;
  }
  
//   cout << "Number of bisectors is " << bisectors->size() << endl;
  assert(bisectors->size()<=pow(2,DIMENSION));
  
//   if(numOfBisectors == 0)
//   {
//     Q->print();
//     mbr->print();
//   }
  
  return numOfBisectors;

}

CPoint* CRectangle::getFrontierPoint(CRectangle* mbr)
{
  CPoint* fp = new CPoint();
  
  for(int i = 0; i < DIMENSION; i++)
  {
    if(mbr->getLow(i)>= this->getHigh(i))
    {
      fp->coord[i] = (this->getHigh(i) + mbr->getHigh(i))/2;
    }
    if(mbr->getHigh(i) <= this->getLow(i))
    {
      fp->coord[i] = (this->getLow(i) + mbr->getLow(i))/2;
    }
  }
  
  return fp;

}

void CRectangle::blowHuge(void)
{
  
  for(int i = 0 ; i < DIMENSION; i++)
  {
    this->m_low[i] = -INFINITY;
    this->m_high[i] = INFINITY;
  
  }

}



void CRectangle::pointCenteric(CPoint* center, double side_length)
{
  for(int i=0; i < DIMENSION; i++)
  {
    this->m_low[i] = center->coord[i] - side_length/2;
    this->m_high[i] = center->coord[i] + side_length/2;
    
    if(this->m_low[i] < 0)
      this->m_low[i] =0;
    if(this->m_high[i] >= 1)
      this->m_high[i] = SPACE_MAX;
  }

}

void CRectangle::coverWholeSpace()
{
  for(int i=0; i<DIMENSION; i++)
  {
    this->m_low[i] = 0;
    this->m_high[i] = SPACE_MAX;
  }

}



void CRectangle::getMaxMinAngle ( double* minAngle, double* maxAngle, CPoint* p )
{ 
	*minAngle = 500;
	*maxAngle = -500;
	
	CPoint* Corners[4];
	this->getCorners(Corners);
	
	double temp1[4];
	double temp2[4];
	int m=0;
	int n=0;
	
	for(int i = 0; i < 4; i++)
	{
	    float angle = p->getAngle(Corners[i]);
	    if(angle>*maxAngle)
	    {
	      *maxAngle = angle;
	    }
	    
	    if(angle<*minAngle)
	    {
	      *minAngle = angle;
	    }
	    
	    if( angle < 90)
	    {
	      temp1[m] = angle;
	      m++;
	    }
	    else
	    {
	      temp2[n] = angle;
	      n++;
	    }
	}
	
	if( *maxAngle - *minAngle > 180)
	{
	    if( *minAngle < temp1[0] )
	    {
	      *minAngle = temp1[0];
	    }
	    else{
	      *minAngle = temp1[1];
	    }
	    
	    if( *maxAngle > temp2[0] )
	    {
		*maxAngle = temp2[0];
	    }
	    else{
		*maxAngle = temp2[1];
	    }	
	}
	
	 delete Corners[0];
	 delete Corners[1];
	 delete Corners[2];
	 delete Corners[3];
}


// works when rectangle is a square
bool CRectangle::isCell()
{
  if(  fabs((this->m_high[0] - this->m_low[0])*CGlobal::GRID_WIDTH - 1) < EPSILON )
    return true;
  return false;

}

CRectangle* CRectangle::getChild(int index)
{
  CRectangle* rec = new CRectangle(this);
  
  assert(index>0 && index<5);
  
  double x_mid = (this->m_low[0] + this->m_high[0])/2;
  double y_mid = (this->m_low[1] + this->m_high[1])/2;
  
  switch (index)
  {
    case 1:
      rec->m_low[0] = x_mid;
      rec->m_low[1] = y_mid;
      break;
    case 2:
      rec->m_high[0] = x_mid;
      rec->m_low[1] = y_mid;
      break;
    case 3:
      rec->m_high[0] = x_mid;
      rec->m_high[1] = y_mid;
      break;
    case 4:
      rec->m_low[0] = x_mid;
      rec->m_high[1] = y_mid;
      break;
  }
  
  return rec;
  

}

int CRectangle::getCellID()
{
  
  assert(this->isCell());
  
  int cellID = 0;
  cellID = (int)((this->m_low[0] + this->m_high[0])*CGlobal::GRID_WIDTH/2);
  cellID = cellID * CGlobal::GRID_WIDTH + (int)((this->m_low[1] + this->m_high[1])*CGlobal::GRID_WIDTH/2);
  return cellID;
  

}

bool CRectangle::contains(CPoint* p)
{
  for (long i = 0; i < DIMENSION; ++ i)
  {
    if (m_low[i] > p->coord[i] ||
        m_high[i] < p->coord[i])
    {
      return false;
    }
  }
  
  return true;  
}

double CRectangle::getPointDist(CPoint* p)
{
  double dist = 0.0;
  
  for (long i = 0; i < DIMENSION; ++ i)
  {
    dist += pow( double (this->getLow(i) - p->coord[i]) , 2.0);
  }
  return dist;
}


float CRectangle::MinDistAtAngle(CPoint* p, float angle)
{ 
  float minDist = 50000;
  float d;
  CEquation* line = new CEquation(p,angle,false);
  //cout<<"line from q"<<endl;
  //line->print();
  // INTERSECTION WITH LINE 1 (LL,HL)
  CPoint* p1 = new CPoint(m_low[0],m_low[1]);
  CPoint* p2 = new CPoint(m_high[0],m_low[1]);
  CEquation* recLine = new CEquation(p1,p2,false); 
  
  
  
  CPoint* intersection = line->getIntersection(recLine);
  //cout<<"=============="<<endl;
  //cout<<"intersect with Line 1"<<endl;
  //recLine->print();
 // cout<<"intersect points:"<<endl;
 // intersection->print();
//  cout<<endl;
  // intersection lies in between
  if(intersection->coord[0]<=m_high[0] && intersection->coord[0] >= m_low[0])
  {
    d = intersection->getDist(p);
    if(d<minDist)
      minDist = d;
  }
  
  
  delete p1;
  delete p2;
  delete recLine;
  delete intersection;
  

  // INTERSECTION WITH LINE 2 (HL,HH)
  p1 = new CPoint(m_high[0],m_low[1]);
  p2 = new CPoint(m_high[0],m_high[1]);
  recLine = new CEquation(p1,p2,false); 
   
  intersection = line->getIntersection(recLine);
  
  //cout<<"=============="<<endl;
  //cout<<"intersect with Line 2"<<endl;
 // recLine->print();
  //cout<<"intersect points:"<<endl;
  //intersection->print();
  //cout<<endl;
  // intersection lies in between
  if(intersection->coord[1]<=m_high[1] && intersection->coord[1] >= m_low[1])
  {
    d = intersection->getDist(p);
    if(d<minDist)
      minDist = d;
  }
  
  
    delete p1;
  delete p2;
  delete recLine;
  delete intersection;
  
  // INTERSECTION WITH LINE 3 (HH,LH)
  p1 = new CPoint(m_high[0],m_high[1]);
  p2 = new CPoint(m_low[0],m_high[1]);
  recLine = new CEquation(p1,p2,false); 
  
   
  
  intersection = line->getIntersection(recLine);
// cout<<"=============="<<endl;
//  cout<<"intersect with Line 3"<<endl;
 // recLine->print();
 // cout<<"intersect points:"<<endl;
 //intersection->print();
  //cout<<endl;
  
  // intersection lies in between
  if(intersection->coord[0]<=m_high[0] && intersection->coord[0] >= m_low[0])
  {
    d = intersection->getDist(p);
    if(d<minDist)
      minDist = d;
  }
  
  
    delete p1;
  delete p2;
  delete recLine;
  delete intersection;
  
  // INTERSECTION WITH LINE 4 (LH,LL)
  p1 = new CPoint(m_low[0],m_high[1]);
  p2 = new CPoint(m_low[0],m_low[1]);
  recLine = new CEquation(p1,p2,false); 
  
  
   
  
  intersection = line->getIntersection(recLine);
 // cout<<"=============="<<endl;
 // cout<<"intersect with Line 4"<<endl;
 // recLine->print();
 // cout<<"intersect points:"<<endl;
 // intersection->print();
 // cout<<endl;
  
  // intersection lies in between
  if(intersection->coord[1]<=m_high[1] && intersection->coord[1] >= m_low[1])
  {
    d = intersection->getDist(p);
    if(d<minDist)
      minDist = d;
  }
  
  
  
  delete p1;
  delete p2;
  delete recLine;
  delete intersection;
  delete line;

  return minDist;
}



bool CRectangle::liesInAngleRange(CPoint* q, double start_angle, double end_angle)
{
  
  // if the rectangle contains the point then it lies within the range
  if(this->contains(q))
    return true;
  
  
  
  CPoint* Corners[4];
  this->getCorners(Corners);
  
  double minAngle = INFINITY;
  double maxAngle = -1*INFINITY;
  for(int i = 0; i < 4; i++)
  {
    double angle = q->getAngle(Corners[i]);
    if(angle>maxAngle)
      maxAngle = angle;
    if(angle<minAngle)
      minAngle = angle;
  }
  
  // rec start and end angles
  double rec_s;
  double rec_e;
  // check if the angles overlap with zero degree angle
  if(this->m_low[1] <= q->coord[1] && this->m_high[1] >= q->coord[1])
  {
    rec_s = maxAngle;
    rec_e = minAngle;
//     cout << "recAngleRange(" << maxAngle << "," << minAngle << ")" << endl;
    // the angle range also overlaps zero degree
    if(start_angle >= end_angle)
      return true;
    
    // Now start_angle && end_angle cannot contain zero
    if(rec_s >= start_angle  && rec_s <= end_angle )
    {
      return true;
    }
    // if the start_angle is between the rectangle rangel; return true
    if(start_angle >= rec_s && start_angle <= 360)
    {
      return true;
    }
    if(start_angle >= 0 && start_angle <= rec_e)
    {
      return true;
    }
    return false;

  }
  else
  {
    rec_s = minAngle;
    rec_e = maxAngle;
    
//     cout << "recAngleRange(" << minAngle << "," << maxAngle << ")" << endl;
    // if the angle range contains zero degree
    if(start_angle>=end_angle)
    {
      // if the start_angle lies within the rec range
      if(start_angle >= rec_s && start_angle <= rec_e)
      {
        return true;
      }
      // check if rec_s lies within the range
      if(rec_s >= start_angle && rec_s <=360)
      {
        return true;
      }
      if(rec_s >=0 && rec_s <=end_angle)
      {
        return true;
        
      }
      return false;
    }
    else
    {
      // both the angles do not overlap zero degree
      if(start_angle >=rec_s && start_angle <= rec_e)
      {
         return true;
      }
      if(rec_s >= start_angle && rec_s <= end_angle)
      {
        return true;
      }
      return false;
    }
  }
  

}

// unoin this rectangle with rect and the result rectangle stored in newRect
// return false if the newRect is empty

bool CRectangle::unoinRec ( CRectangle* rect , CRectangle* newRect )
{
    
    for( int i = 0 ; i <DIMENSION ; i++)
    {
       newRect->m_low[i] =  min( this->m_low[i], rect->m_low[i] );
       newRect->m_high[i] = max( this->m_high[i], rect->m_high[i]);
    }
    
    for( int j = 0 ; j < DIMENSION ; j++)
    {
        if(  newRect->m_low[j] >= newRect->m_high[j])
	{
	  newRect->isEmpty = true;
	  return false;
	}
    }
    newRect->isEmpty = false;
    return true; 
}


bool CRectangle::checkEmpty()
{
    return isEmpty;
}

